

; Calculating Bspline/Bezier curve procedures
; author : Maciej Guba aka macgub
; web    : http://macgub.vxm.pl



init_long_pipe_derives:
; procedure make circle using all derives vertices
; in: esi - first derive vertex
;     edi - ptr to destination derives
;     ecx - derives vertices count
   push  ebp
   mov   ebp,esp
   sub   esp,128

   .cnt   equ dword[ebp-4]
   .one_d equ dword[ebp-8]
   .sinx  equ dword[ebp-12]
   .cosx  equ dword[ebp-16]
   .siny  equ dword[ebp-20]
   .cosy  equ dword[ebp-24]
   .sinz  equ dword[ebp-28]
   .cosz  equ dword[ebp-32]
   .mx    equ [ebp-68]
   .cnt2  equ dword[ebp-72]
   .eesi  equ dword[ebp-76]
   .eedi  equ dword[ebp-80]
   .init_vert  equ [ebp-96]


   mov   .eesi,esi
   mov   .eedi,edi
;   movups xmm0,[esi]
;   movups .init_vert,xmm0
   fninit
   mov   .cnt,ecx
   shr   .cnt,1
   fldpi
   fidiv .cnt
   fstp  .one_d
   fldz
   fsincos
   fstp  .siny
   fstp  .cosy
   fldz
   fsincos
   fstp  .sinz
   fstp  .cosz

   mov   .cnt2,0
;   mov   ebx,2
;   shr   ecx,1
;  .loop1:
;   mov   .cnt2,0
;   push  ebx
  @@:
   push  ecx
   fld   .one_d
   fimul .cnt2
   fsincos
   fstp  .sinx
   fstp  .cosx

   pushad
   push  .sinx
   push  .cosx
   push  .siny
   push  .cosy
   push  .sinz
   push  .cosz
   lea   edi,.mx
   call  make_rotation_matrix
   popad

   push  ebp
   mov   ecx,1
   mov   esi,.eesi
   mov   edi,.eedi
   lea   ebx,.mx
   call  rotary
   pop   ebp

   add   .eedi,12
   pop   ecx
   add   .cnt2,1
   cmp   ecx,.cnt2
   jnz   @b
;   movups   xmm0,.init_vert
;   addps xmm1,[lpipe_init_add]
;   movups .init_vert,xmm0
;   pop   ebx
;   sub   ebx,1
;   jnz   .loop1

   mov   esp,ebp
   pop   ebp
ret



calc_long_pipe:
; in:
;  eax - bspline one segment steps count, must be divisable by 4
;  ebx - bspline segments count
;  ecx - Bspline_derives ptr, must end redundand
;        first two derives vertices (closed curve)
;  edx - Bspline_vertices ptr, must be allocated enough memory
;  esi - curve type, 0 - bezier, 1 - bspline
    push    ebp
    mov     ebp,esp
    sub     esp,256
    and     ebp,0xfffffff0

    .step4  equ [ebp-16]
    .t1     equ [ebp-32]   ; t1, t2, t3, t4
    .Bspline_one_seg_count equ dword[ebp-36]
    .Bspline_segs_count    equ dword[ebp-40]
    .Bspline_vertices_ptr  equ dword[ebp-44]
    .Bspline_derives_ptr   equ dword[ebp-48]
    .init_t_value          equ [ebp-64]
    .curv_type             equ dword[ebp-68]

    mov  .Bspline_one_seg_count,eax
    mov  .Bspline_segs_count,ebx
    mov  .Bspline_vertices_ptr,edx
    mov  .Bspline_derives_ptr,ecx
    mov  .curv_type,esi



;    mov    ebx,.Bspline_one_seg_count
    ; Bspline one segement steps count,  must be divisable by 4
    cvtsi2ss xmm0,.Bspline_one_seg_count
    rcpss    xmm0,xmm0
    shufps   xmm0,xmm0,0
    movaps   xmm1,xmm0
    mulps    xmm0,[const4]
    movaps   .step4,xmm0
    mulps    xmm1,[const1234]
    movaps   .init_t_value,xmm1

    movaps   .t1,xmm1
    mov      ecx,.Bspline_segs_count
  ;  lea      ecx,[ecx*4-1]
 ;   shl      ecx,2
     ; Bspline segments count
    mov      ebx,.Bspline_derives_ptr  ; must end redundand
                                       ; first two derives vertices
    lea      esi,.t1
    mov      edi,.Bspline_vertices_ptr    ; must be alloc enough mem
 .next_seg:
    movaps   xmm0,.init_t_value
    movaps   .t1,xmm0
    mov      eax,.Bspline_one_seg_count
    shr      eax,2
 .nexxt:
    pushad
    cmp      .curv_type,0
    je       @f
    call     calc_Bspline
    jmp      .dalej
  @@:
    call     calc_bezier_curve  ;
  .dalej:
    popad
    add     edi,48
    movaps  xmm0,.t1
    addps   xmm0,.step4
    movaps  .t1,xmm0
    sub     eax,1
    jnz     .nexxt
    cmp     .curv_type,0
    je      @f
    add     ebx,12 ;  ; offset in Bspline derives list
    jmp    .nn
  @@:
    add     ebx,36   ; offset in Bezier derives list
  .nn:
    sub     ecx,1   ; Bspline segments count
    jnz     .next_seg

    add     esp,256
    pop     ebp
ret


if 1

calc_Bspline:

; in  ebx - derives: dword float x1,y1,z1,x2,y2,z2
;                                x3,y3,z3,x4,y4,z4
;     esi - aligned t1,t2,t3,t4
; out edi - points (vertices) as dword float
;           x1,y1,z1 x 16
;
  push  ebp
  mov   ebp,esp
  sub   esp,256
  and   ebp,0xfffffff0

  .a1   equ [ebp-16]
  .b1   equ [ebp-32]
  .c1   equ [ebp-48]
  .d1   equ [ebp-64]
     ; a = 1/6*(1-3*t+3*t^2-t^3)
     ; b = 1/6*(4-6*t^2+3*t^3)
     ; c = 1/6*(1+3*t+3t^2-3t^3)
     ; d = 1/6*(t^3)
  movaps  xmm0,[const1]
  movaps  xmm1,[esi]
  movaps  xmm2,xmm1
  mulps   xmm2,xmm2
  movaps  xmm3,xmm2
  mulps   xmm3,xmm1
  mulps   xmm2,[const3]
  mulps   xmm1,[const3]
  subps   xmm0,xmm1
  addps   xmm0,xmm2
  subps   xmm0,xmm3
  mulps   xmm0,[constrecipr6]
  movaps  .a1,xmm0
  movaps  xmm0,[const4]
  subps   xmm0,xmm2
  subps   xmm0,xmm2
  movaps  xmm4,xmm3
  mulps   xmm4,[const3]
  addps   xmm0,xmm4
  mulps   xmm0,[constrecipr6]
  movaps  .b1,xmm0
  movaps  xmm0,[const1]
  addps   xmm0,xmm1
  addps   xmm0,xmm2
  subps   xmm0,xmm4
  mulps   xmm0,[constrecipr6]
  movaps  .c1,xmm0
  mulps   xmm3,[constrecipr6]
  movaps  .d1,xmm3


  movss   xmm0,[ebx]  ;[p1x]
  shufps  xmm0,xmm0,0
  mulps   xmm0,.a1       ;   px = a*x1 + b*x2 + c*x4 + d*x4
  movss   xmm1,[ebx+12] ;[p2x]
  shufps  xmm1,xmm1,0
  mulps   xmm1,.b1
  movss   xmm2,[ebx+24] ;[p3x]
  shufps  xmm2,xmm2,0
  mulps   xmm2,.c1
  movss   xmm3,[ebx+36]  ;[p4x]
  shufps  xmm3,xmm3,0
  mulps   xmm3,.d1
  addps   xmm0,xmm1
  addps   xmm2,xmm3
  addps   xmm0,xmm2
  movss   [edi],xmm0    ;  4 x coordinates
  shufps  xmm0,xmm0,00111001b
  movss   [edi+12],xmm0
  shufps  xmm0,xmm0,00111001b
  movss   [edi+24],xmm0
  shufps  xmm0,xmm0,00111001b
  movss   [edi+36],xmm0



  movss   xmm0,[ebx+4]   ;[p1y]
  shufps  xmm0,xmm0,0
  mulps   xmm0,.a1       ;   py = a*y1 + b*y2 + c*py3 + d*y4
  movss   xmm1,[ebx+16]  ;[p2y]
  shufps  xmm1,xmm1,0
  mulps   xmm1,.b1
  movss   xmm2,[ebx+28]  ;[p3y]
  shufps  xmm2,xmm2,0
  mulps   xmm2,.c1
  movss   xmm3,[ebx+40] ;;[p4y]
  shufps  xmm3,xmm3,0
  mulps   xmm3,.d1
  addps   xmm0,xmm1
  addps   xmm2,xmm3
  addps   xmm0,xmm2
  movss   [edi+4],xmm0    ;  4 x coordinates
  shufps  xmm0,xmm0,00111001b
  movss   [edi+16],xmm0
  shufps  xmm0,xmm0,00111001b
  movss   [edi+28],xmm0
  shufps  xmm0,xmm0,00111001b
  movss   [edi+40],xmm0


  movss   xmm7,[ebx+8]  ;[p1z]  ; z coords
  shufps  xmm7,xmm7,0
  mulps   xmm7,.a1
  movss   xmm6,[ebx+20] ;[p2z]
  shufps  xmm6,xmm6,0
  mulps   xmm6,.b1
  movss   xmm5,[ebx+32] ;[p3z]
  shufps  xmm5,xmm5,0
  mulps   xmm5,.c1
  movss   xmm4,[ebx+44]   ;[p4z]
  shufps  xmm4,xmm4,0
  mulps   xmm4,.d1
  addps   xmm4,xmm5
  addps   xmm6,xmm7
  addps   xmm4,xmm6
  movss   [edi+8],xmm4    ;  4 x coordinates
  shufps  xmm4,xmm4,00111001b
  movss   [edi+20],xmm4
  shufps  xmm4,xmm4,00111001b
  movss   [edi+32],xmm4
  shufps  xmm4,xmm4,00111001b
  movss   [edi+44],xmm4

  add     esp,256
  pop     ebp
ret
end if
calc_bezier_curve:
; in  ebx - derives: dword float x1,y1,z1,x2,y2,z2
;                                x3,y3,z3,x4,y4,z4
;     esi - aligned t1,t2,t3,t4
; out edi - points (vertices) as dword float
;           x1,y1,z1 x 16
;
  push  ebp
  mov   ebp,esp
  sub   esp,256
  and   ebp,0xfffffff0

  .a1   equ [ebp-16]
  .b1   equ [ebp-32]
  .c1   equ [ebp-48]
  .d1   equ [ebp-64]

 movaps   xmm0,[const1]
 movaps   xmm1,[esi]
 subps    xmm0,xmm1                                 ;   a  = (1-t)^3
 movaps   xmm7,xmm0                                 ;   b  = 3*(1-t)^2*t
 movaps   xmm6,xmm0                                 ;   c  = 3*(1-t)*t^2
 mulps    xmm6,xmm6                                 ;   d  = t^3
 mulps    xmm7,xmm6
 movaps   .a1,xmm7
 mulps    xmm6,[const3]
 mulps    xmm6,[esi]
 movaps   .b1,xmm6
 mulps    xmm0,[const3]
 mulps    xmm1,xmm1
 mulps    xmm0,xmm1
 movaps   .c1,xmm0
 mulps    xmm1,[esi]
 movaps   .d1,xmm1


  movss   xmm0,[ebx]  ;[Bp1x]
  shufps  xmm0,xmm0,0
  mulps   xmm0,.a1       ;   px = a*p1x + b*p2x + c*p3x + d*p4x
  movss   xmm1,[ebx+12]  ;[Bp2x]
  shufps  xmm1,xmm1,0
  mulps   xmm1,.b1
  movss   xmm2,[ebx+24]  ;[Bp3x]
  shufps  xmm2,xmm2,0
  mulps   xmm2,.c1
  movss   xmm3,[ebx+36]  ;[Bp4x]
  shufps  xmm3,xmm3,0
  mulps   xmm3,.d1
  addps   xmm0,xmm1
  addps   xmm2,xmm3
  addps   xmm0,xmm2
  movss   [edi],xmm0    ;  4 x coordinates
  shufps  xmm0,xmm0,00111001b
  movss   [edi+12],xmm0
  shufps  xmm0,xmm0,00111001b
  movss   [edi+24],xmm0
  shufps  xmm0,xmm0,00111001b
  movss   [edi+36],xmm0



  movss   xmm0,[ebx+4] ;[Bp1y]
  shufps  xmm0,xmm0,0
  mulps   xmm0,.a1       ;   px = a*p1x + b*p2x + c*p3x + d*p4x
  movss   xmm1,[ebx+16]  ;[Bp2y]
  shufps  xmm1,xmm1,0
  mulps   xmm1,.b1
  movss   xmm2,[ebx+28]  ;[Bp3y]
  shufps  xmm2,xmm2,0
  mulps   xmm2,.c1
  movss   xmm3,[ebx+40] ;[Bp4y]
  shufps  xmm3,xmm3,0
  mulps   xmm3,.d1
  addps   xmm0,xmm1
  addps   xmm2,xmm3
  addps   xmm0,xmm2
  movss   [edi+4],xmm0    ;  4 x coordinates
  shufps  xmm0,xmm0,00111001b
  movss   [edi+16],xmm0
  shufps  xmm0,xmm0,00111001b
  movss   [edi+28],xmm0
  shufps  xmm0,xmm0,00111001b
  movss   [edi+40],xmm0



  movss   xmm7,[ebx+8] ;[Bp1z]
  shufps  xmm7,xmm7,0
  mulps   xmm7,.a1
  movss   xmm6,[ebx+20]  ;[Bp2z]
  shufps  xmm6,xmm6,0
  mulps   xmm6,.b1
  movss   xmm5,[ebx+32] ;[Bp3z]
  shufps  xmm5,xmm5,0
  mulps   xmm5,.c1
  movss   xmm4,[ebx+44]
  shufps  xmm4,xmm4,0
  mulps   xmm4,.d1
  addps   xmm4,xmm5
  addps   xmm6,xmm7
  addps   xmm4,xmm6
  movss   [edi+8],xmm4    ;  4 x coordinates
  shufps  xmm4,xmm4,00111001b
  movss   [edi+20],xmm4
  shufps  xmm4,xmm4,00111001b
  movss   [edi+32],xmm4
  shufps  xmm4,xmm4,00111001b
  movss   [edi+44],xmm4

  add     esp,256
  pop     ebp

ret
draw_long_pipe_derv_lines:
; in esi - derives vertices, each derive 4 vertices,
;          each derive vertex 3xdword float
;    ecx - vertices (!not derives!) count
;    edi - screen buffer ptr
;    eax - screen width
; derives list must be scaled and translated into screen coords

   pushad
   emms
 ;  mov      [line_color],0x0000ff00
 ;  mov      [lbuffer_ptr],edi
 ;  mov      [screen_width_var],eax
   movd     mm2,eax
   movd     mm1,edi
   mov      eax,0x0000ff00
   movd     mm3,eax
 .aagain:
   push     ecx
   movlps   xmm0,[esi]
   movhps   xmm0,[esi+12]
   cvtps2dq xmm0,xmm0

   pxor      xmm3,xmm3    ; check if in sreeen area
   movd      xmm4,[xres_vard]
   punpcklwd xmm4,xmm3
   movlhps   xmm4,xmm4
   pcmpgtd   xmm4,xmm0
   pcmpgtd   xmm3,xmm0
   pxor      xmm3,xmm4
   pmovmskb  eax,xmm3
   and       eax,0x0000ffff
   cmp       eax,0x0000ffff
   jne       @f

   movd     eax,xmm0
   psrldq   xmm0,4
   movd     ebx,xmm0
   psrldq   xmm0,4
   movd     ecx,xmm0
   psrldq   xmm0,4
   movd     edx,xmm0
   pushad
   call     lineaa
   popad
  @@:
   pop      ecx
   add      esi,12
   sub      ecx,1
   jnz      .aagain
   popad
ret

if 0
translate_spline_derv:
; in esi - ptr to derives
;    edi - ptr to to translated derives
;    eax - x add
;    ebx - y add
;    ecx - derives count

     cvtsi2ss xmm0,eax
     cvtsi2ss xmm1,ebx
     shufps   xmm1,xmm1,11110011b
     por      xmm0,xmm1
     shl      ecx,2
   @@:
     movups  xmm1,[esi]
     addps   xmm1,xmm0
     movlps  [edi],xmm0
     movhlps xmm0,xmm0
     movss   [edi],xmm0
     add     esi,12
     add     edi,12
     sub     ecx,1
     jnz     @b
ret
end if

rotate_long_pipe_derv:
; rotate and translate Bspline/Bezier derives
; procedure uses some global variables

   push   ebp
   mov    ebp,esp
   sub    esp,255

   .sinx  equ dword[ebp-4]
   .cosx  equ dword[ebp-8]
   .siny  equ dword[ebp-12]
   .cosy  equ dword[ebp-16]
   .sinz  equ dword[ebp-20]
   .cosz  equ dword[ebp-24]


   fninit
   fld   [one_deg]
   fld   st
   fimul dword[main_rotary+8]   ; z
   fsincos
   fstp  .sinx
   fstp  .cosx
   fld   st
   fimul dword[main_rotary] ; x
   fsincos
   fstp  .siny
   fstp  .cosy
   fimul dword[main_rotary+4] ; y
   fsincos
   fstp  .sinz
   fstp  .cosz

   pushad
   push   .sinx
   push   .cosx
   push   .siny
   push   .cosy
   push   .sinz
   push   .cosz
   lea    edi,[matrix_scaled]
   call   make_rotation_matrix
   ; edi -  desired matrix
   popad


   lea  esi,[matrix_scaled] ;esi - pointer to 3x3 matrix
   mov  ebx,scale     ;ebx - ptr to scale
   call add_scale_to_matrix


   mov    esi,long_pipe_derv ; bssurface_derives
    ; rotate bezier surf. nodes
   mov    edi,long_pipe_derv_rotated
   mov    ecx,[long_pipe_segs_c]
   lea    ecx,[ecx*3+1]
   lea    ebx,[matrix_scaled]
   call   rotary


    movdqu   xmm3,[xxadd]
    cvtdq2ps xmm3,xmm3
    xor      ecx,ecx
    mov      eax,[long_pipe_segs_c]
    lea      eax,[eax*3+1]
    mov      ebx,long_pipe_derv_rotated ;bssurface_derives_rotated
  @@:
    movups   xmm1,[ebx]
    addps    xmm1,xmm3
    movlps   [ebx],xmm1
    movhlps  xmm1,xmm1
    movss    [ebx+8],xmm1
    add      ebx,12
    add      ecx,1
    cmp      ecx,eax
    jnz      @b

    mov      esp,ebp
    pop      ebp
ret
do_long_pipe:
; in:  esi  - ptr to curve vertices - middle points of pipe
;      ecx  - curve vertices count
;      eax  - rotary steps count as integer
;      xmm0 - radius, lowest dword float
;      ebx  - ptr to triangles list
;      edi  - ptr to pipe vertices
; out: ecx  - triangles count
;      ebx  - vertices count
; I am pretty sure that this pipe routines can be done
; more simply using cylinder equations. This can be
; shorter, faster and more nice. Feel free to make your
; own code instead mine - poor and humble.

  push ebp
  mov  ebp,esp
  sub  esp,256
  and  ebp,0xfffffff0

  .curve_vert_ptr equ [ebp-4]
  .curve_vert_c   equ [ebp-8]
  .pipe_vert_ptr  equ [ebp-12]
  .radius         equ [ebp-16]
  .rot_steps_c    equ [ebp-20]
  .one_deg        equ [ebp-24]
  .vert1          equ [ebp-40]
  .vert2          equ [ebp-56]
  .cross          equ [ebp-72]
  .sinus          equ [ebp-76]
  .cosinus        equ [ebp-80]
  .mx             equ [ebp-116]
  .counter        equ dword [ebp-120]
  .counter_edi    equ dword [ebp-124]
  .xedi           equ dword [ebp-128]
  .xesi           equ dword [ebp-132]
  .tri_list       equ dword [ebp-136]
  .radius_vec     equ [ebp-160]
  .point_to_rot   equ [ebp-176]
  .vert_cnt       equ dword [ebp-180]
  .axx            equ [ebp-184]



  mov   .curve_vert_ptr,esi
  mov   .pipe_vert_ptr,edi
  mov   .curve_vert_c,ecx
  mov   .tri_list,ebx
  mov   .rot_steps_c,eax
  movss .radius,xmm0
  mov   .xesi,esi
  mov   .xedi,edi

  movss    xmm0,[two_pi]
  cvtsi2ss xmm1,.rot_steps_c
  divss    xmm0,xmm1
  movss    .one_deg,xmm0
  fninit


  mov   .counter,0
  mov   .vert_cnt,0
 .loop2:
  lea   ebx,.vert1
  mov   esi,.xesi
  lea   edi,[esi+12]
  call  make_vector_r
  lea   edi,.vert1
  call  normalize_vector
  lea   ebx,.vert2
  mov   esi,.xesi
  lea   edi,[esi+12]
  add   esi,24
  call  make_vector_r
  lea   edi,.vert2
  call  normalize_vector
  lea   esi,.vert1
  lea   edi,.vert2
  lea   ebx,.radius_vec
  call  cross_product
  lea   edi,.radius_vec
  call  normalize_vector
  movss  xmm1,.radius
  shufps xmm1,xmm1,0
  mulps  xmm1,.radius_vec
;  mov    eax,0x7fffffff
;  mov    ebx,0xffffffff
;  movd   xmm2,eax
;  movd    xmm3,ebx
;  shufps  xmm2,xmm2,11001100b
;  shufps  xmm3,xmm3,0
;  movlhps xmm3,xmm2
;  andps   xmm1,xmm3

  movaps .point_to_rot,xmm1

  mov    .counter_edi,0  ; rotary counter
 .loop1:
  fld    dword .one_deg
  fimul  .counter_edi
  fsincos
  fstp  dword .sinus
  fstp  dword .cosinus
  lea   esi,.cosinus
  lea   edi,.mx
  lea   ebx,.vert1
  call  make_arbitrary_mx

  lea   esi,.point_to_rot   ; esi - point to rotate
  mov   edi,.xedi
  mov   ecx,1
  lea   ebx,.mx
  call  rotary  ;mul_matrix
  mov    esi,.xesi
  mov    edi,.xedi
  movups xmm0,[edi]  ; translate point by vector [esi]
  movups xmm1,[esi]
  addps  xmm0,xmm1
  movlps [edi],xmm0
  movhlps xmm0,xmm0
  movss  [edi+8],xmm0
  add   .vert_cnt,1
  add   .xedi,12
  add   .counter_edi,1
  mov   ecx,.counter_edi
  cmp   ecx,.rot_steps_c
  jnz   .loop1

  add   .xesi,12
  add   .counter,1
  mov   ecx,.counter
  cmp   ecx,.curve_vert_c
  jnz   .loop2

; do triangles list

  mov  .counter,0
  mov  esi,.tri_list
  mov  ebx,.rot_steps_c
  xor  eax,eax
  mov  ecx,.curve_vert_c
  sub  ecx,2

 .oop:
  push ecx
  mov  ecx,.rot_steps_c
  mov  .axx,eax  ; .axx - variable nedeed to last triangle
  sub  ecx,1    ;  of turn

 @@:
  mov  dword[esi],eax
  mov  edx,eax
  add  edx,ebx
  add  eax,1
  mov  dword[esi+4],eax
  mov  dword[esi+8],edx
  mov  dword[esi+16],edx
  add  edx,1
  mov  dword[esi+12],edx
  mov  dword[esi+20],eax
  add  .counter,2
  add  esi,24
  sub  ecx,1
  jnz  @b

  mov  dword[esi],eax
  mov  edx,eax
  add  edx,ebx
  add  eax,1
  mov  dword[esi+4],eax
  mov  dword[esi+8],edx
  mov  ecx,.axx      ; last  triangle - special cause
  mov  [esi+12],ecx
  add  ecx,ebx
  mov  [esi+16],ecx
  sub  ecx,1
  mov  [esi+20],ecx
  add  .counter,2
  add  esi,24
  pop  ecx
  sub  ecx,1
  jnz  .oop

  mov  ecx,.counter
  mov  ebx,.vert_cnt

; Vertices and triangles describing pipe done.
; ecx, ebx ->  triangles count and vertices count of pipe.


  add  esp,256
  pop  ebp
ret


; data segment
;Bsplines_vertices  rd 12 * derives all segemnt count
;                       * one segment steps count
;Bspline_derives 4 * 12 * all segments count
;Bspline_one_ss_count dw 10
;Bspline_seg__count dw 12 ; all segments