;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Ave Cruce Salus Mea
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; program: Win3ds23 - *.3ds / *.asc files viewer + demonstration of
;;          several graphics effects + tiny based on Bezier curves
;;          3d objects editor.
;; author:  Maciej Guba aka macgub
;; web   :  www.macgub.vxm.pl - if this site not works, search the web
;;          for keyword 'macgub'.
;; email :  macgub3@wp.pl
;; This application don't uses any particular 3d graphic library.
;; Written from scratch in pure assembler. Thanks to all, who was
;; helping me to do it. Especially:
;; Jan Pawel II, Tomasz Grysztar, Madis Kalme, Mikolaj Feliks, Lostcauz,
;; Brian Paul, Reverend, Pablo Reda, MHajduk, Ica, James Foley,
;; Andries van Dam, Steven Feiner, John Hughes, Richard Phillips,
;; tthsqe, J. Burkardt, Morgan McGuire and others...
;;
;; Program needs SSE3 extension.
;;
;; Coding style is terribly ugly. Other words it's my own mess. ;)
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Keys / GUI Buttons description:
;;          SPACE key / 'dr. model' button -  switching beetwen displaying
;;             models: flat, smooth shading, spherical environment,
;;             texture, bump, smooth + texture, two texture and bump
;;             + texture mapping pararell, edges only (smooth shaded,
;;             textured lines), flat with float single precision Z
;;             coordinate, point light with shadow light position
;;             dependent, real Phongs shading with float normal vector
;;             interpolation... Description of actual displaying model
;;             on bar located at top of app window;
;;
;;          TAB / 'ran. light' - generate three unlinear directional
;;             lights (Phongs illumination);
;;
;;          F1 / 'mot. blur' - motion blur on/off;
;;
;;          F2 / 'cur object' - cyclic switching between objects (from
;;             file and predefined: heart (Sanctissimi Cordis Christi)
;;             and tetrahedron).
;;          F3/F4(+/-) keys / 'zoom in/out' buttons - no comment;
;;
;;          F5 / 'tex. mapp.' button - choose   bumps also texture mapping
;;             planar or spherical (planar according to axis X, Y, Z);
;;
;;          F6 / 'bumps' - switching between random bumps and texture
;;             dependent;
;;
;;          F7 / 'texture' - select  texture: red white stripes, xored face
;;              also blank (witch can by used in smoothing objects edges)
;;              or texture from external file (X key);
;;
;;          F8 / 'save obj.' - save 3d currently shown object to file
;;             "object.asc" and "object.3ds";
;;
;;          F9 / 'cur.segm.' - disabled.
;;
;;          F11 / 'run/stop' button - main rotary on/off;
;;
;;          F12 / 'rotary' - main rotary cyclic switching - around axis
;;             X, Y, Z;
;;
;;          D  / 'do displ.' - do displacement mapping texture dependent;
;;
;;          T  / 'do triang.' - triangulize objects faces from teslation
;;             area; it's mean divide each face inside area to four
;;             smaller faces;
;;
;;
;;          U  / 'rand. shd.' button - random point (spot) light position
;;             and obliquity of plane on wihch shadow falls;
;;
;;          V / 'edit mesh.' button - edit option, posibility to edit
;;            according to vertex, single face or edge. Use left mouse
;;            button and yellow handlers;
;;
;;          W / 'set t. area' button - disabled.
;;
;;          C / 'set t. col' - set color that will by bypassed when
;;            teslation occured, options - black, white;
;;
;;          F / 'dsp. factor' - set diplacement factor, determine bumps
;;            that make displacement higher or lower;
;;
;;          B / 'bezier sr.' button - select predefined object, assembled
;;            using bezier patches;
;;
;;          A / 'set ap tol' button - set approximation tolerance of adding
;;            near bezier vertices, some of them are redundant (on tangent
;;            of bezier patches), this value is used also in removing
;;            redundend vertices option;
;;
;;          E / 'b. derives' button - draw bezier derives, selected by B
;;            key;
;;
;;            't&d wh|pos' button - displacement options if set on
;;            'posi' - only positive z cooficient vertices are affected,
;;            if set on 'whol' all vertices of displacement process are
;;            affected;
;;
;;          S / 'speed' button - toggle speed;
;;
;;            'rm unu ver' button - remove unused vertices, other words remove
;;             this vertices that are unused in triangles list;
;;
;;          P / 'Long pipe ' button - make long multi segment pipe, editing
;;             pipe - hold left mouse button on yellow bar - moving node,
;;             hold right mouse button and move them - rotate pipe project,
;;             pressing once again 'Long pipe ' menu button - render pipe;
;;
;;            'Lp segs cn' menu button - determin long pipe segments count;
;;
;;            'Lp Bez/Bsp' button - determine curve type long pipe based,
;;             options Bezier / Bspline;
;;
;;            'Lp quality' - determine long pipe quality (traiangles count
;;             per segment;
;;
;;            'NextM edit' menu button - edit next (earliest loaded joined)
;;             mesh and: blue bar determines X,Y position, violet bar
;;             determines Z position, yellow bar determies scale of new
;;             loaded freshly new part of mesh, move bars with mouse and
;;             its left button, in violet and yellow bars cause only vertiical
;;             move is allowed;
;;
;;            'Rm red ver' button - removing redundand vertices, according
;;             to aproximation tolerancy value;
;;
;;            'Draw norm.' button - drawing faces and vertices normal vectors;
;;
;;            'Norm. care' button - take care about normal vector, when
;;             is performing remove unused vertices option - 'rm unu ver'
;;             button, if 'Norm. care' button is on - merging vertices only
;;             with equal normal vector.
;;
;;          X key - load texture (raw format 512X512 pixels 24bit 0xRRGGBB)
;;
;;          N key - load new mesh (old will be abandoned);
;;
;;          M key - load new mesh and join it to the old;
;;
;;
;;          ESC key - exit;
;;
;;          Right mouse button and move - setting position of texture,
;;                (works in all drawing models, but is visable only in
;;                texture displaying models).
;;
;;          After run program shows dialog to open file.
;;          Choose file in 3ds or asc format and ...admire. ;)
;;
;;
;;
;;                                    "No work is ever complete,
;;                                     and this one is no exception."
;;                                                      Pierre Bezier
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

format PE GUI 4.0
entry start

include '/include/win32a.inc'

BSEGMENTS = 2       ; Bezier curve segments count
BPIPES    = 1          ; Bezier pipes count (only in solid and pipes
                    ; pararell editing cause).
BROTSOLIDS =  BSEGMENTS - BPIPES
BXRES      equ 400 ;1600 ;1200 ;800 ;choose window size of bezier curve editor area
BYRES      equ 400 ; 900 ; 675 ;450
BSIZE_X    equ BXRES
BSIZE_Y    equ BYRES
MXRES      equ BXRES  ; menu window size
MYRES      equ 250 ;BYRES
BROUND     equ 8

BROTARY_STEPS = BSTEPS   ; BROTARY_STEPS and BSTEPS  must have the same
          ; value (both determines object quality| - pipes and rot. solids)
BSTEPS = 12

TEX_X  = 512    ; texture size
TEX_Y  = 512
TEXTURE_SIZE    EQU (TEX_X * TEX_Y)-1
TEX_SHIFT       = 9
TRIANGLES_COUNT equ 4
POINTS_COUNT    equ 4
MAX_COUNTER     equ 16    ; b_surface.inc  - determines triangles count
ROUND           equ 8
NON    = 0  ; --\
MMX    = 1  ;    > this definitions are only for upside down
SSE    = 2  ;    > compatible reasons
SSE2   = 3  ; --/
SSE3   = 4
SSE4.1 = 5
AVX    = 6
AVX2   = 7
Ext    =  SSE3   ; as minimal
;HEAP_MAX_SIZE = 1024 * 1024 * 20 ;1024  ; = 1GB address space.
WM_FINISH equ WM_USER+100h


section '.text' code readable executable
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; start and message loop
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
start:
    invoke GetModuleHandle,0
    mov [wc.hInstance],eax
    invoke LoadIcon,0,IDI_APPLICATION
    mov [wc.hIcon],eax
    invoke LoadCursor,0,IDC_ARROW
    mov [wc.hCursor],eax

     invoke  GetSystemMetrics,SM_CXSCREEN
 ;    mov     eax,600
     and     eax,0xfffffffe
     mov     [whole_xres],ax
     sub     eax,MXRES    ; place for menu and bezier curves area
     and     eax,-4       ; motion blur need that
     mov     ebx,eax
     mov     [xres_var],ax
     shr     eax,1
     mov     [xxadd],eax
     invoke  GetSystemMetrics,SM_CYSCREEN
 ;    mov     eax,600
     and     eax,0xfffffffe
     mov     [whole_yres],ax
     sub     eax,100
     mov     [yres_var],ax
     mov     ecx,eax
     shr     ecx,1
     mov     [yyadd],ecx

     lea     ecx,[ebx-3]
     cvtsi2ss xmm0,ecx
     movss   [max_clip_seg],xmm0
     lea     ecx,[eax-3]
     cvtsi2ss xmm2,ecx
     movss   [max_clip_seg+4],xmm2

     movzx   eax,[xres_var]
     movzx   ebx,[yres_var]
     sub     eax,6
     sub     ebx,6
     mov     [max_work_x],ax
     mov     [max_work_y],bx
     sub     eax,1
     sub     ebx,1
     cvtsi2ss xmm0,eax
     cvtsi2ss xmm1,ebx
     movss   [tri_area_x2],xmm0
     movss   [tri_area_y2],xmm1

     movzx   eax,[xres_var]
     movzx   ebx,[yres_var]
     mul     ebx
     push    eax

     invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     mov     [slices_counter_ptr],eax
     pop     ebx
     shl     ebx,2       ;dbg
     invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     or  eax,eax
     jnz @f
   @@:

    mov     [screen_ptr],eax
    invoke   VirtualAlloc,NULL,ebx,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
    mov      [slices_ptrs_buff_ptr],eax


    shl     ebx,2  ; in Z buffer , also 4 slices buffer
    invoke  VirtualAlloc,NULL,ebx,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
    mov     [Zbuffer_ptr],eax
;    shr     ebx,1
    lea     eax,[65536+ebx]

    invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
    mov     [edit_buf_ptr],eax
 ;   shr     ebx,2
    invoke  VirtualAlloc,NULL,ebx,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
    mov     [shd_buffer_ptr],eax


    mov    eax,4000000  ; mem for long pipe size determines with Lp variables
    invoke VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
    mov   [long_pipe_vert_ptr],eax  ; mem for long pipe vertices


    mov   [bmi.biSize],sizeof.BITMAPINFOHEADER
    movzx eax,[xres_var]
    mov   [bmi.biWidth],eax  ;XRES
    movzx edx,[yres_var]

 ;   add   edx,30
    neg   edx

    mov [bmi.biHeight],edx ;-YRES-30
    mov [bmi.biPlanes],1
    mov [bmi.biBitCount],32
    mov [bmi.biCompression],BI_RGB

    mov [bmi.biSizeImage],0
    mov [bmi.biXPelsPerMeter],0
    mov [bmi.biYPelsPerMeter],0
    mov [bmi.biClrUsed],0
    mov [bmi.biClrImportant],0


    mov [Bbmi.biSize],sizeof.BITMAPINFOHEADER
    mov [Bbmi.biWidth],BXRES
    mov [Bbmi.biHeight],-BYRES-30
    mov [Bbmi.biPlanes],1
    mov [Bbmi.biBitCount],32
    mov [Bbmi.biCompression],BI_RGB

    mov [Bbmi.biSizeImage],0
    mov [Bbmi.biXPelsPerMeter],0
    mov [Bbmi.biYPelsPerMeter],0
    mov [Bbmi.biClrUsed],0
    mov [Bbmi.biClrImportant],0

    mov [Mbmi.biSize],sizeof.BITMAPINFOHEADER
    mov [Mbmi.biWidth],MXRES
    mov [Mbmi.biHeight],-MYRES ;-30 ;-BYRES-30
    mov [Mbmi.biPlanes],1
    mov [Mbmi.biBitCount],32
    mov [Mbmi.biCompression],BI_RGB

    mov [Mbmi.biSizeImage],0
    mov [Mbmi.biXPelsPerMeter],0
    mov [Mbmi.biYPelsPerMeter],0
    mov [Mbmi.biClrUsed],0
    mov [Mbmi.biClrImportant],0

      mov  [bmiheader.biSize],sizeof.BITMAPINFOHEADER
                        movzx  eax,[xres_var]
                        mov  [bmiheader.biWidth],eax
                        movzx  eax,[yres_var]
                        neg  eax
                        mov  [bmiheader.biHeight],eax
                        mov  [bmiheader.biPlanes],1
                        mov  [bmiheader.biBitCount],32
                        mov  [bmiheader.biCompression],BI_RGB
                        movzx  eax,[xres_var]
                        movzx  ebx,[yres_var]
                        imul  eax,ebx
                        lea  eax,[4*eax+16]
                        mov  [bmiheader.biSizeImage],eax
                        mov  [bmiheader.biXPelsPerMeter],0
                        mov  [bmiheader.biYPelsPerMeter],0
                        mov  [bmiheader.biClrUsed],0
                        mov  [bmiheader.biClrImportant],0
                     invoke  CreateDIBSection,0,bmiheader,0,screen_ptr,0,0


    mov   [rect.left],0
    mov   [rect.top],0
    movzx eax,[whole_xres]
    mov   [rect.right],eax
    movzx edx,[whole_yres]
    add   edx,30
;    neg edx
    mov   [rect.bottom],edx ;YRES+30

    invoke AdjustWindowRect,rect,WS_POPUP or WS_SYSMENU or WS_CAPTION,0

    invoke RegisterClass,wc



    mov ecx,[rect.bottom]
    mov eax,[rect.right]
    sub ecx,[rect.top]
    sub eax,[rect.left]
    ; rasizable window
    invoke  CreateWindowEx, 0,_class, _title, WS_VISIBLE + WS_OVERLAPPEDWINDOW,\
                                0, 0, eax, ecx, NULL, NULL, [wc.hInstance], NULL

;    invoke  CreateWindowEx, 0,_class, _title, WS_VISIBLE + WS_OVERLAPPEDWINDOW,\
;                                CW_USEDEFAULT, CW_USEDEFAULT, eax, ecx, NULL, NULL, [wc.hInstance], NULL
    ; non resizable window
 ;   invoke CreateWindowEx,0,_class,_title,WS_OVERLAPPEDWINDOW\
 ;          xor WS_THICKFRAME xor WS_MAXIMIZEBOX,0,0,eax,ecx,NULL,NULL,0,0
    mov [hwnd],eax
    invoke GetDC,[hwnd]
    mov [hdc],eax
    invoke ShowWindow,[hwnd],SW_SHOW

    ; fix bezier rect files
     mov   ebx,tspoon_rect
     mov   ecx,16+26+32 ; spoon + cup + pot
     shl   ecx,4
   @@:
     sub  word[ebx],1
     add  ebx,2
     sub  ecx,1
     jnz  @b

    call long_pipe_init ; init control points of long pipe
                        ; dependent on segments count
    pushad
    movzx  eax,[yres_var]  ;YRES / 2
    shr    eax,1
    cvtsi2ss xmm0,eax
    movss  [scale],xmm0

    mov    edi,segments_rotary   ; zero segment add struct
    xor    eax,eax
    mov    ecx,BSEGMENTS*3
    cld
    rep    stosd


    mov    edi,segments_add     ; zero segments add struct
    xor    eax,eax
    mov    ecx,BSEGMENTS*3
    rep    stosd


    mov    edi,menu_screen               ; zero menu screen
    mov    ecx,MXRES * MYRES
    rep    stosd



    call   normalize_plane_equation

    call   Binit_derives


    call   open_file
    cmp    [file_ext_flag],0
    jne    @f
; in:   eax -> ptr to 3ds file
;       ebx -> file size
; out:  ecx -> points count
;       ebx -> triangles count
    mov    ebx,[FileSize]
    mov    eax,[File_ptr]
    call   read_tp_variables1
    mov    [points_count_var],ecx
    mov    [triangles_count_var],ebx
    xor    eax,eax
    call   alloc_mem_for_tp1
; read from 3ds file
;in:  eax - ptr to triangles
;     ebx - ptr to vertices
;     ecx - ptr to file begin
;     edx - ptr to file end
;out: ecx - points (vertices) count
;     edx - triangles count

    mov    eax,[triangles_ptr]
    mov    ebx,[points_r_ptr]
    mov    ecx,[File_ptr]
    call   read_from_file     ; 3ds
    mov    [EndFile],ebx


    jmp    .norm
  @@:
    call   read_asc_tp_variables
    xor    eax,eax
    call   alloc_mem_for_tp1
    call   read_asc             ; asc
  .norm:

    call   prepare_object
    call   normalize_all_light_vectors
    mov    edi,lights_aligned
    mov    ecx,3
    call   norm_aligned_lv

    call   init_envmap2
    call   do_color_buffer
    mov    edi,bump_map
    call   calc_bumpmap
    call   generate_texture2


    call   draw_menu

    popad








;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;   msgLoop
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
msg_loop:


   cmp    [lpipe_flag],1 ;   long pipe cause
   jne    .no_pipe_edit

          call    rotate_long_pipe_derv   ; rotate derives, scale
                                        ; and translate on vector

          cmp        [edit_vertex_no],0
          je         @f

          movd       xmm0,dword[edit_end_x]
          movd       xmm1,dword[edit_start_x]
          pxor       xmm2,xmm2
          punpcklwd  xmm0,xmm2
          punpcklwd  xmm1,xmm2
          psubd      xmm0,xmm1

          cvtdq2ps   xmm0,xmm0

          mov        eax,[edit_vertex_no]
          sub        eax,1
          shl        eax,2
          lea        eax,[eax*3]
          add        eax,long_pipe_derv_rotated
          movlps     xmm1,[eax]

          addps      xmm1,xmm0
          movlps     [eax],xmm1

        @@:
          mov    edx,[long_pipe_vert_ptr]
          mov    eax,[long_pipe_one_seg_c]
          mov    ebx,[long_pipe_segs_c]   ;segments count
          cmp    [lpcurve_tp],0
          je     @f
          lea    ebx,[ebx*3-2]
        @@:
          mov    ecx,long_pipe_derv_rotated

          mov    eax,[long_pipe_one_seg_c]
          pushad
          movzx  esi,[lpcurve_tp]
          call   calc_long_pipe ;Bsplines or bezier curves
          popad
          imul   eax,ebx
          mov    [points_count_var],eax
          mov    [triangles_count_var],3

          xor    eax,eax          ; clear draw buffer
          mov    edi,[screen_ptr]
          movzx  ebx,[xres_var]
          movzx  ecx,[yres_var]
          imul   ecx,ebx
          call   fill_aligned

             ;  draw_vertices
          mov      esi,[long_pipe_vert_ptr]  ;[points_rotated_ptr]
          mov      ecx,[points_count_var]
          movzx    edx,[xres_var]

          pxor      xmm3,xmm3
          movd      xmm4,[xres_vard]
          punpcklwd xmm4,xmm3
        .again_vertex:
          movups    xmm0,[esi]
          cvtps2dq  xmm0,xmm0
          pxor      xmm3,xmm3
          movd      xmm4,[xres_vard]
          punpcklwd xmm4,xmm3
          pcmpgtd  xmm4,xmm0
          pcmpgtd  xmm3,xmm0
          pxor     xmm3,xmm4
          pmovmskb eax,xmm3
          and      eax,0x000000ff
          cmp      eax,0xff
          jne      @f

          movd     eax,xmm0
          psrldq   xmm0,4
          movd     ebx,xmm0
          imul     ebx,edx
          add      ebx,eax
          shl      ebx,2
          add      ebx,[screen_ptr]
          mov      [ebx],dword 0x00ffffff
        @@:                         ;  skip vertex
          add      esi,12
          sub      ecx,1
          jnz      .again_vertex

   ;      mov   [triangles_count_var],32*32*32  ; will be enought
   ;      mov   [points_count_var],32*32*32
   ;      call  alloc_mem_for_tp1

          ; in esi - derives vertices, each derive 4 vertices,
          ;         each derive vertex 3xdword float
          ;    ecx - derives count
          ;    edi - screen buffer ptr
          ;    eax - screen width
          ; derives list must be scaled and translated into screen coords
          mov   esi,long_pipe_derv_rotated
          mov   ecx,[long_pipe_segs_c]
          lea   ecx,[ecx*3]

          mov   edi,[screen_ptr]
          movzx eax,[xres_var]
          call  draw_long_pipe_derv_lines
          call  clear_edit_buf

          movzx    eax,[xres_var]
          movzx    ebx,[yres_var]
          mov      ecx,[long_pipe_segs_c]
          lea      ecx,[ecx*3+1]
          mov      edx,[screen_ptr]
          mov      esi,long_pipe_derv_rotated
          mov      edi,[edit_buf_ptr]
          call     write_edit_bars_vert  ; write handlers to derives

          jmp      .no_draw

 .no_pipe_edit:

    stdcall g_draw

 .no_draw:

    cmp   [edit_flag],0
    je    .skip_ed
 ;   jne   @f
 ;   cmp   [lpipe_flag],1
 ;   je    @f
 ;   jmp   .skip_ed
 ; @@:
    call  clear_edit_buf
    call  write_edit_bars
  .skip_ed:
    inc   [temp_fur]  ; delete me
    movzx eax,[ccounter_inc]
    ; if custom rotate
    cmp   [lpipe_flag],1
    je    .custom
    cmp   [edit_flag],0
    je    @f
    xor   eax,eax
  @@:
    cmp   [rotary_flag],0
    jne   @f
    add   dword[main_rotary],eax
    and   dword[main_rotary],0xff
    jmp   .end_rot_inc
  @@:
    cmp   [rotary_flag],1
    jne   @f
    add   dword[main_rotary+4],eax
    and   dword[main_rotary+4],0xff
    jmp   .end_rot_inc
  @@:
    cmp   [rotary_flag],2
    jne   @f
    add   dword[main_rotary+8],eax
    and   dword[main_rotary+8],0xff
    jmp   .end_rot_inc
  @@:
    cmp   [rotary_flag],3
    jne   .end_rot_inc
  .custom:       ; custom rotate
    push  [custom_obj_angle]
    pop   eax
    and   eax,0x00ff00ff
    mov   ebx,eax
    shr   ebx,16
    mov   dword[main_rotary],ebx
    and   dword[main_rotary],0xff
    mov   dword[main_rotary+4],eax
    and   dword[main_rotary+4],0xff


  @@:
  .end_rot_inc:

; draw image

    movzx  eax,[xres_var]
    movzx  ebx,[yres_var]
;    rdtsc
;    push   eax
;    invoke StretchDIBits,[hdc],0,30,eax,ebx,0,0,eax,ebx,[screenDIB_ptr],bmi,0,SRCCOPY
;    pop    ebx
;    rdtsc
;    sub    eax,ebx
;    push   eax
;    rdtsc
;    push   eax
    invoke StretchDIBits,[hdc],0,30,eax,ebx,0,0,eax,ebx,[screen_ptr],bmi,0,SRCCOPY
;    pop    ebx
;    rdtsc
;    sub    eax,ebx   ;  eax - dib,   ebx - vitual
;    pop    ebx
;int3
  ;  invoke BitBlt,[hdc],0,30,eax,ebx,[screen_ptr],bmi,0,SRCCOPY
 ;    call draw_menu
    movzx    eax,[whole_xres]
    mov      ecx,eax
    sub      eax,MXRES

    invoke StretchDIBits,[hdc],eax,0,MXRES,MYRES,0,0,MXRES,MYRES,menu_screen,Mbmi,0,SRCCOPY
    cmp    [tex_button_pres],1
    jne    @f

  @@:

    cmp  [hrt_flag],3            ; hrt_flag indicates what object is current
    jl   @f

    call   Bdraw_all
    movzx  eax,[whole_xres]
    movzx  ebx,[whole_yres]
    sub    ebx,(100+BYRES) ; 30
    sub    eax,BXRES


    invoke StretchDIBits,[hdc],eax,ebx,BXRES,BYRES,0,0,BXRES,BYRES,Bbuffer,Bbmi,0,SRCCOPY




  @@:
; .no_editor:
    movzx  eax,[draw_flag]
    mov    ebx,36  ; 36 - text chain lenght
    mul    ebx     ; draw render mode
    add    eax,sh_model
    invoke TextOut,[hdc],10,10,eax,36
    cmp    [edit_flag],0
    jne    @f
    cmp    [speed_flag],1
    je     @f
    invoke Sleep,3
  @@:
    invoke PeekMessage,msg,0,0,0,PM_NOREMOVE
    or eax,eax
    je msg_loop

    invoke GetMessage,msg,NULL,0,0
    or eax,eax
    je end_loop

    invoke TranslateMessage,msg
    invoke DispatchMessage,msg
    jmp msg_loop

.error:
;    call error
    invoke MessageBox,NULL,_error,NULL,MB_ICONERROR+MB_OK

end_loop:

    movzx    ebx,[xres_var]
    movzx    eax,[yres_var]
    mul      ebx
    shl      eax,2
    invoke  VirtualFree,[screen_ptr], eax, MEM_DECOMMIT
    xor     eax,eax
    call    free_mem_for_tp1

    invoke ReleaseDC,[hwnd],[hdc]
    invoke DestroyWindow,[hwnd]
    invoke ExitProcess,[msg.wParam]



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; WindowProc
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
proc WindowProc uses ebx esi edi, hwnd,wmsg,wparam,lparam
     cmp [wmsg],WM_KEYDOWN
     je .wmkeydown


   @@:
     cmp     [wmsg],WM_LBUTTONDOWN
     jne     .next_mess345

     cmp     [derive_flag],0
     jne     .no_edit_derives
     cmp     [ed_bpatch_derive_on],1
     jne     .no_edit_derives
     push    [lparam]
     pop     dword[ed_bpatch_derives_x_start]
     mov     word[ed_bpatch_derive_on],1
     jmp     .next_mess345
  .no_edit_derives:
     cmp     [NextMed_flag],0   ; next mesh edition routines
     jne     @f
     mov     eax,[lparam]
     mov     ecx,eax
     sub     ax,[whole_xres]
     shr     ecx,16              ;
     add     ax,BXRES
     sub     cx,[whole_yres]     ;
     shl     eax,16              ;
     add     cx,BYRES+100              ;
     mov     ax,cx
     ror     eax,16

     pxor    xmm1,xmm1
     movd    xmm0,eax
     mov     ebx,BYRES shl 16 + BXRES
     movd    xmm2,ebx
     pcmpgtw xmm1,xmm0
     pcmpgtw xmm2,xmm0
     pxor    xmm1,xmm2
     movd    ebx,xmm1
     mov     ecx,ebx
     shr     ecx,16
     and     cx,bx
     cmp     cx,-1
     jne     @f
     mov     ebx,eax
     mov     edx,BYRES
     movzx   eax,ax
     shr     ebx,16
     imul    ebx,edx
     add     eax,ebx
     shl     eax,2
     add     eax,Bbuffer
     mov     ebx,[eax]
     shr     ebx,24
     cmp     ebx,1   ; 1 - xy, 2 - z, 3 - scale
     je      @f
     cmp     ebx,2
     je      @f
     cmp     ebx,3
     je      @f
     jmp     .pipe
    @@:

     mov     [NextMotion],bx

   .pipe:
     cmp   [lpipe_flag],1
     je    .edit
     cmp   [rotary_flag],3
     jne   @f
     mov   eax,[lparam]
     mov   dword[custom_rot_start_x],eax
    @@:
       cmp   [edit_flag],0
       jne   .edit
       cmp   [set_tri_area],0
    ;   jne   .no_edit
    ; jmp   .no_edit
     .edit:
        mov     ebx,[lparam]
        emms
        movd    mm1,ebx
        mov     ebx,30 shl 16 + 1
        movd    mm0,ebx
        movd    mm3,dword[xres_var]
        paddw   mm3,mm0
        pcmpgtw mm0,mm1
        pcmpgtw mm3,mm1
        pxor    mm3,mm0
        movd    eax,mm3
        mov     cx,ax
        shr     eax,16
        and     ax,cx
        or      ax,ax
        jz      .no_edit
        movd    mm0,ebx
        psubw   mm1,mm0

        movd    eax,mm1
        push    eax
       ; eax - position of cursor in "screen/work" area
        cmp    [set_tri_area],0
        jne    .edit_mesh
        movzx  ebx,ax   ; check if edition triangulize area
        mov    ecx,eax
        shr    ecx,16
        movzx  edx,word[xres_var]
        imul   edx,ecx
        add    ebx,edx
        mov    ecx,ebx
        lea    ecx,[ebx*2]
        add    ecx,ecx
        add    ecx,[edit_buf_ptr]
        mov    ecx,dword[ecx]
        cmp    ecx,65532   ; change it coz 32 bit !!
        jne    @f
        mov    [edit_tri_area],1
      @@:
        cmp    ecx,65533
        jne    @f
        mov    [edit_tri_area],2
        jmp    .next_msg
    @@:
      .edit_mesh:
        pop     eax
        xor     edx,edx
        cmp     [lpipe_flag],1
        je      @f
        mov     esi,[points_r_ptr]
        mov     edi,[points_rotated_ptr]
        call    edit_init_and_submit
        jmp     .no_edit
      @@:
        mov     esi,long_pipe_derv
        mov     edi,long_pipe_derv_rotated
        call    edit_init_and_submit

   .no_edit:
     mov   eax,[lparam]
     sub   ax,[whole_xres]
     cmp   ax,-MXRES
     jl    .bez
     ror   eax,16
     cmp   ax,MYRES  ; menu work
     jg    .bez
     ror   eax,16
     mov   esi,menu_data
     mov   eax,[lparam]
     sub   ax,[whole_xres]
     add   ax,MXRES
  @@:
     push  eax
     mov   ecx,[esi+2]
     mov   edx,[esi+6]
     call  xy_in_rect
     pop   eax
     or    ebx,ebx
     jnz   @f
     add   esi,33               ; size of one position in menu
     cmp   [esi],word -1        ; end mark
     jnz   @b
     jmp  .bez
  @@:
     mov   ax,[esi+2]
     add   ax,2
     shl   eax,16
     mov   ax,[esi+6]
     sub   ax,2
     mov   bx,[esi+4]
     add   bx,2
     shl   ebx,16
     mov   bx,[esi+8]
     sub   bx,2
     emms
     push  edx
     mov   edx,MXRES
     movd  mm2,edx
     mov   edx,menu_screen
     movd  mm1,edx
     mov   edx,0x0000ff00
     movd  mm3,edx
     pop   edx





     pushad
     call  write_frame
     popad
     mov   word[esi],1      ;mark button chosen

     movzx    eax,[whole_xres]
     sub      eax,MXRES

     invoke StretchDIBits,[hdc],eax,0,MXRES,MYRES,0,0,MXRES,MYRES,menu_screen,Mbmi,0,SRCCOPY

   .bez:
     cmp   [hrt_flag],3  ; pipes, rot solid cause
     jl    .next_mess345
     mov   eax,[lparam]
     sub   ax,[whole_xres]  ;XRES-BXRES
     add   ax,BXRES
     ror   eax,16
     sub   ax,[whole_yres]  ;YRES+30-BYRES
     add   ax,BYRES+100
     ror   eax,16
     movd  xmm0,eax
     pxor  xmm1,xmm1
     mov   ebx,BYRES shl 16 + BXRES
     movd  xmm2,ebx
     pcmpgtw xmm1,xmm0
     pcmpgtw xmm2,xmm0
     pxor    xmm1,xmm2
     movd    ebx,xmm1
     mov     ecx,ebx
     shr     ecx,16
     and     cx,bx
     cmp     cx,-1
     jne     @f




     movzx ebx,ax
     shr   eax,16
     mov   edx,BXRES
     mul   edx
     add   eax,ebx
     shl   eax,2
     add   eax,Bbuffer
     mov   ebx,[eax]
     and   ebx,0xff000000
     cmp   ebx,0xff000000
     je    .next_mess345
     shr   ebx,24
     mov   [Bcrucial_point],ebx
     mov   [Bmove_flag],1
     jmp   .next_mess345
  @@:
     cmp   [current_seg],255
     je    @f


     mov   [seg_rot_flag_button_down],1
  @@:








   .next_mess345:
if 1 ;dd couunt
     cmp     [wmsg],WM_MOUSEMOVE
     jne     .next_msg
     cmp     [derive_flag],0    ; edit bezier patch
     jne     @f
     push    [lparam]
     pop     dword[ed_bpatch_derives_x_end]
    @@:
if 1
     cmp     [NextMed_flag],0   ; next mesh edition routines
     jne     .l_pipe
     cmp     [NextMotion],1
     je      @f
     cmp     [NextMotion],2
     je      @f
     cmp     [NextMotion],3
     jne     .l_pipe
  @@:
     mov   eax,[lparam]
     sub   ax,[whole_xres]  ;XRES-BXRES
     add   ax,BXRES
     ror   eax,16
     sub   ax,[whole_yres]  ;YRES+30-BYRES
     add   ax,BYRES+100
     ror   eax,16
     movd  xmm0,eax

     cmp     [NextMotion],1
     jne     @f
     mov     [NextMxadd],eax
     jmp     .callM
   @@:
     cmp     [NextMotion],2
     jne     @f
     shr     eax,16
     mov     [NextMzadd],ax
     jmp     .callM
   @@:
     cmp     [NextMotion],3
     jne     .l_pipe
     shr     eax,16
     mov     [NextScale],ax

   .callM:
     call      NextMdraw_all
     movzx     eax,[whole_xres]
     movzx     ebx,[whole_yres]
     sub       ebx,(BYRES+100) ; 30
     sub       eax,BXRES

     invoke  StretchDIBits,[hdc],eax,ebx,BXRES,BYRES,0,0,BXRES,BYRES,Bbuffer,Bbmi,0,SRCCOPY
     jmp  .end
end if
   .l_pipe:

     cmp   [lpipe_flag],1
     je    @f
     cmp   [rotary_flag],3
     jne   .tri_rect
   @@:
     cmp   [custom_rot_start_x],0
     je    .tri_rect
     emms
     movd  mm0,[lparam]
     psubw mm0,qword[custom_rot_start_x]
     psrlw mm0,2
     movd  [custom_obj_angle],mm0
  .tri_rect:
     cmp   [edit_tri_area],0
     je    @f
        mov     ebx,[lparam]
        emms
        movd    mm1,ebx
        mov     ebx,30 shl 16 + 1
        movd    mm0,ebx
        psubw   mm1,mm0
        pmaxsw  mm1,qword[min_work_x]
        pminsw  mm1,qword[max_work_x]
        movd    eax,mm1
        call    set_triangulize_rect
     @@:
    cmp    [tex_button_pres],1
    jne    @f

    movd   xmm0,[lparam]
    movd   xmm1,[tex_mov_begin_x]
    psubw  xmm1,xmm0
    movd   eax,xmm1
    movd   [tex_mov_begin_x],xmm0
    mov    ebx,eax
    sar    ebx,16
    cwde
    call   move_texture
  @@:
     cmp   [edit_vertex_no],0
     jne   @f
     cmp   [edit_edge_no],0
     jne   @f
     mov   eax,dword[edit_face_no]
     ; only one index, rest to done !
     ; what if face struct contains index 0 vertex ?
     ; it should be fixed for example xword -1 !
     or    eax,eax
     jz   .next_one
   @@:









     emms
     movd  mm0,[lparam]
     mov   eax,30 shl 16 + 0
     movd  mm1,eax
     psubw mm0,mm1
     pmaxsw  mm0,qword[min_work_x]
     pminsw  mm0,qword[max_work_x]

     movd  dword[edit_end_x],mm0
   ;  call edit_move
   .next_one:
     cmp   [seg_rot_flag_button_down],1
     jne   .Bezier_mov
     cmp   [hrt_flag],3
     jge   @f
     jmp   .Bezier_mov
   @@:
     cmp   [hrt_flag],6
     jge   .Bezier_mov
     mov   eax,[lparam]    ; rotate single segment of bezier
                                            ; rotated solid
     ror   eax,16
     sub   ax,100
     movzx ebx,ax
     shr   eax,16
     movzx ecx,[current_seg]
     shl   ecx,2
     lea   ecx,[ecx*3]
     and   eax,0xff
     and   ebx,0xff
     mov   [segments_rotary+ecx],eax
     mov   [segments_rotary+ecx+4],ebx
     pushad
     call  Bdraw_all
     popad
     cmp   [hrt_flag],3
     jne   @f
     call  save_object
     jmp   .Bezier_mov
   @@:
     cmp   [hrt_flag],4
     jne   @f
     call  save_pipe
     jmp   .Bezier_mov
   @@:
     call  save_pipe_rot_solid
   .Bezier_mov:
     cmp   [Bmove_flag],1
     jne   .next_msg
     cmp   [hrt_flag],6
     jge   .next_msg
     mov   ecx,[lparam]
     mov   ax,[whole_xres]
     sub   ax,BXRES
     cmp   cx,ax  ;XRES-BXRES
     jl    .next_msg
     ror   ecx,16
     mov   bx,[whole_yres]
     sub   bx,BYRES
     sub   bx,100
     cmp   cx,bx  ;YRES+30-BYRES
     jl    .next_msg
     sub   cx,bx   ;YRES+30-BYRES
     ror   ecx,16
     sub   cx,ax   ;XRES-BXRES

     mov   ebx,[Bcrucial_point]
     cmp   ebx,253                 ; check if axis
     jae   @f
     shl   ebx,2
     mov   [ebx+Biderives],ecx
     jmp   .draw
    @@:
     sub   ebx,253
     shl   ebx,2
     mov   [ebx+Bix1_axis],ecx
   .draw:
     pushad
     call  Bdraw_all
     popad
     cmp   [hrt_flag],4
     jne   @f
     call  save_pipe        ; save changes - on the fly write to points / triangles lists
     jmp   .next_msg
   @@:


     cmp   [hrt_flag],5
     jne   @f

     call  save_pipe_rot_solid
     jmp  .next_msg
   @@:
     call  save_object   ; if hrt_flag = 3 save solid rotared curve

 ; .no_rot_solids_and_pipes:     ; no job with rotated solids and pipes
end if ; dd count
if 1 ; dd count
  .next_msg:
     cmp   [wmsg],WM_LBUTTONUP
     jne   .rbutd
     mov   [ed_bpatch_derive_on],0
     mov   [NextMotion],0   ; next mesh x y
     cmp   [lpipe_flag],1
     je    @f
     mov   [custom_rot_start_x],0
   @@:
     mov    [edit_tri_area],0
     cmp    [edit_vertex_no],0
     jne    @f
     mov    eax,dword[edit_face_no]
     or     eax,eax   ; only one index, rest to done !
     jne    @f
     cmp    [edit_edge_no],0
     jne    @f
     jmp    .segment_rot
  @@:
     emms
     movd  mm0,[lparam]
     mov   eax,30 shl 16 + 0
     movd  mm1,eax
     psubw mm0,mm1
     movd  eax,mm0

     mov    edx,1
     mov    esi,[points_r_ptr]
     mov    edi,[points_rotated_ptr]
     cmp    [lpipe_flag],1
     jne    @f
     mov    esi,long_pipe_derv
     mov    edi,long_pipe_derv_rotated
   @@:
     call   edit_init_and_submit
     call   calc_bumpmap_coords  ; bump , tex
     mov    [edit_vertex_no],0
     mov    [edit_edge_no],0
;     pxor   xmm0,xmm0
     mov    dword[edit_face_no],0  ;xmm0
     mov    dword[edit_end_x],0
     mov    dword[edit_start_x],0




   .segment_rot:

     mov   [seg_rot_flag_button_down],0
     mov   [Bmove_flag],0

     ; GUI buttons work
     mov   eax,[lparam]     ; check if sth in menu marked
     sub   ax,[whole_xres]
     cmp   ax,-MXRES
     jl    .dr_mn
     ror   eax,16
     cmp   ax,MYRES
     jg    .dr_mn
     ror   eax,16
     mov   esi,menu_data
     mov   eax,[lparam]
     sub   ax,[whole_xres]
     add   ax,MXRES
  @@:
     push  eax
     mov   ecx,[esi+2]
     mov   edx,[esi+6]
     call  xy_in_rect
     pop   eax
     or    ebx,ebx
     jnz   @f
     add   esi,33               ; size of one position in menu
     cmp   [esi],word -1
     jnz   @b
     jmp  .dr_mn
  @@:
     cmp   word[esi],1
     jne   .dr_mn
     ; this button was pressed (pointed in esi)
                            ; update flags
     mov   word[esi],0      ; mark
     mov   al,byte[esi+22]
     inc   al
     cmp   al,[esi+21]
     jne   @f
     xor   al,al
   @@:
     mov   [esi+22],al
     mov   ax,[esi+27]
     mov   [menu_button_number],ax
   .dr_mn:


     call   draw_menu
  .rbutd:
    cmp   [wmsg],WM_RBUTTONDOWN
    jne   .r_up
    cmp   [lpipe_flag],1
    jne   @f
    mov   eax,[lparam]
    mov   dword[custom_rot_start_x],eax
  @@:
    mov   [tex_button_pres],1
    mov   eax,[lparam]
    mov   [tex_mov_begin_x],eax

  .r_up:
    cmp   [wmsg],WM_RBUTTONUP
    jne   @f
    mov   [tex_button_pres],0
    cmp   [lpipe_flag],1
    jne    @f
    mov   [custom_rot_start_x],0
   @@:
end if ; dd count
    cmp    [wmsg],WM_DESTROY
    je     .wmdestroy

  .defwndproc:
    invoke DefWindowProc,[hwnd],[wmsg],[wparam],[lparam]
    jmp    .finish

  .wmkeydown:
    cmp   [wparam],VK_UP
    jne   @f
    xor   eax,eax
    jmp   .finish
  @@:

    mov    eax,[wparam]
 ;   int3
    call   update_flags_after_keypress
if 1 ; dd counter
    cmp    [wparam],0x4d  ;  m
    jne    @f
    xor    eax,eax
    call   open_next_mesh
    mov    [NextPointsCount],eax
    mov    [NextTrianglesCount],ebx
    call   prepare_object
   @@:
    cmp    [wparam],0x4e  ;  n
    jne    @f
    mov    eax,-1
    call   open_next_mesh
    call   prepare_object
   @@:

    cmp   [wparam],0x58 ;  'x'
    jne   @f
    call   open_texture
  @@:
    cmp   [wparam],VK_ESCAPE
    jne   .finish
end if  ; dd count

  .wmdestroy:
    invoke PostQuitMessage,0
    xor    eax,eax

  .finish:    ; GUI buttons
 if 1   ; dd counter
    cmp   [menu_button_number],32  ; remove  redundant  vertices
    jne   .no_remove

      cmp   [tolerancy_flag],0
      jne   @f
      mov   eax,0.00009
      jmp   .rem_redu
    @@:
      cmp   [tolerancy_flag],1
      jne   @f
      mov   eax,0.0009
      jmp   .rem_redu
    @@:
      cmp   [tolerancy_flag],2
      jne   @f
      mov   eax,0.009
      jmp   .rem_redu
         @@:
      cmp   [tolerancy_flag],3
      jne   @f
      mov   eax,0.02
      jmp   .rem_redu
         @@:
      cmp   [tolerancy_flag],4
      jne   @f
      mov   eax,0.03
      jmp   .rem_redu
    @@:
      cmp   [tolerancy_flag],5
      jne   @f
      mov   eax,0.04
      jmp   .rem_redu
                  @@:
      cmp   [tolerancy_flag],6
      jne   @f
      mov   eax,0.1
      jmp   .rem_redu
   .rem_redu:
   pushad




      call   remove_redundant_vert

if 0
;        eax - approximation tolerancy of triangle vect
;              (as dword float)
;        ebx - vertices list ptr
;        esi - triangles list ptr
;        edi - vertex normal vectors ptr
;        ecx - vertices count  - integer
;        edx - triangles count - integer
;        xmm0 -approximation toleranc of normal vector
;               as dwoord float

         mov   ebx,0.0009
         movd  xmm0,ebx
         mov   ebx,[points_r_ptr]
         mov   esi,[triangles_ptr]
         mov   edi,[points_normals_ptr]
         movzx ecx,[points_count_var]
         movzx edx,[triangles_count_var]
         call  merge_triangles
       ;  mov   [points_count_var],dx

end if


      cmp    [triangles_count_var],100
      jge    @f
      mov    eax,100
      call   init_triangles_normals2
      jmp    .skip
    @@:
      call   init_tri_n_th
    .skip:
      cmp    [points_count_var],100
      jnl    @f
      mov    eax,100
      call   init_point_normals
      jmp    .skip2
    @@:
      call   init_pt_norm_th
    .skip2:
      call   calc_bumpmap_coords
    popad
      mov    eax,[edges_ptr]
      mov    ebx,[triangles_ptr]
      mov    ecx,[triangles_count_var]
      call   do_edges_list
      mov   [edges_count],ecx

     jmp    .end

    .no_remove:

    cmp   [menu_button_number],26  ; remove unused vertices
    jne   @f
    mov   eax,[triangles_ptr]
    mov   ebx,[points_r_ptr]
    mov   ecx,[triangles_count_var]
    mov   edx,[points_count_var]
    call  remove_unused_vertices
    mov   [points_count_var],ecx
    call  prepare_object
    jmp   .end
  @@:
    cmp   [menu_button_number],31  ; edit next mesh
    jne   @f
    call  NextMdraw_all
    movzx  eax,[whole_xres]
    movzx  ebx,[whole_yres]
    sub    ebx,(BYRES+100) ; 30
    sub    eax,BXRES

    invoke StretchDIBits,[hdc],eax,ebx,BXRES,BYRES,0,0,BXRES,BYRES,Bbuffer,Bbmi,0,SRCCOPY
    jmp    .end
  @@:
    cmp   [menu_button_number],3  ; zoom in
    jne   @f
    fninit
    fld   [scale]
    fmul  [scale_plus]
    fstp  [scale]
    jmp   .end
  @@:
    cmp   [menu_button_number],4  ; zoom out
    jne   @f
    fninit
    fld   [scale]
    fmul  [scale_minus]
    fstp  [scale]
    jmp   .end
  @@:

    cmp   [menu_button_number],5  ; tex map
    jne   @f         ; bump, texture - spherical or planar mapping
    call  calc_bumpmap_coords
    jmp   .end
  @@:
    cmp   [menu_button_number],6  ; random directional lights
    jne   @f
    call  make_random_lights
    call  normalize_all_light_vectors
    call  do_color_buffer
    call  init_envmap2
    jmp   .end
  @@:
    cmp   [menu_button_number],30 ; long pipe derivesv
    jne   @f                 ; init work, depend segs cnt.
    movzx eax,[lp_quality]
    add   eax,1
    shl   eax,2
    mov   [long_pipe_one_seg_c],eax
    jmp   .end


  @@:
    cmp   [menu_button_number],28 ; long pipe derivesv
    jne   .nx_but27               ; init work, depend segs cnt.
    call  long_pipe_init

    jmp   .end

  .nx_but27:
    cmp   [menu_button_number],27
    jne   .nx_but7

    cmp   [lpipe_flag],1
    jne   @f
    pushad
    push  dword long_pipe_derv ;[derv_nodes_ptr]
    mov   ecx,[long_pipe_segs_c]
    lea   ecx,[ecx*3+1]
    push  ecx
    call  normalize_object1
    popad
    jmp   .end
  @@:
    cmp   [lpipe_flag],2
    jne   .active_object_do    ;  .end

       ; calculate long pipe
       mov    esi,[long_pipe_vert_ptr]   ;[points_rotated_ptr]
       mov    ecx,[long_pipe_segs_c]
       cmp    [lpcurve_tp],0   ; bspline or bezier long pipe ?
       je     @f
       lea    ecx,[ecx*3-2]
       jmp    .bspl
     @@:
       sub    ecx,1
    .bspl:


       ; memory work
       pushad
       xor     eax,eax
       call    free_mem_for_tp1
       mov     ebx,[long_pipe_segs_c]
       mov     ecx,8   ; triangles per one turn
       imul    ebx,ecx
      ; int3
       mov     ecx,[long_pipe_one_seg_c]
       imul    ebx,ecx
       movzx   ecx,[lp_quality]
       add     ecx,2
       movzx   edx,[lpcurve_tp]
       add     edx,1
       imul    edx,edx
       imul    ebx,ecx
       imul    ebx,edx
       mov     [triangles_count_var],65500
       mov     [points_count_var],65500
       xor     eax,eax
       call    alloc_mem_for_tp1
       popad



       ; calc long pipe
       imul   ecx,[long_pipe_one_seg_c]
       mov    eax,4       ; rotary count per one turn
       mov    edx,15.15   ; radius
       movd   xmm0,edx
       mov    ebx,[triangles_ptr]
       mov    edi,[points_r_ptr]
; in:  esi  - ptr to curve vertices - middle points of pipe
;      ecx  - curve vertices count
;      eax  - rotary steps count as integer
;      xmm0 - radius, lowest dword float
;      ebx  - ptr to triangles list
;      edi  - ptr to pipe vertices
; out: ecx  - triangles count
;      ebx  - vertices count

       call   do_long_pipe
       mov    [triangles_count_var],ecx
       mov    [points_count_var],ebx
       call   prepare_object

    jmp   .end

   .nx_but7:
    cmp   [menu_button_number],7  ; displaying models
    jne   .nx_but8
    movzx  eax,[xres_var]   ; clear text with render metod text
    invoke StretchDIBits,[hdc],0,0,eax,30,0,0,1,1,fakebitmap,bmi,0,SRCCOPY
    cmp    [draw_flag],1
    jne    @f
    call   do_color_buffer
   @@:
    jmp   .end
  .nx_but8:
    cmp   [menu_button_number],8  ; bumps random/texture dependent
    jne   @f
    call  calc_bumpmap
  @@:
    cmp   [menu_button_number],9  ; bumps random/texture dependent
    jne   .next_no
    cmp   [tex_flag],0
    jne   @f
    call  generate_texture2    ; draw stripes
    jmp   .tex_end
  @@:
    cmp   [tex_flag],1
    jne   @f
    call  generate_texture3     ; draw xor
    jmp   .tex_end
  @@:
    call  generate_texture1     ; white area
  .tex_end:
    call  calc_bumpmap
  .next_no:
    cmp   [menu_button_number],10  ; save 3d object
    jne   @f
    call  write_asc
    push  [triangles_count_var]   ; pushes prevent crash in
    push  [points_count_var]      ; above 65535 vertices or tris
    call  save3ds                 ; situation
    pop   [points_count_var]
    pop   [triangles_count_var]
    jmp   .end
  @@:
    cmp   [menu_button_number],12  ; active segment chosen
    jne   .seg_ch_end
    cmp   [hrt_flag],3
    jl    .seg_ch_end
    inc   [current_seg]
    cmp   [current_seg],BSEGMENTS
    jne   @f
    mov   [current_seg],0
  @@:
    pushad
    call  Bdraw_all
    popad
    cmp  [hrt_flag],3
    jne  @f
    call save_object
    jmp   .end
  @@:
    cmp   [hrt_flag],4
    jne   @f
        ;call  save_pipe
    jmp   .end
  @@:
     ;   call  save_pipe_rot_solid
    jmp  .end
  .seg_ch_end:

    cmp   [menu_button_number],13  ; do displacement
    jne   @f
    pushad
    call  apply_displacement_from_tex
    popad
    call   prepare_object
    jmp   .end
   @@:


   cmp   [menu_button_number],14  ; do teslation
    jne   .triang_end
    call  triangulize_all_faces


    mov   eax,1
    call  alloc_mem_for_tp1

; remove slots proc
; in:
;     eax - ptr to edges
;     ebx - ptr to triangles
;     ecx - edges count
;     edx - ptr to vertices (points)
;     esi - triangles count
;     memory pointed by ebx, must be enough to write new
;     triangles in it.
; out:
;     ecx - new triangles count

 ;    pushad
 ;    mov   eax,[edges_ptr]
 ;    mov   ebx,[triangles_ptr]
 ;    mov   ecx,[edges_count]
 ;    mov   edx,[points_r_ptr]
 ;    movzx esi,[triangles_count_var]
 ;    mov   edi,0.0005
 ;    call  remove_slots
 ;    mov   [triangles_count_var],cx
 ;    popad


    xor    eax,eax
    call   prepare_object
  ;  mov   eax,[edges_ptr]
  ;  mov   ebx,[triangles_ptr]
  ;  mov   ecx,[triangles_count_var]
  ;  call  do_edges_list
  ;  mov   [edges_count],ecx
    jmp   .end

  .triang_end:


    cmp   [menu_button_number],15  ; current object
    jne   .active_obj_end
    mov   [NextPointsCount],0 ; disable next mesh
  .active_object_do:
    cmp   [hrt_flag],1
    jne   @f
    call  read_hrt
    xor   eax,eax
    call   prepare_object
    jmp    .end

  @@:
    cmp   [hrt_flag],2
    jne    @f
    call   load_tetrahedron
    xor    eax,eax
    call   prepare_object
    jmp    .end

  @@:
     cmp   [hrt_flag],3
     jne   @f
     mov   [triangles_count_var],BROTARY_STEPS*BSTEPS*BSEGMENTS*2
     mov   [points_count_var],BROTARY_STEPS*BSTEPS*BSEGMENTS*5/4
     xor   eax,eax
     call  alloc_mem_for_tp1
     pushad
     call  Bdraw_all
     popad
     call  save_object        ; save changes - on the fly write to points / triangles lists
     call  prepare_object

     jmp   .end

 @@:
     cmp   [hrt_flag],4
     jne   @f
     pushad
     call  Bdraw_all
     popad
     call  save_pipe        ; save changes - on the fly write to points / triangles lists
     call  prepare_object

     mov   eax,0.2
     movd  xmm0,eax
     movss xmm1,[scale]  ; proteza
     mulss xmm0,xmm1
     movss [scale],xmm0

     jmp   .end

  @@:
     cmp   [hrt_flag],5
     jne   @f
     pushad
     call  Bdraw_all
     popad
  ;   call  save_pipe_rot_solid       ; save changes - on the fly write to points / triangles lists
     call  prepare_object

     mov   eax,5.0
     movd  xmm0,eax
       ; proteza !!!
     mulss xmm0,[scale]
     movss [scale],xmm0


     jmp   .end
  @@:


    cmp    [file_ext_flag],0
    jne    @f
    pushad
    mov    ebx,[FileSize]
    mov    eax,[File_ptr]
    call   read_tp_variables1
    mov    [points_count_var],ecx
    mov    [triangles_count_var],ebx      ; 3ds
    xor    eax,eax
    call   alloc_mem_for_tp1

; read from 3ds file
;in:  eax - ptr to triangles
;     ebx - ptr to vertices
;     ecx - ptr to file begin
;     edx - ptr to file end
;out: ecx - points (vertices) count
;     edx - triangles count
    mov    eax,[triangles_ptr]
    mov    ebx,[points_r_ptr]
    mov    ecx,[File_ptr]
    call   read_from_file     ; 3ds
    mov    [EndFile],ebx
  ;  mov    [points_count_var],cx
  ;  mov    [triangles_count_var],dx
    popad
    xor    eax,eax
    call   prepare_object
    jmp    .end
   @@:
    call   read_asc_tp_variables  ; asc
    xor    eax,eax
    call   alloc_mem_for_tp1
    call   read_asc
    xor    eax,eax
    call   prepare_object
    jmp   .end
  .active_obj_end:

    cmp   [menu_button_number],16  ; random point light position
    jne   @f                       ; and plane equation
    call   random_point_light_position
    call   normalize_plane_equation
    jmp   .end
  @@:
    cmp   [menu_button_number],19  ; transparent displacement color
    jne   .next
    cmp   [disp_col_flag],0
    jne   @f
    mov   [displ_transparent_col],0x00ffffff
  @@:
    cmp   [disp_col_flag],1
    jne   @f
    mov   [displ_transparent_col],0
  @@:
    jmp   .end
  .next:

     ; ********check if draw bezier surfaces****************************
     cmp   [menu_button_number],21
     jne   .no_bez
     cmp   [bezier_flag],1
     jne   @f
     push  dword tpot_rect
     pop   dword[derv_rect_ptr]
     push  dword tpot_nodes
     pop   dword[derv_nodes_ptr]
     mov   [bez_patch_count],32
     mov   [bez_nodes_count],306
     jmp   .forward_bez
   @@:
     cmp   [bezier_flag],2
     jne   @f
     push  dword tcup_rect
     pop   dword[derv_rect_ptr]
     push  dword tcup_nodes
     pop   dword[derv_nodes_ptr]
     mov   [bez_patch_count],26
     mov   [bez_nodes_count],251
     jmp   .forward_bez
   @@:
     cmp   [bezier_flag],3
     jne  .active_object_do
     push  dword tspoon_rect
     pop   dword[derv_rect_ptr]
     push  dword tspoon_nodes
     pop   dword[derv_nodes_ptr]
     mov   [bez_patch_count],16
     mov   [bez_nodes_count],256

   .forward_bez:


     mov   [triangles_count_var],32*32*32*32
     mov   [points_count_var],32*32*32
     call  alloc_mem_for_tp1

    pushad
    push  [derv_nodes_ptr]
    movzx ecx,[bez_nodes_count]
    push  ecx
    call  normalize_object1
    popad


;     unpack list
    ;  bssurface_derv_unpacked

     mov   ebx,[derv_rect_ptr]   ;  ;derives_indexes
     mov   edi,bssurface_derv_unpacked
     mov   edx,bssurface_derv_unpacked_c     ; copy
     mov   esi,[derv_nodes_ptr]   ;derives_vertices
     movzx ecx,[bez_patch_count]
     shl   ecx,4
     push  ebp  ; ebp temporaly use
     xor   ebp,ebp

  .loop_ed_bpatch:
     movzx   eax,word[ebx]
     cmp     ebp,0
     jne     @f
     cmp     [derive_flag],0        ; bezier editing patch code
     jne     @f
     cmp     [ed_bpatch_derive_on],1
     jne     @f
     movzx   eax,[ed_bpatch_derive_no]
     cmp     ax,bx
     jne     @f
     movd    xmm0,[ed_bpatch_derives_x_start]
     movd    xmm1,[ed_bpatch_derives_x_end]
     psubw   xmm1,xmm0
     punpcklwd  xmm1,[the_zero]
     cvtdq2ps   xmm1,xmm1
  ;   movzx      eax,[ed_bpatch_derive_no]
     imul       eax,[i12]
     add        eax,[derv_nodes_ptr]
     movlps     xmm0,[eax]
     addps      xmm0,xmm1
     movlps     [eax],xmm0
     mov        ebp,1
 @@:





     imul    eax,[i12]
     add     eax,esi
     movups  xmm0,[eax]
     movlps  [edx],xmm0
     movlps  [edi],xmm0
     movhlps xmm0,xmm0
     movss   [edi+8],xmm0
     movss   [edx+8],xmm0
     add     edi,12
     add     edx,12
     add     ebx,2
     sub     ecx,1
     jnz     .loop_ed_bpatch


     pop     ebp

;  add to copy edition coordinates

;.....


;       esi - ptr  -> x11, y11, z11 .... x44, y44, z44 ; dword
;       edx - quality -> s shl 16 t
;       edi - ptr to area of calculated vertices
;       ebx - ptr to area of calculated faces (triangles)
;  out :
;       ecx - vertices count
;       edx - triangles count

     mov    [points_count_var],0
     mov    [triangles_count_var],0

     mov   esi,bssurface_derv_unpacked_c   ;bssurface_derives
     mov   edi,[points_r_ptr]
     mov   ebx,[triangles_ptr]
     push  ebp
     mov   ebp,esp
     sub   esp,16
     .bez_cnt   equ dword[ebp-4]
     .cur_tri_l equ dword[ebp-8]
     mov   .bez_cnt, 0   ; counter
     mov   .cur_tri_l,0


   .bez_again:

     mov    edx,8 shl 16 + 8
     push   ebx
     push   esi
     call   calc_bezier_surface          ; calc_bssurface
     add    [points_count_var],ecx
     add    [triangles_count_var],edx
   ;  int3
     pop    esi  ; esi - derives
     add    esi,12*16
     pop    .cur_tri_l
     add    .bez_cnt,1
     cmp    .bez_cnt,1
     jz     .bez_again

;  .bez_again2:
         ; modify triangles_list
     push   ebx  ; ebx - cur tri ptr
     push   ecx
     push   edx
   ;  movzx  ecx,cx
     mov    ebx,.bez_cnt
     sub    ebx,1
     imul   ecx,ebx   ;.bez_cnt
     mov    ebx,.cur_tri_l
  ;     ecx - vertex count, pro one patch
  ;     edx - cur triangles count ( pro one patch )

   @@:
     add    [ebx],ecx
     add    [ebx+4],ecx
     add    [ebx+8],ecx
     add    ebx,12
     sub    edx,1
     jnz    @b

     pop    edx
     pop    ecx
     pop    ebx
     movzx  eax,[bez_patch_count]
     cmp    .bez_cnt,eax
     jne    .bez_again



  .end_bez:

 ; remove redundant vertices witch are on border
 ; of each 'sticked' patch
      cmp   [tolerancy_flag],0
      jne   @f
      mov   eax,0.009
      jmp   .rem_red
    @@:
      cmp   [tolerancy_flag],1
      jne   @f
      mov   eax,0.0009
      jmp   .rem_red
    @@:
      cmp   [tolerancy_flag],2
      jne   @f
      mov   eax,0.00009
      jmp   .rem_red

   .rem_red:
      call  remove_redundant_vert
   ; normals thread work
      cmp    [triangles_count_var],100
      jge    @f
      mov    eax,100
      call   init_triangles_normals2
      jmp    .skip_n
    @@:
      call   init_tri_n_th
    .skip_n:
      cmp    [points_count_var],100
      jnl    @f
      mov    eax,100
      call   init_point_normals
      jmp    .skip_n2
    @@:
      call   init_pt_norm_th
    .skip_n2:

      call   calc_bumpmap_coords

      mov    eax,[edges_ptr]
      mov    ebx,[triangles_ptr]
      mov    ecx,[triangles_count_var]
      call   do_edges_list
      mov   [edges_count],ecx

      add    esp,16
      pop    ebp
     jmp   .end
  .no_bez:

    cmp   [menu_button_number],20  ; displacement factor
    jne   .next1
    cmp   [disp_fac_flag],0
    jne   @f
    mov   [displac_div],6800
  @@:
    cmp   [disp_fac_flag],1
    jne   @f
    mov   [displac_div],13000
  @@:
    cmp   [disp_fac_flag],2
    jne   @f
    mov   [displac_div],30000
  @@:
    cmp   [disp_fac_flag],3
    jne   @f
    mov   [displac_div],60000
  @@:
;   jmp    .end
  .next1:
;    cmp   [menu_button_number],31  ; load next mesh
;    jne   .next2
;    call  open_next_mesh



;   jmp    .end
;  .next2:



;    jmp   .end
end if  ; dd counter

  .end:
    mov   [menu_button_number],0
    ret

endp

;******************************************************
;*****************                      ***************
;*****************  Various procedures  ***************
;*****************                      ***************
;******************************************************
if 1
open_next_mesh:
  ; in - eax=0 -> join to existing mesh, other load new mesh
  ;
  ; allocate mem, open from file and join to existing object  if eax=0
  ; out - eax - points count  of new mesh
  ;       ebx - triangles count of new mesh

  push  ebp
  mov   ebp,esp
  sub   esp,64
  .FileSize      equ  [ebp-4]
  .hfile         equ  [ebp-8]
  .pmem          equ  [ebp-12]
  .file_ptr      equ  [ebp-16]
  .BytesRead     equ  [ebp-20]
  .file_ext_flag equ  [ebp-24]
  .points_count  equ  [ebp-28]
  .tri_count     equ  [ebp-32]
  .vertices_ptr  equ  [ebp-36]
  .tri_ptr       equ  [ebp-40]
  .JoinNew       equ  [ebp-44]
  .hheap         equ  [ebp-48]
  .BytesRead     equ  [ebp-52]
  .NMmem         equ  [ebp-56]

   mov     .JoinNew,eax

   mov     [ofn.Flags],OFN_FILEMUSTEXIST+OFN_PATHMUSTEXIST+OFN_EXPLORER
   invoke  GetOpenFileName,ofn
   or      eax,eax
   jz      .end

   invoke  CreateFile,fname,GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0
   mov     [hfile],eax
   cmp     eax,-1
   je      .failed
   invoke  GetFileSize,[hfile],0
   cmp     eax,-1
   je      .failed
   mov     [FileSize],eax
   invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
   mov     [File_ptr],eax
   invoke  ReadFile,[hfile],eax,[FileSize],BytesRead,0
   or      eax,eax
   jz      .read_error
   invoke  CloseHandle,[hfile]
   mov     [hfile],0





   mov    ebx,[FileSize]   ;.BytesRead   ;.FileSize
   mov    eax,[File_ptr]  ;[NMmem]
   call   read_tp_variables1
   mov    .points_count,ecx
   mov    .tri_count,ebx

   cmp    dword .JoinNew,0  ; join to existing object
   je     @f                ; or load diffrent?
   mov    [points_count_var],0
   mov    [triangles_count_var],0
 ;  invoke  HeapFree,.hheap,0,.FileSize
;   invoke  VirtualFree, [File_ptr] , 0,  MEM_RELEASE
if 0
   push   dword File_ptr
   pop    [File_ptr]
   push   dword .FileSize
   pop    [FileSize]
end if
 @@:

; alloc some mem to new and old vertices   and new and old triangles lists
   mov     eax,dword .points_count
   mov     ebx,[points_count_var]
   add     eax,ebx
;   cmp     eax,65534
;   ja      .failed
   imul    eax,[i12]
   add     eax,20
   invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
   or      eax,eax
   jz      .failed
   mov     .vertices_ptr,eax

   mov     eax,dword .tri_count
   mov     ebx,[triangles_count_var]
   add     eax,ebx
;   cmp     eax,65534
;   ja      .failed
   imul    eax,[i12]
   add     eax,65536
   invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
   or      eax,eax
   jz      .failed
   mov     .tri_ptr,eax

   cmp     dword .JoinNew,0
   jne     .nocopy
;   copy old to new lists
     mov   edi,.tri_ptr
     mov   esi,[triangles_ptr]
     mov   ecx,[triangles_count_var]
  @@:
     movq   xmm0,[esi]
     movd   xmm1,[esi+8]
     movq   [edi],xmm0
     movd   [edi+8],xmm1
     add    edi,12
     add    esi,12
     sub    ecx,1
     jnz    @b

     mov   edi,.vertices_ptr
     mov   esi,[points_r_ptr]
     mov   ecx,[points_count_var]
  @@:
     movq   xmm0,[esi]
     movd   xmm1,[esi+8]
     movq   [edi],xmm0
     movd   [edi+8],xmm1
     add    edi,12
     add    esi,12
     sub    ecx,1
     jnz    @b
     pushad
     push  dword .vertices_ptr
     push  [points_count_var]
     call  normalize_object1
     popad
  .nocopy:

; read from second file
;in:  eax - ptr to triangles
;     ebx - ptr to vertices
;     ecx - ptr to file begin
;     edx - ptr to file end
;out: ecx - points (vertices) count
;     edx - triangles count
   mov   eax,.tri_ptr
;   sub   eax,6
   mov   ebx,[triangles_count_var]
   imul  ebx,[i12]
   add   eax,ebx   ; offset in triangles list
;   add   eax,12

   push  eax
   mov   ebx,.vertices_ptr
   mov   edx,[points_count_var]
   imul  edx,[i12]
   add   ebx,edx    ; offset in vartices list
   pushad
   push  ebx
   mov   ecx,[File_ptr]
   mov   edx,ecx
   add   edx,[FileSize]
   call  read_from_file

;   vertices ptr previously pushed
 ;  mov   ecx,.points_count
 ;  inc   ecx
   push  ecx
   call  normalize_object1
   popad

;fix triangles list
   pop   ebx
   mov   eax,[points_count_var]
   mov   ecx,dword .tri_count
   lea   ecx,[ecx*3]
 @@:
   add  dword [ebx],eax
   add  ebx,4
   sub  ecx,1
   jnz  @b

 ;  mov   eax,[points_count_var]
 ;  mov   ebx,[triangles_count_var]
   mov   eax,.points_count
   mov   ebx,.tri_count
   add   [points_count_var],eax
   add   [triangles_count_var],ebx

  xor eax,eax
  call free_mem_for_tp1 ; release all buffers mem
  mov  eax,1
  call alloc_mem_for_tp1 ; alloc all but triangles_ptr, points_r_ptr
                        ; and shadoe vertices ptr

   mov     eax,[points_count_var]
   imul    eax,[i12]
   add     eax,20
   invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
   or      eax,eax
   jz      .failed
   mov     [shadow_vertices_ptr],eax


   push  dword .vertices_ptr
   pop   [points_r_ptr]
   push  dword .tri_ptr
   pop   [triangles_ptr]


   invoke  CloseHandle,[hfile]
   mov     [hfile],0

 .init_ptr:
   invoke  InvalidateRect,[hwnd],0,1
   xor     eax,eax
   jmp     .end
 .read_error:
 .failed:
   cmp     dword [hfile],0
   je      @f
   invoke  CloseHandle,[hfile]
  @@:
   jmp     .end
   invoke  MessageBox,[hwnd],_not3ds,0,MB_OK
 ;  .cancel:
 ;  invoke  ReleaseDC,[hwnd],[hdc]
 ;  invoke  DestroyWindow,[hwnd]
 ;  invoke  ExitProcess,[msg.wParam]
  .end:
   mov    eax,.points_count
   mov    ebx,.tri_count

   mov    esp,ebp
   pop    ebp
ret

end if

long_pipe_init: ; calc long pipes control points
                ; segments count depend


    cmp   [lpsegs_flag],0
    jne   @f
    mov   [long_pipe_segs_c],2
    jmp   .calll
  @@:
    cmp   [lpsegs_flag],1
    jne   @f
    mov   [long_pipe_segs_c],4
    jmp   .calll
  @@:
    cmp   [lpsegs_flag],2
    jne   @f
    mov   [long_pipe_segs_c],8
    jmp   .calll
  @@:
  ;  cmp   [lpsegs_flag],2
  ;  jne   @f
    mov   [long_pipe_segs_c],11
  .calll:
    mov   esi,long_pipe_derv_init
    mov   edi,long_pipe_derv+12
    mov   ecx,[long_pipe_segs_c]
    lea   ecx,[ecx*3]
    add   ecx,1
    call  init_long_pipe_derives
ret

prepare_object:
;    in: eax=1 - no edges
    push   ebp
    mov    ebp,esp
    sub    esp,8
    .edg   equ dword[ebp-4]
    mov    .edg,eax
    pushad
    push   dword [points_r_ptr]
    mov    ecx,[points_count_var]
    push   ecx
    call   normalize_object1
    cmp    [triangles_count_var],200
    jge    @f
    mov    eax,100
    call   init_triangles_normals2
    jmp    .skip
  @@:
    call   init_tri_n_th
  .skip:
    cmp    [points_count_var],200
    jnl    @f
    mov    eax,100
    call   init_point_normals
    jmp    .skip2
  @@:
    call   init_pt_norm_th
  .skip2:
   popad
   call    calc_bumpmap_coords
; do edges list
; in  ->
;        eax - ptr to edges list in memory (there must be
;              declared enough size of mem)
;        ebx - ptr to faces
;        ecx - faces(triangles) count
;
; out ->
;        desired list, each edge in list as two words - indices
;          two position in vertices list, pointed by [points_r_ptr]
;        ecx - edges count
    cmp   .edg,1
    je    @f
    mov   eax,[edges_ptr]
    mov   ebx,[triangles_ptr]
    mov   ecx,[triangles_count_var]
    call  do_edges_list
    mov   [edges_count],ecx
    @@:
    mov   esp,ebp
    pop   ebp
ret

update_flags_after_keypress:
  ;   in ax - wparam
     pushad
     mov   esi,menu_data
   @@:
     cmp   word[esi+29],ax
     je    .updat
     cmp   word[esi+31],ax
     je    .updat
     add   esi,33
     cmp   word[esi],-1
     jnz   @b
     jmp   .end
  .updat:
     mov   al,byte[esi+22]
     inc   al
     cmp   al,[esi+21]
     jne   @f
     xor   al,al
   @@:
     mov   [esi+22],al
     mov   ax,[esi+27]
     mov   [menu_button_number],ax

     mov    edi,menu_screen
     mov    ecx,MXRES*MYRES
     xor    eax,eax
   @@:
     mov    [edi],eax
     sub    ecx,1
     jnz    @b
     call   draw_menu
   .end:
     popad
ret


open_texture:

   mov     [ofntex.Flags],OFN_FILEMUSTEXIST+OFN_PATHMUSTEXIST+OFN_EXPLORER
   invoke  GetOpenFileName,ofntex
   or      eax,eax
   jz      .cancel

   invoke  CreateFile,fnametex,GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0
   mov     [hfiletex],eax
   cmp     eax,-1
   je      .failed
   invoke  GetFileSize,[hfiletex],0
   cmp     eax,-1
   je      .failed
   cmp     eax,TEX_X*TEX_Y*3
   jnae    .failed
   mov     [FileSize],eax
   cmp     [pmemtex],0   ;pmem !!! non pmemtex ?
   jne     .heap_realloc
   invoke  GetProcessHeap
   mov     [hheap],eax
   invoke  HeapAlloc,[hheap],0,[FileSize]
   jmp     .read_file

 .heap_realloc:
   invoke  HeapReAlloc,[hheap],0,[pmem],[FileSize]
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; read raw file
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 .read_file:      ; read texture
   mov     [pmemtex],eax
   mov     ebx,eax
   invoke  ReadFile,[hfiletex],[pmemtex],[FileSize],BytesRead,0
   or      eax,eax
   jz      .read_error
   invoke  CloseHandle,[hfiletex]
   mov     [hfiletex],0

   mov     esi,[pmemtex]  ;ebx
   mov     edi,texmap
   cld

   mov      ecx,TEX_X*TEX_Y  ;eax

   xor      eax,eax
 @@:
   movsw
   movsb
   stosb
   loop  @b

   mov     eax,[BytesRead]
  .init_ptr:
   mov     [File_tex_ptr],ebx
   invoke  InvalidateRect,[hwnd],0,1
   xor     eax,eax
   ret


 .read_error:
   invoke  HeapFree,[hheap],0,[pmemtex]
   mov     [pmemtex],0

 .failed:
   cmp     [hfiletex],0
   je      @f
   invoke  CloseHandle,[hfiletex]
  @@:
   invoke  MessageBox,[hwnd],_notraw,0,MB_OK
   .cancel:
ret


open_file:

   mov     [ofn.Flags],OFN_FILEMUSTEXIST+OFN_PATHMUSTEXIST+OFN_EXPLORER
   invoke  GetOpenFileName,ofn
   or      eax,eax
   jz      .cancel

   invoke  CreateFile,fname,GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0
   mov     [hfile],eax
   cmp     eax,-1
   je      .failed
   invoke  GetFileSize,[hfile],0
   cmp     eax,-1
   je      .failed
   mov     [FileSize],eax
   invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
   mov     [File_ptr],eax
   invoke  ReadFile,[hfile],eax,[FileSize],BytesRead,0
   or      eax,eax
   jz      .read_error
   invoke  CloseHandle,[hfile]
   mov     [hfile],0

   mov     eax,[BytesRead]
   mov     ebx,[File_ptr]
   mov     cx,WORD [ebx]
   cmp     cx,'MM'
   je     @f
   mov     [file_ext_flag],1  ; asc
   jmp     .init_ptr
 @@:
   mov     [file_ext_flag],0
 .init_ptr:
   invoke  InvalidateRect,[hwnd],0,1
   xor     eax,eax
   ret
 .read_error:
 .failed:
   cmp     [hfile],0
   je      @f
   invoke  CloseHandle,[hfile]
  @@:
   invoke  MessageBox,[hwnd],_not3ds,0,MB_OK
   .cancel:
   invoke  ReleaseDC,[hwnd],[hdc]
   invoke  DestroyWindow,[hwnd]
   invoke  ExitProcess,[msg.wParam]
ret

rotate_normals:
   push   ebp
   mov    esi,[points_normals_ptr]           ; rotate normals
   mov    edi,[points_normals_rotated_ptr]
   mov    ecx,[points_count_var]
   lea    ebx,[matrix]
   call   rotary

   cmp    [normals_flag],1
   jne    @f
   mov    esi,[triangles_normals_ptr]           ; rotate normals
   mov    edi,[triangles_normals_rotated_ptr]
   mov    ecx,[triangles_count_var]
   lea    ebx,[matrix]
   call   rotary
 @@:
   pop ebp
ret

rotate_points:
   mov  esi,matrix
   mov  edi,matrix_scaled
   mov  ecx,36/4
 @@:
   push dword[esi]
   pop  dword[edi]
   add  esi,4
   add  edi,4
   sub  ecx,1
   jnz  @b

   lea  esi,[matrix_scaled] ;esi - pointer to 3x3 matrix
   mov  ebx,scale     ;ebx - ptr to scale
   call add_scale_to_matrix

   cmp    [bezier_flag],0
   je     .no_bez_rot

   mov    esi,[derv_nodes_ptr] ; bssurface_derives
    ; rotate bezier surf. nodes
   mov    edi,derv_nodes_rot
   movzx  ecx,[bez_nodes_count]
   lea    ebx,[matrix_scaled]
   call   rotary

    movdqu   xmm3,[xxadd]
    cvtdq2ps xmm3,xmm3
    xor      ecx,ecx
    movzx    eax,[bez_nodes_count]
    mov      ebx,derv_nodes_rot  ;bssurface_derives_rotated
  @@:
    movups   xmm1,[ebx]
    addps    xmm1,xmm3
    movlps   [ebx],xmm1
    movhlps  xmm1,xmm1
    movss    [ebx+8],xmm1
    add      ebx,12
    add      ecx,1
    cmp      ecx,eax
    jnz      @b
 .no_bez_rot:
   lea    ebx,[matrix_scaled]
   mov    ecx,[points_count_var]
   sub    ecx,[NextPointsCount]
   mov    esi,[points_r_ptr]
   mov    edi,[points_rotated_ptr]
   call   rotary

   cmp      [NextPointsCount],0    ; Next mesh custom scale
   je       @f
   movsx    eax,[NextScale]
 ;  sub      ax,BYRES
 ;  cwde
   mov      ebx,BYRES/2
   cvtsi2ss xmm0,eax
   cvtsi2ss xmm1,ebx
   divss    xmm0,xmm1
   movss    [NextScaleR],xmm0

   movd     xmm7,esi
   lea      esi,[matrix_scaled]  ;esi - pointer to 3x3 matrix
   mov      ebx,NextScaleR       ;ebx - ptr to scale
   call     add_scale_to_matrix

   movd   esi,xmm7
 ;  sub    esi,12
 ;  sub    edi,12
   lea    ebx,[matrix_scaled]
   mov    ecx,[NextPointsCount]

   call   rotary

 @@:
ret

th_peel0:
  mov  eax,0
  call peel
ret
th_peel1:
  mov  eax,1
  call peel
ret
th_peel2:
  mov  eax,2
  call peel
ret
th_peel3:
  mov  eax,3
  call peel
ret
peel:
   push  ebp
   mov   ebp,esp
   sub   esp,128
   and   ebp,-16
.th_no  equ dword [ebp-32]
.x_max  equ dword [ebp-4]
.x_min  equ dword [ebp-8]
.y_max  equ dword [ebp-12]
.y_min  equ [ebp-16]
.index1 equ [ebp-20]
.index2 equ [ebp-24]
.index3 equ [ebp-28]
.coord1 equ [ebp-40]
.coord2 equ [ebp-44]
.coord3 equ [ebp-48]

    mov    .th_no,eax
    movzx  ebx,[yres_var]
    movzx  eax,[yres_var]
    shr    ebx,1   ; shr ebx,n ;n=1 - 2 threads, n=2 - 4 threads...
    movzx  edx,[xres_var]
    movzx  ecx,[xres_var]
    shr    edx,1

    cmp    .th_no,0
    jne    @f
    mov    .y_min,dword 0
    mov    .x_min,0
    mov    .y_max,ebx
    mov    .x_max,edx
    jmp    .skip_div
   @@:
    cmp    .th_no,1
    jne    @f
    mov    .y_min,dword 0
    mov    .x_min,edx
    mov    .x_max,ecx
    mov    .y_max,ebx
    jmp    .skip_div
   @@:
    cmp    .th_no,2
    jne    @f
    mov    .y_max,eax
    mov    .y_min,ebx
    mov    .x_min,0
    mov    .x_max,edx
    jmp    .skip_div
  @@:
    cmp    .th_no,3
    jne    @f

    mov    .y_max,eax
    mov    .y_min,ebx
    mov    .x_min,edx
    mov    .x_max,ecx
   ; jmp    .skip_div
  @@:
  .skip_div:
    mov    esi,[triangles_ptr]
    mov    ecx,[triangles_count_var]
  .again_stencil_tri:
    push  esi
    push  ecx
    mov   eax,dword[esi]
    mov   ebx,dword[esi+4]
    mov   ecx,dword[esi+8]
    imul  eax,[i12]
    imul  ebx,[i12]
    imul  ecx,[i12]
    add   eax,[points_rotated_ptr]
    add   ebx,[points_rotated_ptr]
    add   ecx,[points_rotated_ptr]
    movss  xmm3,[eax+8]
    movss  xmm4,[ebx+8]
    movss  xmm5,[ecx+8]
    movlps xmm0,[eax]
    movlps xmm1,[ebx]
    movlps xmm2,[ecx]

    shufps  xmm4,xmm4,11110011b
    orps    xmm3,xmm4
    movlhps xmm3,xmm5  ; xmm3 - z float pack

    shufps xmm0,xmm0,11100001b
    shufps xmm1,xmm1,11100001b
    shufps xmm2,xmm2,11100001b

    cvtps2dq xmm0,xmm0
    cvtps2dq xmm1,xmm1
    cvtps2dq xmm2,xmm2
    packssdw xmm0,xmm0
    packssdw xmm1,xmm1
    packssdw xmm2,xmm2
    movd     eax,xmm0
    movd     ebx,xmm1
    movd     ecx,xmm2

    movaps   xmm0,xmm3

    movdqa     xmm5,.y_min
    mov      edi,[slices_ptrs_buff_ptr]
    mov      esi,[slices_counter_ptr]
    call     slices_tri_var
  .next_stc:
    pop      ecx
    pop      esi
    add      esi,12
    sub      ecx,1
    jnz      .again_stencil_tri



if 1

    mov    esi,[triangles_ptr]
    mov    ecx,[triangles_count_var]
  .again_peel_tri:
    push  esi
    push  ecx
    mov   eax,dword[esi]
    mov   ebx,dword[esi+4]
    mov   ecx,dword[esi+8]
    imul  eax,[i12]
    imul  ebx,[i12]
    imul  ecx,[i12]
    mov   .index1,eax
    mov   .index2,ebx
    mov   .index3,ecx
    add   eax,[points_normals_rotated_ptr]
    add   ebx,[points_normals_rotated_ptr]
    add   ecx,[points_normals_rotated_ptr]
    movups xmm0,[eax]
    movups xmm1,[ebx]
    movups xmm2,[ecx]
    andps  xmm0,[zero_hgst_dd]
    andps  xmm1,[zero_hgst_dd]
    andps  xmm2,[zero_hgst_dd]
    movups xmm3,[lights]
    andps  xmm3,[zero_hgst_dd]

     mov   eax,.index1
     mov   ebx,.index2
     mov   ecx,.index3
     add   eax,[points_rotated_ptr]
     add   ebx,[points_rotated_ptr]
     add   ecx,[points_rotated_ptr]

     movups   xmm4,[eax]
     movups   xmm5,[ebx]
     movups   xmm6,[ecx]
     cvtps2dq xmm4,xmm4
     cvtps2dq xmm5,xmm5
     cvtps2dq xmm6,xmm6
     pshufd   xmm4,xmm4,11100001b
     pshufd   xmm5,xmm5,11100001b
     pshufd   xmm6,xmm6,11100001b
     packssdw xmm4,xmm4
     packssdw xmm5,xmm5
     packssdw xmm6,xmm6
     movd     .coord1,xmm4
     movd     .coord2,xmm5
     movd     .coord3,xmm6
     movss    xmm4,[eax+8]
     movss    xmm5,[ebx+8]
     movss    xmm6,[ecx+8]
     shufps   xmm5,xmm5,11100001b  ; z coord work
     orps     xmm4,xmm5
     movlhps  xmm4,xmm6
     movdqa   xmm5,.y_min

     mov    eax, .index3      ; bump map coords
     cdq
     idiv   [i3]
     add    eax,[tex_points_ptr]
     push   dword[eax]

     mov    eax, .index2      ; bump map coords
     cdq
     idiv   [i3]
     add    eax,[tex_points_ptr]
     push   dword[eax]

     mov    eax, .index1      ; bump map coords
     cdq
     idiv   [i3]
     add    eax,[tex_points_ptr]
     push   dword[eax]
     movdqu xmm6,[esp]
     add    esp,12

     movzx      edx,[xres_var]
     movd       xmm7,edx
     pshufd     xmm7,xmm7,00111111b
     pand       xmm6,[zero_hgst_dd]
     por        xmm6,xmm7
     movd       xmm7,[slices_counter_ptr]

     mov      eax,.coord1
     mov      ebx,.coord2
     mov      ecx,.coord3
;     ror      eax,16
;     ror      ebx,16
;     ror      ecx,16

     mov      edx,texmap
     mov      edi,[screen_ptr]
     mov      esi,[slices_ptrs_buff_ptr]
     call     glass_tex_tri_sl_var

     pop      ecx
     pop      esi
     add      esi,12
     sub      ecx,1
     jnz      .again_peel_tri
end if

    add  esp,128
    pop  ebp
ret
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; g_draw - draw to screen buffer; main program proc
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
proc g_draw ;uses ebx esi edi
local .temp dw ?, .rot_flag dw ?
;local .first_p_index dd ?, .second_p_index dd ?, .third_p_index dd ?
;local .zz1 dw ?, .zz2 dw ?, .zz3 dw ? ;-> don't change order
local .zz1f dd ?, .zz2f dd ?,.zz3f dd ?
local .xx1f dd ?,.xx2f dd ?,.xx3f dd ?
local .yy1f dd ?,.yy2f dd ?,.yy3f dd ?
local .xx1 dw ?, .xx2 dw ?, .xx3 dw ?
local .yy1 dw ?, .yy2 dw ?, .yy3 dw ?
local .xy1 dd ?, .xy2 dd ?, .xy3 dd ?
local .dpr dd ?, .dpr_p dd ?
local .x_temp_var dd ?, .y_temp_var dd ?
local .counter dd ?
local .sinx dd ?, .cosx dd ?
local .siny dd ?, .cosy dd ?
local .sinz dd ?, .cosz dd ?
;local .matrix rb 36
local .shadow_matrix rb 64
local .dot_pr_light   dd ?
;local .p_light_vec rd 4
local .shd_vertices rd  12
local .transparent_voxels_count dd ?
;local .x_max dw ?,.x_min dw ?,y_max dw ?,y_min dw ?
;local .shd_counter dd ?
; local .temp_vect dqword ?
    pushad
    cmp   [blur_flag],1
    jne   @f
    xor   eax,eax          ; clear draw buffer
    mov   edi,[screen_ptr]
    movzx ebx,[xres_var]
    movzx ecx,[yres_var]
    imul  ecx,ebx
    call  fill_aligned
    jmp   .no_blur
  @@:
    mov   ecx,1
    call  blur_screen
  .no_blur:
    cmp   [draw_flag],9
    je    @f
    cmp   [draw_flag],10
    je    @f
    cmp   [draw_flag],13 ;5
    jge   @f
    mov   eax,5000*65536    ; make background in Zbuffer
    jmp   .m_back
  @@:
    mov   eax,6000.1
  .m_back:
    mov   edi,[Zbuffer_ptr]
    movzx ecx,[xres_var]
    movzx ebx,[yres_var]
    imul  ecx,ebx
    cmp   [draw_flag],19
    jne   @f
    shl   ecx,2  ; slices need more memory
   @@:
    call  fill_aligned



    cmp   [draw_flag],15
    ; real phong with slices cause
    jne   @f
    mov   edi,[slices_ptrs_buff_ptr]
   ; i use temporaly this buff as stencil buff
    movzx ecx,[xres_var]
    movzx ebx,[yres_var]
    imul  ecx,ebx
    call  fill_aligned
   @@:
  .rotate:

   fninit
   fld   [one_deg]
   fld   st
   fimul dword[main_rotary+8]   ; z
   fsincos
   fstp  [.sinx]
   fstp  [.cosx]
   fld   st
   fimul dword[main_rotary] ; x
   fsincos
   fstp  [.siny]
   fstp  [.cosy]
   fimul dword[main_rotary+4] ; y
   fsincos
   fstp  [.sinz]
   fstp  [.cosz]

   pushad
   push   [.sinx]
   push   [.cosx]
   push   [.siny]
   push   [.cosy]
   push   [.sinz]
   push   [.cosz]
   lea    edi,[matrix]
   call   make_rotation_matrix
   ; edi -  desired matrix
   popad
if 0    ; I roll off threads to easier debugging
    invoke CreateThread,NULL,NULL,rotate_points,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID0]
    mov    [hthread0],eax
    invoke SetThreadPriority,hthread0,THREAD_PRIORITY_TIME_CRITICAL
    invoke SetThreadAffinityMask,[hthread0],1
    invoke ResumeThread,[hthread0]

    invoke CreateThread,NULL,NULL,rotate_normals,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID1]
    mov    [hthread1],eax
    invoke SetThreadPriority,hthread1,THREAD_PRIORITY_TIME_CRITICAL
    invoke SetThreadAffinityMask,[hthread1],2
    invoke ResumeThread,[hthread1]



    invoke WaitForMultipleObjectsEx,2,hthread0,1,-1,0
    invoke CloseHandle,[hthread0]
    invoke CloseHandle,[hthread1]
end if
     call rotate_points
     call rotate_normals


    cmp    [draw_flag],0
    jne    .no_shadow
    ;  draw background image
        pushad
        push    [yres_var]
        push    word 0
        push    word 5500
        push    word 5500
        push    word 5500
        push    dword 1 * 65536 + 511
        push    dword 511 * 65536 + 1
        push    dword 1 * 65536 + 1

        mov    eax, 1 * 65536 + 1
        mov    bx, [xres_var]
        shl    ebx,16
        mov    bx, 1
        mov    ecx, 1 * 65536
        mov    cx,[yres_var]

        mov     edi,[screen_ptr]   ; screen buffer
        mov     esi,texmap
        mov     edx,[Zbuffer_ptr]
        call    tex_triangle_z

        push    [yres_var]
        push    word 0
        push    word 5500
        push    word 5500
        push    word 5500
        push    dword 1 * 65536 + 511
        push    dword 511 * 65536 + 511
        push    dword 511*65536 + 1

        mov    ax,[xres_var]
        shl    eax,16
        mov    ax,1
        mov    bx, [xres_var]
        shl    ebx,16
        mov    bx, [yres_var]
        mov    ecx, 1 * 65536
        mov    cx,[yres_var]

        mov     edi,[screen_ptr]   ; screen buffer
        mov     esi,texmap
        mov     edx,[Zbuffer_ptr]
        call    tex_triangle_z
        popad


        ;make background in shd buff
        movzx eax,[xres_var]
        movzx ecx,[yres_var]
        mul   ecx
        mov   ecx,eax
        shr   ecx,2
        xor   eax,eax
        mov   edi,[shd_buffer_ptr]
        cld
        rep   stosd


        mov   ecx,[triangles_count_var]
        mov   esi,[triangles_ptr]

   .again_shd_tri:
        push   esi
        push   ecx
        mov    [shd_counter],0
      .next_shd_vert:
        mov     eax,[shd_counter]
        shl     eax,2
        mov     edx,dword[eax+esi]
        imul    edx,[i12]
        add     edx,[points_rotated_ptr]  ; [edx] - cur vertex
        movups  xmm0,[edx]
        andps   xmm0,[zero_hgst_dd]
        subps   xmm0,[point_light_coords]
        movaps  xmm1,xmm0
        mulps   xmm1,xmm1
        haddps  xmm1,xmm1
        haddps  xmm1,xmm1
        andps   xmm1,[zero_hgst_dd]
        rsqrtps xmm1,xmm1
        mulps   xmm0,xmm1    ; xmm0 - normalized light vector
        andps   xmm0,[zero_hgst_dd]
        movups  [p_light_vec],xmm0  ;+eax

        pushad
        lea    ebx,[.shadow_matrix]
        lea    esi,[p_light_vec]   ; +eax
        call   calc_shadow        ; calculate 4x4 shadow matrix
        popad
        pushad
        mov    esi,edx
        mov    ecx,1
        mov    eax,[shd_counter]
        shl    eax,4
        lea    ebx,[.shadow_matrix]
        lea    edi,[.shd_vertices]
        add    edi,eax
        call   mul_matrix4x4
        popad
        add    [shd_counter],1
        cmp    [shd_counter],3
        jnz    .next_shd_vert

        lea      ebx,[.shd_vertices]
        movups   xmm0,[ebx]
        movdqu   xmm3,[xxadd]
        movups   xmm1,[ebx+16]
        movups   xmm2,[ebx+32]

        cvtps2dq xmm0,xmm0
        cvtps2dq xmm1,xmm1
        cvtps2dq xmm2,xmm2
        paddd    xmm0,xmm3
        paddd    xmm1,xmm3
        paddd    xmm2,xmm3

        packssdw xmm0,xmm0
        packssdw xmm1,xmm1
        packssdw xmm2,xmm2
        movd     eax,xmm0
        movd     ebx,xmm1
        movd     ecx,xmm2
        ror      eax,16
        ror      ebx,16
        ror      ecx,16
        mov      edx,0xffafafaf
        mov      edi,[screen_ptr]
        mov      esi,[shd_buffer_ptr]

        pushad
        call    triangle_shd
        popad

        pop     ecx
        pop     esi
        add     esi,12
        dec     ecx
        jnz    .again_shd_tri

.no_shadow:


  @@:


    cmp        [edit_vertex_no],0
    ; what if you want check vertex no 0 ?
    ; It should be dd -1 instead
    je         @f
    movd       xmm0,dword[edit_end_x]
    movd       xmm1,dword[edit_start_x]
    pxor       xmm2,xmm2
    punpcklwd  xmm0,xmm2
    punpcklwd  xmm1,xmm2
    psubd      xmm0,xmm1

    cvtdq2ps   xmm0,xmm0

    mov        eax,[edit_vertex_no]
    dec        eax
    shl        eax,2
    lea        eax,[eax*3]
    add        eax,[points_rotated_ptr]
    movlps     xmm1,[eax]

    addps      xmm1,xmm0
    movlps     [eax],xmm1


  @@:     ; face temporaly edition routines
    mov        eax,dword[edit_face_no]
    or         eax,eax          ; only 1 index is checked !
    jz         @f

    movd       xmm0,dword[edit_end_x]
    movd       xmm1,dword[edit_start_x]
    pxor       xmm2,xmm2
    punpcklwd  xmm0,xmm2
    punpcklwd  xmm1,xmm2
    psubd      xmm0,xmm1
    cvtdq2ps   xmm0,xmm0
    mov        eax,dword[edit_face_no]
    shl        eax,2
    lea        eax,[eax*3]
    add        eax,[triangles_ptr]


    mov        ebx,dword[eax+4]
    mov        ecx,dword[eax+8]
    mov        eax,dword[eax]
    shl        eax,2
    shl        ebx,2
    shl        ecx,2
    lea        eax,[eax*3]
    lea        ebx,[ebx*3]
    lea        ecx,[ecx*3]
    add        eax,[points_rotated_ptr]
    add        ebx,[points_rotated_ptr]
    add        ecx,[points_rotated_ptr]




    movlps     xmm1,[eax]
    movlps     xmm2,[ebx]
    movlps     xmm3,[ecx]
    addps      xmm1,xmm0
    addps      xmm2,xmm0
    addps      xmm3,xmm0
    movlps     [eax],xmm1
    movlps     [ebx],xmm2
    movlps     [ecx],xmm3
  @@:
    cmp        dword[edit_edge_no],0
    je         @f
    movd       xmm0,dword[edit_end_x]
    movd       xmm1,dword[edit_start_x]
    pxor       xmm2,xmm2
    punpcklwd  xmm0,xmm2
    punpcklwd  xmm1,xmm2
    psubd      xmm0,xmm1
    cvtdq2ps   xmm0,xmm0
    mov        eax,dword[edit_edge_no]
    dec        eax
    shl        eax,3
    add        eax,[edges_ptr]
    mov        ebx,[eax+4]
    mov        eax,[eax]

    shl        eax,2
    shl        ebx,2
    lea        eax,[eax*3]
    lea        ebx,[ebx*3]
    add        eax,[points_rotated_ptr]
    add        ebx,[points_rotated_ptr]

    movlps     xmm1,[eax]
    movlps     xmm2,[ebx]
    addps      xmm1,xmm0
    addps      xmm2,xmm0
    movlps     [eax],xmm1
    movlps     [ebx],xmm2

  @@:                     ; translate vertices into 2d
    movdqu   xmm3,[xxadd]
    cvtdq2ps xmm3,xmm3
    xor      ecx,ecx
    mov      ebx,[points_rotated_ptr]
    mov      eax,[points_count_var]
    sub      eax,[NextPointsCount]
  @@:
    movups   xmm1,[ebx]
    addps    xmm1,xmm3
    movlps   [ebx],xmm1
    movhlps  xmm1,xmm1
    movss    [ebx+8],xmm1
    add      ebx,12
    add      ecx,1
    cmp      ecx,eax  ;[points_count_var]
    jnz      @b
if  1

    mov      eax,[NextPointsCount]
    or       eax,eax
    jz       .end_add
    xor      ecx,ecx
    movq     xmm7,[NextMxadd]
    movq     xmm5,[NextMsub]
    pxor      xmm6,xmm6
    punpcklwd xmm7,xmm6
    punpcklwd xmm5,xmm6
    psubd     xmm7,xmm5
    pslld     xmm7,1
    cvtdq2ps  xmm7,xmm7
    addps     xmm3,xmm7
  @@:
    movups   xmm1,[ebx]
    addps    xmm1,xmm3
    movhlps  xmm2,xmm1
    movlps   [ebx],xmm1
    movss    [ebx+8],xmm2
    add      ebx,12
    add      ecx,1
    cmp      ecx,eax
    jnz      @b
   .end_add:
end if


     cmp    [draw_flag],20    ; count  transparent objects slices
     jne    .no_counter

    mov   eax,0x0          ; clear draw buffer
    mov   edi,[slices_counter_ptr]
    movzx ebx,[xres_var]
    movzx ecx,[yres_var]
    imul  ecx,ebx
    shr   ecx,2
    call  fill_aligned

    mov    esi,[triangles_ptr]
    mov    ecx,[triangles_count_var]
  .again_counter_tri:
    push  esi
    push  ecx
    mov   eax,dword[esi]
    mov   ebx,dword[esi+4]
    mov   ecx,dword[esi+8]
    shl   eax,2
    shl   ebx,2
    shl   ecx,2
    lea   eax,[eax*3]
    lea   ebx,[ebx*3]
    lea   ecx,[ecx*3]
    add   eax,[points_rotated_ptr]
    add   ebx,[points_rotated_ptr]
    add   ecx,[points_rotated_ptr]
    movlps xmm0,[eax]
    movlps xmm1,[ebx]
    movlps xmm2,[ecx]
    shufps xmm0,xmm0,11100001b
    shufps xmm1,xmm1,11100001b
    shufps xmm2,xmm2,11100001b

    cvtps2dq xmm0,xmm0
    cvtps2dq xmm1,xmm1
    cvtps2dq xmm2,xmm2
    packssdw xmm0,xmm0
    packssdw xmm1,xmm1
    packssdw xmm2,xmm2
    movd     eax,xmm0
    movd     ebx,xmm1
    movd     ecx,xmm2
    mov      edi,[slices_counter_ptr]


    call     slices_counter_tri
    pop      ecx
    pop      esi
    add      esi,12
    sub      ecx,1
    jnz      .again_counter_tri

    movzx    eax,[xres_var]
    movzx    ebx,[yres_var]
    mov      edi,[slices_counter_ptr]
    call     count_transparent_voxels
    mov      [.transparent_voxels_count],ecx
;    push     ecx
    shl      ecx,2
    invoke   VirtualAlloc,NULL,ecx,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
    or       eax,eax
    jnz      @f
   ; int3
    invoke MessageBox,NULL,_error_mem,NULL,MB_ICONERROR+MB_OK
  @@:
    mov      [voxels_depth_ptr],eax

; make background in slices history buff
    mov   eax,60000.1          ; clear draw buffer
    mov   edi,[voxels_depth_ptr]
 ;   pop   ecx
    mov   ecx,[.transparent_voxels_count]
    call  fill_aligned

; fill pointers buff
    mov      eax,[voxels_depth_ptr]
    mov      esi,[slices_counter_ptr]
    mov      edi,[slices_ptrs_buff_ptr]
    movzx    edx,[xres_var]
    movzx    ecx,[yres_var]
    imul     ecx,edx
    call     do_variable_slices_ptr_buff  ; file procs_b.inc

  .no_counter:


     cmp    [draw_flag],20
     jne    .no_peel

     invoke CreateThread,NULL,NULL,th_peel0,\
            NULL,NORMAL_PRIORITY_CLASS,\
            [ThreadID0]
     mov    [hthread0],eax
     invoke SetThreadPriority,hthread0,THREAD_PRIORITY_TIME_CRITICAL


     invoke CreateThread,NULL,NULL,th_peel1,\
            NULL,NORMAL_PRIORITY_CLASS,\
            [ThreadID1]
     mov    [hthread1],eax
     invoke SetThreadPriority,hthread1,THREAD_PRIORITY_TIME_CRITICAL



     invoke CreateThread,NULL,NULL,th_peel2,\
            NULL,NORMAL_PRIORITY_CLASS,\
            [ThreadID2]
     mov    [hthread2],eax
     invoke SetThreadPriority,hthread2,THREAD_PRIORITY_TIME_CRITICAL


     invoke CreateThread,NULL,NULL,th_peel3,\
            NULL,NORMAL_PRIORITY_CLASS,\
            [ThreadID3]
     mov    [hthread3],eax
     invoke SetThreadPriority,hthread3,THREAD_PRIORITY_TIME_CRITICAL

     invoke SetThreadAffinityMask,[hthread0],0 ;1
     invoke SetThreadAffinityMask,[hthread1],0 ;2
     invoke SetThreadAffinityMask,[hthread2],0 ;3
     invoke SetThreadAffinityMask,[hthread3],0 ;4

     invoke ResumeThread,[hthread0]
     invoke ResumeThread,[hthread1]
     invoke ResumeThread,[hthread2]
     invoke ResumeThread,[hthread3]

     invoke WaitForMultipleObjectsEx,4,hthread0,1,-1,0
     invoke CloseHandle,[hthread0]
     invoke CloseHandle,[hthread1]
     invoke CloseHandle,[hthread2]
     invoke CloseHandle,[hthread3]

   ;  call   th_peel0
   ;  call   th_peel1
   ;  call   th_peel2
   ;  call   th_peel3

     jmp    .no_stencil

 .no_peel:
     cmp    [draw_flag],21
     jge    .no_stencil
     cmp    [draw_flag],17    ; glass render
     jnge    .no_stencil
     mov    esi,[triangles_ptr]
     mov    ecx,[triangles_count_var]
  .again_stencil_tri:
    push  esi
    push  ecx
    mov   eax,dword[esi]
    mov   ebx,dword[esi+4]
    mov   ecx,dword[esi+8]
    shl   eax,2
    shl   ebx,2
    shl   ecx,2
    lea   eax,[eax*3]
    lea   ebx,[ebx*3]
    lea   ecx,[ecx*3]
    add   eax,[points_rotated_ptr]
    add   ebx,[points_rotated_ptr]
    add   ecx,[points_rotated_ptr]
    movss  xmm3,[eax+8]
    movss  xmm4,[ebx+8]
    movss  xmm5,[ecx+8]
    movlps xmm0,[eax]
    movlps xmm1,[ebx]
    movlps xmm2,[ecx]



  ;  movlhps xmm4,xmm4
    shufps  xmm4,xmm4,11110011b
    orps    xmm3,xmm4
;    shufps  xmm3,xmm3,11111000b
    movlhps xmm3,xmm5  ; xmm3 - z float pack
  ;  orps     xmm3,[zero_hgst_dd]
  ;  shufps  xmm3,xmm3,11000110b

    shufps xmm0,xmm0,11100001b
    shufps xmm1,xmm1,11100001b
    shufps xmm2,xmm2,11100001b

    cvtps2dq xmm0,xmm0
    cvtps2dq xmm1,xmm1
    cvtps2dq xmm2,xmm2
    packssdw xmm0,xmm0
    packssdw xmm1,xmm1
    packssdw xmm2,xmm2
    movd     eax,xmm0
    movd     ebx,xmm1
    movd     ecx,xmm2
    movaps   xmm0,xmm3
;    cmp      [draw_flag],20
;    jne      @f
;    movd     xmm5,[xres_vard]
;    punpcklwd xmm5,[the_zero]
;    pshufd    xmm5,xmm5,00110111b
;    mov      edi,[slices_ptrs_buff_ptr]
;    mov      esi,[slices_counter_ptr]
;    call     slices_tri_var
;    jmp      .next_stc
;  @@:
    cmp      [draw_flag],19
    je       .slic
    cmp      [draw_flag],15
    jne      @f
    mov      esi,[slices_ptrs_buff_ptr]
    call     stencil_tri
    jmp      .next_stc
  @@:
    mov      esi,[Zbuffer_ptr]
    call     stencil_tri
    jmp      .next_stc
  .slic:
    mov      edi,[Zbuffer_ptr]
    call     slices_tri
  .next_stc:
    pop      ecx
    pop      esi
    add      esi,12
    sub      ecx,1
    jnz      .again_stencil_tri

  .no_stencil:


;    cmp [draw_flag],14
;    jne @f
;    cmp [vert_flag],1
;    je  .vert_from_vert_lst  ; coordinates no from triangles list
;                             ; but from vertices list
;  @@:
   ; jmp  .skip_th   ;dbg
    cmp [draw_flag],21
    je  .skip_th
    cmp [draw_flag],20
    je  .skip_dr
    cmp [draw_flag],15
    jge @f
    cmp [draw_flag],9
    jge .skip_th
;    cmp [draw_flag],16

   @@:

    invoke CreateThread,NULL,NULL,draw_th0,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID0]
    mov    [hthread0],eax
    invoke  SetThreadPriority,hthread0,THREAD_PRIORITY_TIME_CRITICAL

    invoke CreateThread,NULL,NULL,draw_th1,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID1]
    mov  [hthread1],eax
    invoke  SetThreadPriority,hthread1,THREAD_PRIORITY_TIME_CRITICAL
    cmp [draw_flag],15   ; phong shading
    jl  @f
if 1
    invoke CreateThread,NULL,NULL,draw_th2,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID2]
    mov  [hthread2],eax
    invoke  SetThreadPriority,hthread2,THREAD_PRIORITY_TIME_CRITICAL

;    mov  eax,draw_th3

    invoke CreateThread,NULL,NULL,draw_th3,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID3]
    mov  [hthread3],eax
    invoke  SetThreadPriority,hthread3,THREAD_PRIORITY_TIME_CRITICAL
end if
;        invoke CreateThread,NULL,NULL,draw_th3,\
;           NULL,NORMAL_PRIORITY_CLASS,\
;           [ThreadID4]
;    mov  [hthread4],eax
;    invoke  SetThreadPriority,hthread4,THREAD_PRIORITY_TIME_CRITICAL
  @@:

    invoke SetThreadAffinityMask,[hthread0],0 ;1
    invoke SetThreadAffinityMask,[hthread1],0 ;2
    cmp [draw_flag],15   ; phong shading
    jl  @f
if 1
    invoke SetThreadAffinityMask,[hthread2],0 ;3
    invoke SetThreadAffinityMask,[hthread3],0 ;4
;    invoke SetThreadAffinityMask,[hthread4],5
end if
  @@:
    invoke ResumeThread,[hthread0]
    invoke ResumeThread,[hthread1]
    cmp [draw_flag],15   ; phong shading
    jl  @f
if 1
    invoke ResumeThread,[hthread2]
    invoke ResumeThread,[hthread3]
end if
;    invoke ResumeThread,[hthread4]
  @@:
    cmp [draw_flag],15   ; phong shading
    jl  @f
if 1
    invoke WaitForMultipleObjectsEx,4,hthread0,1,-1,0  ;TRUE,0
    jmp    .close_h
end if
   @@:
    invoke  WaitForMultipleObjectsEx,2,hthread0,1,-1,0
  .close_h:
    invoke CloseHandle,[hthread0]
    invoke CloseHandle,[hthread1]
    cmp [draw_flag],15   ; phong shading
    jl  @f
if 1
    invoke CloseHandle,[hthread2]
    invoke CloseHandle,[hthread3]
end if
   @@:
    jmp  .skip_dr
  .skip_th:
    mov  eax,1
    call draw_thread




;    jmp  .skip_dr
  ;  mov    ecx,[trans_voxels_count]

;  .vert_from_vert_lst:

  .skip_dr:
    cmp    [draw_flag],20
    jne    @f
    invoke  VirtualFree, [voxels_depth_ptr] , 0,  MEM_RELEASE
  @@:

    cmp    [normals_flag],1
    jne    @f
    call   draw_triangles_normals
  @@:
    cmp    [normals_flag],2
    jne    @f
    call   draw_vertices_normals

  @@:


    cmp    [bezier_flag],0
    je     .no_derives
    cmp    [derive_flag],1
    je     .no_derives

;in  eax - dword value to fill,
;    ecx - area size in dwords
;    edi - mem adress
;    addres must be aligned
    mov   edi,[edit_buf_ptr]
    movzx ecx,[xres_var]
    movzx eax,[yres_var]
    imul  ecx,eax
    xor   eax,eax
    call  fill_aligned

    mov    edi,[screen_ptr]
    mov    ebx,[derv_rect_ptr]
    xor    ecx,ecx

  @@:
    push   ecx
    push   ebx
    mov    esi,derv_nodes_rot
    call   draw_bezier_derives1  ; only for sigle patch
    pop    ebx
    add    ebx,16*2
    pop    ecx

    add    ecx,1
    cmp    cx, [bez_patch_count]
    jnz    @b

    mov    esi,derv_nodes_rot
    mov    ebx,[derv_rect_ptr]
    movzx  ecx,[bez_patch_count]
; in:
;  esi - ptr to derives vertices list
;  ebx - ptr to derives indexes (16 pro one rectangle)
;  ecx - number of surfaces
;  globals - screen_ptr, i12, ..

    call   draw_bezier_bars


  .no_derives:

    cmp   [set_tri_area],0  ; triangukize area/rect
    jne   @f                       ;
    call   draw_triangulize_rect
  @@:
    cmp  [hrt_flag],3      ; write frame around current segment
    jl   .no_rectangle
    cmp  [hrt_flag],6
    jge  .no_rectangle
  @@:
    cmp  [current_seg],255
    jz   .no_rectangle

    mov   esi,[points_rotated_ptr]
    movzx ebx,[current_seg]

    cmp   [hrt_flag],3
    jne   @f
    mov  eax,BSTEPS*BSTEPS*12
    jmp   .calc
  @@:
    cmp   [hrt_flag],4
    jne   @f
    mov   eax,[rot_pipe_points_number]
    shl   eax,2
    lea   eax,[eax*3]
    jmp   .calc
  @@:
    cmp   ebx,BPIPES ;-1
    jle   .only_pipes
    mov   eax,[rot_pipe_points_number]
    imul  eax,BPIPES*12
    sub   ebx,BPIPES
    imul  ebx,BSTEPS*BSTEPS*12
    add    eax,ebx
    jmp   .skip_mul

  .only_pipes:
    mov   eax,[rot_pipe_points_number]
    shl   eax,2
    lea   eax,[eax*3]


 .calc:
    mul   ebx
 .skip_mul:
   add   esi,eax
   movups xmm0,[esi]
   movups xmm1,[esi]
   add    esi,12
   cmp    [hrt_flag],3
   jg     @f
   mov    ecx,(BSTEPS*(BSTEPS))-1
   jmp    .find
 @@:
   cmp    [hrt_flag],4
   jg     @f
   mov    ecx,[rot_pipe_points_number]
   dec    ecx
   jmp    .find
 @@:
   cmp    [current_seg],BPIPES
   jge    @f
   mov    ecx,[rot_pipe_points_number]
   dec    ecx
   jmp    .find
 @@:
   mov    ecx,(BSTEPS*(BSTEPS))-1
 .find:
 @@:
   movups xmm2,[esi]
   maxps  xmm0,xmm2
   minps  xmm1,xmm2
   add    esi,12
   sub    ecx,1
   jnz    @b
 ;  xmm0 - max
 ;  xmm1 - min

   movups   xmm2,[max_clip_seg]
   movups   xmm3,[min_clip_seg]
   minps    xmm0,xmm2
   maxps    xmm1,xmm3
   cvtps2dq xmm0,xmm0
   cvtps2dq xmm1,xmm1
   movups   [cur_seg_max_rot],xmm0
   movups   [cur_seg_max_rot+12],xmm1



    ; max min coordinate should be normalized like whole object
    ; (witch is before rotating, rescaling and translating) in <-1.0 ; 1.0>


    .xsegmax equ [cur_seg_max_rot]
    .ysegmax equ [cur_seg_max_rot+4]
    .xsegmin equ [cur_seg_max_rot+12]
    .ysegmin equ [cur_seg_max_rot+16]

;    movzx     eax,[xres_var]
;    mov       [screen_width_var],eax ;,XRES




;    mov      eax,[cur_seg_max_rot+12]   ; xmin
;    mov      ebx,[cur_seg_max_rot+16]   ; ymin
;    mov      edx,ebx                    ; ymin
;    mov      ecx,[cur_seg_max_rot]      ; xmax

    mov      ax,.xsegmin
    shl      eax,16
    mov      ax,.xsegmax
    mov      bx,.ysegmin
    shl      ebx,16
    mov      bx,.ysegmax



     movd     mm1,[screen_ptr]
  ;  movd     mm0,[Zbuffer_ptr]
     movzx    esi,[xres_var]
     movd     mm2,esi
     mov      esi,0x0000ff00
     movd     mm3,esi




    call     write_frame


 @@:
   .no_rectangle:
    popad

   ; invoke      MessageBox,[hwnd],STRdata,_tsc_title,MB_RTLREADING
;   invoke stop_ticks
ret
endp

normalize_all_light_vectors:
        mov     edi,lights
     @@:
        call    normalize_vector           ;       3dmath.inc
        add     edi,LIGHT_SIZE
        cmp     edi,lightsend
        jl      @b
ret
read_hrt:
        fninit
        mov    esi,hrt_triangles
        mov    edi,[triangles_ptr]
        mov    [triangles_count_var],hrt_triangles_count
        mov    ecx,hrt_triangles_count
        cld
        lea    ecx,[ecx*3]
     @@:
        movzx  eax,word[esi]
        mov    [edi],eax
        add    esi,2
        add    edi,4
        loop   @b
        movsd


        mov    esi,hrt_points
        mov    edi,[points_r_ptr]
        mov    ecx,hrt_points_count
        mov    [points_count_var],0
      @@:
        fild   word[esi]
        fstp   dword[edi]
        fild   word[esi+2]
        fstp   dword[edi+4]
        fild   word[esi+4]
        fstp   dword[edi+8]
        inc    [points_count_var]
        add    edi,12
        add    esi,6
        dec    ecx
        jne    @b

ret

alloc_mem_for_tp1:
;; in eax=1 - full alloc, else alloc only for triangles(faces)
;; and points(vertices);
if 0 ;abandon text below
;; and in this cause (eax<>1) - in: eax -> points (vertices) count
;;                                  ebx -> triangles count
;;                              out: eax -> point (vertices) ptr
;;                                   ebx -> triangles ptr
end if ; abandon text above
      ; cmp triangulize cause
     cmp     eax ,1
     jne     @f
     call    free_mem_for_tp1
     jmp     .no_tp
  @@:


     mov     eax,[points_count_var]
     imul    eax,[i12]
     add     eax,20
     invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     or  eax,eax
     jnz @f
   ;  int3
     invoke MessageBox,NULL,_error_mem,NULL,MB_ICONERROR+MB_OK
   @@:
     mov     [points_r_ptr],eax
if 0
     movzx   eax,[points_count_var]
     imul    eax,[i12]
     add     eax,65536
     invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     or  eax,eax
     jnz @f
   ;  int3
     invoke MessageBox,NULL,_error_mem,NULL,MB_ICONERROR+MB_OK
   @@:
     mov     [shadow_vertices_ptr],eax

end if

     mov     eax,[triangles_count_var]
     imul    eax,[i12]
     add     eax,65536
     invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     or  eax,eax
     jnz @f
     ;int3
      invoke MessageBox,NULL,_error_mem,NULL,MB_ICONERROR+MB_OK
   @@:
     mov     [triangles_ptr],eax

;     movzx   eax,[triangles_count_var]
;     imul    eax,[i12]
;     invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
;     or  eax,eax
;     jnz @f
;     int3
;   @@:
;     mov     [edges_ptr],eax

 .no_tp:
          mov     eax,[triangles_count_var]
     shl     eax,3
     lea     eax,[eax*3]
     add     eax,65536
     invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     or  eax,eax
     jnz @f
     invoke MessageBox,NULL,_error_mem,NULL,MB_ICONERROR+MB_OK
   @@:
     mov     [edges_ptr],eax

     mov     eax,[points_count_var]
     imul    eax,[i12]
     add     eax,20
     invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     or  eax,eax
     jnz @f
     invoke MessageBox,NULL,_error_mem,NULL,MB_ICONERROR+MB_OK
  ;   int3
   @@:
     mov     [shadow_vertices_ptr],eax






     mov     eax,[triangles_count_var]
     imul    eax,[i12]
       add     eax,65536
     invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     or  eax,eax
     jnz @f
     invoke MessageBox,NULL,_error_mem,NULL,MB_ICONERROR+MB_OK
   ;  int3
   @@:
     mov     [triangles_normals_ptr],eax


     mov     eax,[triangles_count_var]
     imul    eax,[i12]
       add     eax,65536
     invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     or  eax,eax
     jnz @f
     invoke MessageBox,NULL,_error_mem,NULL,MB_ICONERROR+MB_OK
   ;  int3
   @@:
     mov     [triangles_normals_rotated_ptr],eax





     mov     eax,[points_count_var]
     imul    eax,[i12]
     add     eax,65536
     invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     or  eax,eax
     jnz @f
     invoke MessageBox,NULL,_error_mem,NULL,MB_ICONERROR+MB_OK
  ;   int3
   @@:
     mov     [points_normals_ptr],eax

     mov     eax,[points_count_var]
     imul    eax,[i12]
       add     eax,65536
     invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     or  eax,eax
     jnz @f
     invoke MessageBox,NULL,_error_mem,NULL,MB_ICONERROR+MB_OK
   ;  int3
   @@:
     mov     [points_normals_rotated_ptr],eax

     mov     eax,[points_count_var]
     imul    eax,[i12]
     add     eax,20
     invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     or  eax,eax
     jnz @f
     invoke MessageBox,NULL,_error_mem,NULL,MB_ICONERROR+MB_OK
   ;  int3
   @@:
     mov     [points_rotated_ptr],eax

     mov     eax,[points_count_var]
     shl     eax,2
     add     eax,20
     invoke  VirtualAlloc,NULL,eax,MEM_COMMIT+MEM_RESERVE,PAGE_EXECUTE_READWRITE
     or  eax,eax
     jnz @f
     invoke MessageBox,NULL,_error_mem,NULL,MB_ICONERROR+MB_OK
  ;   int3
   @@:
     mov     [tex_points_ptr],eax
ret
free_mem_for_tp1:
   ; check triangulize cause
   cmp     eax,1
   je      .no_tp


   mov     ebx,[points_r_ptr]
   invoke  VirtualFree, ebx , 0,  MEM_RELEASE
   or      eax,eax
   jnz     @f
;   int3
 @@:
if 0
 ;    movzx   edx,[points_count_var]
 ;    imul    edx,[i12]
 ;    add     edx,20
     mov     ebx,[shadow_vertices_ptr]
     invoke  VirtualFree, ebx , 0,  MEM_RELEASE
     or  eax,eax
     jnz @f
  ;   int3
   @@:
end if


   mov     ebx,[triangles_ptr]
   invoke  VirtualFree, ebx , 0,  MEM_RELEASE
   or      eax,eax
   jnz     @f
   int3
 @@:


 .no_tp:
      mov     ebx,[edges_ptr]
   invoke  VirtualFree, ebx , 0,  MEM_RELEASE
;   or      eax,eax
;   jnz     @f

   mov     ebx,[shadow_vertices_ptr]
   invoke  VirtualFree, ebx , 0,  MEM_RELEASE
   or  eax,eax
;   jnz @f
;   int3
 @@:

   mov     ebx,[triangles_normals_ptr]
   mov     eax,[triangles_count_var]
   imul    eax,[i12]
   invoke  VirtualFree, ebx , 0,  MEM_RELEASE
   or      eax,eax
   jnz     @f
;   int3
 @@:

   mov     ebx,[triangles_normals_rotated_ptr]
   invoke  VirtualFree, ebx , 0,  MEM_RELEASE
   or      eax,eax
   jnz     @f
;   int3
 @@:



   mov     ebx,[points_normals_ptr]
   invoke  VirtualFree, ebx , 0,  MEM_RELEASE
   or      eax,eax
   jnz     @f
 ;  int3
 @@:

   mov     ebx,[points_normals_rotated_ptr]
   invoke  VirtualFree, ebx , 0,  MEM_RELEASE
   or      eax,eax
   jnz     @f
 ;  int3
 @@:

   mov     ebx,[points_rotated_ptr]
   invoke  VirtualFree, ebx , 0,  MEM_RELEASE
   or      eax,eax
   jnz     @f
;   int3
 @@:

   mov     ebx,[tex_points_ptr]
   invoke  VirtualFree, ebx , 0,  MEM_RELEASE
   or      eax,eax
   jnz     @f
;   int3
 @@:

ret
;in:
read_tp_variables1:            ; read [triangles_count_var] and ...
                              ; ..[points_count_var]
; in:   eax -> ptr to 3ds file
;       ebx -> file size
; out:  ecx -> points count
;       ebx -> triangles count
        push ebp
        mov  ebp,esp
        sub  esp,64
        .tri_count  equ [ebp-4]
        .vert_count equ [ebp-8]
        .FileSize   equ [ebp-12]
        .EndFile    equ [ebp-16]

        mov     .FileSize,ebx
        xor     ebx,ebx
        mov     .vert_count,ebx
        mov     .tri_count,ebx
        mov     esi,eax  ;[File_ptr]


        cmp     [esi],word 4D4Dh
        je      @f ;Must be legal .3DS file
        xor     eax,eax
        ret
    @@:
        mov     eax,dword[esi+2]
        cmp     eax,.FileSize ;This must tell the length
        je      @f
        xor     eax,eax
        ret
     @@:
        add     eax,esi
        mov     .EndFile,eax    ;
        xor     eax,eax
        add     esi,6
      @@:
        cmp     [esi],word 3D3Dh
        je      @f
        add     esi,[esi+2]
        jmp     @b
      @@:
        add     esi,6
      .find4k:
        cmp     [esi],word 4000h
        je      @f
        add     esi,[esi+2]
        cmp     esi,.EndFile
        jc      .find4k
        jmp     .exit
      @@:
        add     esi,6
      @@:
        cmp     [esi],byte 0
        je      @f
        add     esi,1
        jmp     @b
      @@:
        add     esi,1
      @@:
        cmp     [esi],word 4100h
        je      @f
        add     esi,[esi+2]
        jmp     @b
      @@:
        add     esi,6
      @@:
        cmp     [esi],word 4110h
        je      @f
        add     esi,[esi+2]
        jmp     @b
      @@:
        movzx   ecx,word[esi+6]
        add     .vert_count,ecx
        mov     edx,ecx
        add     esi,8
        lea     ecx,[ecx*3]
        add     ecx,ecx
        add     ebx,ecx
        lea     esi,[esi+ecx+ecx]
      @@:
        cmp     [esi],word 4120h
        je      @f
        add     esi,[esi+2]
        jmp     @b
      @@:
        movzx   ecx,word[esi+6]
        add     .tri_count,ecx
        add     esi,8

        shl     ecx,3
        add     esi,ecx
        add     eax,edx
        jmp     .find4k
        mov     eax,-1 ;<---mark if OK
     .exit:
        mov     ecx,dword .vert_count
        mov     ebx,dword .tri_count

        mov   esp,ebp
        pop   ebp
ret


; read from 3ds file
;in:  eax - ptr to triangles
;     ebx - ptr to vertices
;     ecx - ptr to file begin
;     edx - ptr to file end
;out: ecx - points (vertices) count
;     edx - triangles count
read_from_file:
;; this procedure coded orginally by Madis Kalme
;; some changes by me
;; Thanks a lot Madis!!!
        push    ebp
        mov     ebp,esp
        sub     esp,40

        .triangles_p equ [ebp-4]
        .vert_p      equ [ebp-8]
        .file        equ [ebp-12]
        .file_end    equ [ebp-16]
        .vert_cnt    equ [ebp-20]
        .tri_cnt     equ [ebp-24]

        mov     .triangles_p,eax
        mov     .vert_p,ebx
        mov     .file,ecx
        mov     edi,eax  ;[triangles_ptr]
        xor     ebx,ebx
        mov     dword .vert_cnt,0
        mov     dword .tri_cnt,0

        mov     esi,ecx

        cmp     [esi],word 4D4Dh
        jne     .exit ;Must be legal .3DS file
;        cmp     dword[esi+2],EndFile-SourceFile
;        jne     .exit ;This must tell the length
        mov     eax,dword[esi+2]
  ;      cmp     eax,[fsize]
  ;      jne     .exit

        add     eax,esi
        mov     .file_end,eax    ;
        xor     eax,eax
        add     esi,6
      @@:
        cmp     [esi],word 3D3Dh
        je      @f
        add     esi,[esi+2]
        jmp     @b
      @@:
        add     esi,6
      .find4k:
        cmp     [esi],word 4000h
        je      @f
        add     esi,[esi+2]
        cmp     esi,.file_end
        jc      .find4k
        jmp     .exit
      @@:
        add     esi,6
      @@:
        cmp     [esi],byte 0
        je      @f
        add     esi,1
        jmp     @b
      @@:
        add     esi,1
      @@:
        cmp     [esi],word 4100h
        je      @f
        add     esi,[esi+2]
        jmp     @b
      @@:
        add     esi,6
      @@:
        cmp     [esi],word 4110h
        je      @f
        add     esi,[esi+2]
        jmp     @b
      @@:
        movzx   ecx,word[esi+6]       ; word
        add     .vert_cnt,ecx

        mov     edx,ecx
        add     esi,8
        push    edi
        mov     edi,.vert_p
if 1

      @@:
        cmp     ecx,10     ; 120 bytes to transfer
        jle     .one
        movups  xmm0,[esi]
        movups  xmm1,[esi+16]
        movups  xmm2,[esi+32]
        movups  xmm3,[esi+48]
        movups  xmm4,[esi+64]
        movups  xmm5,[esi+80]
        movups  xmm6,[esi+96]
        movlps  xmm7,[esi+112]

        movups  [edi+ebx*2],xmm0
        movups  [edi+ebx*2+16],xmm1
        movups  [edi+ebx*2+32],xmm2
        movups  [edi+ebx*2+48],xmm3
        movups  [edi+ebx*2+64],xmm4
        movups  [edi+ebx*2+80],xmm5
        movups  [edi+ebx*2+96],xmm6
        movlps  [edi+ebx*2+112],xmm7
        add     ebx,60
        add     esi,120
        sub     ecx,10
        jmp     @b
     .one:

        or      ecx,ecx
        jz      .done
end if
     @@:
        movups   xmm0,[esi]
        movhlps  xmm1,xmm0
        movlps   [edi+ebx*2],xmm0
        movss    [edi+ebx*2+8],xmm1
        add     ebx,6
        add     esi,12
        sub     ecx,1
        jnz     @b
     .done:
        pop     edi

      @@:
        cmp     [esi],word 4120h
        je      @f
        add     esi,[esi+2]
        jmp     @b
      @@:
        movzx   ecx,word[esi+6]
        add     .tri_cnt,ecx
        add     esi,8
        movd    xmm1,eax
        pshufd  xmm1,xmm1,0
    if 1
     @@:
        cmp       ecx,4
        jle       .two
        movq      xmm0,[esi]
        movq      xmm2,[esi+8]
        movq      xmm3,[esi+16]
        movq      xmm4,[esi+24]
        punpcklwd xmm0,[the_zero]
        punpcklwd xmm2,[the_zero]
        punpcklwd xmm3,[the_zero]
        punpcklwd xmm4,[the_zero]
        paddd     xmm0,xmm1
        paddd     xmm2,xmm1
        paddd     xmm3,xmm1
        paddd     xmm4,xmm1
        movdqu    [edi],xmm0
        movdqu    [edi+12],xmm2
        movdqu    [edi+24],xmm3
        movdqu    [edi+36],xmm4
        add       edi,48
        add       esi,32
        sub       ecx,4
        jmp       @b
       .two:
  end if
       @@:
        movq      xmm0,[esi]
        punpcklwd xmm0,[the_zero]
        paddd     xmm0,xmm1
        movq      [edi],xmm0
        pshufd    xmm0,xmm0,00000010b
        movd      [edi+8],xmm0
        add       edi,12
        add       esi,8
        sub       ecx,1
        jnz       @b


        add     eax,edx
        jmp     .find4k

        mov     eax,-1 ;<---mark if OK
      .exit:
        mov   ecx,dword .vert_cnt
        mov   edx,dword .tri_cnt
        mov   ebx,.file_end

        mov   esp,ebp
        pop   ebp
ret



init_tri_n_th:   ; init triangles normals
if 1    ;  roll off threads to easier debugging
    invoke CreateThread,NULL,NULL,init_tri_n_th0,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID5]
    mov    [hthread5],eax
    invoke SetThreadPriority,hthread5,THREAD_PRIORITY_TIME_CRITICAL

    invoke CreateThread,NULL,NULL,init_tri_n_th1,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID6]
    mov    [hthread6],eax
    invoke SetThreadPriority,hthread6,THREAD_PRIORITY_TIME_CRITICAL

    invoke CreateThread,NULL,NULL,init_tri_n_th2,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID7]
    mov    [hthread7],eax
    invoke SetThreadPriority,hthread7,THREAD_PRIORITY_TIME_CRITICAL

    invoke CreateThread,NULL,NULL,init_tri_n_th3,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID8]
    mov    [hthread8],eax
    invoke SetThreadPriority,hthread8,THREAD_PRIORITY_TIME_CRITICAL

    invoke SetThreadAffinityMask,[hthread5],[thread_array]
    invoke SetThreadAffinityMask,[hthread6],[thread_array+4]
    invoke SetThreadAffinityMask,[hthread7],[thread_array+8]
    invoke SetThreadAffinityMask,[hthread8],[thread_array+12]

    invoke ResumeThread,[hthread5]
    invoke ResumeThread,[hthread6]
    invoke ResumeThread,[hthread7]
    invoke ResumeThread,[hthread8]

    invoke WaitForMultipleObjectsEx,4,hthread5,1,-1,0
    invoke CloseHandle,[hthread5]
    invoke CloseHandle,[hthread6]
    invoke CloseHandle,[hthread7]
    invoke CloseHandle,[hthread8]
end if
ret

init_tri_n_th0:
  xor    eax,eax
  call   init_triangles_normals2
  invoke SendMessage,hwnd,WM_FINISH,0,0
  invoke ExitThread,0
ret

init_tri_n_th1:
  mov    eax,1
  call   init_triangles_normals2
  invoke SendMessage,hwnd,WM_FINISH,0,0
  invoke ExitThread,0 ;1
ret

init_tri_n_th2:
  mov    eax,2
  call   init_triangles_normals2
  invoke SendMessage,hwnd,WM_FINISH,0,0
  invoke ExitThread,0 ;2
ret

init_tri_n_th3:
  mov    eax,3
  call   init_triangles_normals2
  invoke SendMessage,hwnd,WM_FINISH,0,0
  invoke ExitThread,0 ;3
ret


init_triangles_normals2: ; orginally by Mikolaj Feliks
                         ; some changes by macgub
        fninit
        push     ebp
        mov      ebp,esp
        sub      esp,50
        and      ebp,0xfffffff0
        .vectors equ [ebp-32]

        mov     ecx,[triangles_count_var]
        mov     ebx,[triangles_normals_ptr]
        cmp     eax,100
        jne     @f
        mov     eax,[triangles_ptr]
        jmp     .do
     @@:
        mov     edi,[triangles_count_var]
        and     edi,011b
        shr     ecx,2

        mov     edx,ecx
        imul    edx,eax
        imul    edx,[i12]
        mov     eax,[triangles_ptr]
        add     ebx,edx
        add     eax,edx

        cmp     eax,2
        jng     .do
        add     ecx,edi

     .do:
     @@:
        push    ecx
        push    ebx
        lea     ebx,.vectors
        mov     esi,[eax]          ; first point index
        lea     esi,[esi*3]
        ; esi - pointer to 1st 3d point
        shl     esi,2
        add     esi,[points_r_ptr]
        mov     edi,[eax+4]          ; first point index
        lea     edi,[edi*3]
        shl     edi,2
        add     edi,[points_r_ptr]
        ; second point index
        ; edi - pointer to 2nd 3d point

        call    make_vector_r
        add     ebx,12
        mov     esi,edi
        mov     edi,[eax+8]        ; third point index
        lea     edi,[edi*3]
        shl     edi,2
        add     edi,[points_r_ptr]
        call    make_vector_r
        mov     edi,ebx                 ; edi - pointer to 2nd vector
        mov     esi,ebx
        sub     esi,12                  ; esi - pointer to 1st vector
        pop     ebx
        call    cross_product
        mov     edi,ebx
        call    normalize_vector
        add     eax,12
        add     ebx,12
        pop     ecx
        sub     ecx,1
        jnz     @b

        add     esp,50
        pop     ebp
ret
init_pt_norm_th:
 if 1    ;  roll off threads to easier debugging
    invoke CreateThread,NULL,NULL,init_pt_norm_th0,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID5]
    mov    [hthread5],eax
    invoke SetThreadPriority,hthread5,THREAD_PRIORITY_TIME_CRITICAL

    invoke CreateThread,NULL,NULL,init_pt_norm_th1,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID6]
    mov    [hthread6],eax
    invoke SetThreadPriority,hthread6,THREAD_PRIORITY_TIME_CRITICAL

    invoke CreateThread,NULL,NULL,init_pt_norm_th2,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID7]
    mov    [hthread7],eax
    invoke SetThreadPriority,hthread7,THREAD_PRIORITY_TIME_CRITICAL

    invoke CreateThread,NULL,NULL,init_pt_norm_th3,\
           NULL,NORMAL_PRIORITY_CLASS,\
           [ThreadID8]
    mov    [hthread8],eax
    invoke SetThreadPriority,hthread8,THREAD_PRIORITY_TIME_CRITICAL

    invoke SetThreadAffinityMask,[hthread5],[thread_array]
    invoke SetThreadAffinityMask,[hthread6],[thread_array+4]
    invoke SetThreadAffinityMask,[hthread7],[thread_array+8]
    invoke SetThreadAffinityMask,[hthread8],[thread_array+12]

    invoke ResumeThread,[hthread5]
    invoke ResumeThread,[hthread6]
    invoke ResumeThread,[hthread7]
    invoke ResumeThread,[hthread8]

    invoke WaitForMultipleObjectsEx,4,hthread5,1,-1,0
    invoke CloseHandle,[hthread5]
    invoke CloseHandle,[hthread6]
    invoke CloseHandle,[hthread7]
    invoke CloseHandle,[hthread8]
end if
ret

init_pt_norm_th0:
    xor    eax,eax
    call   init_point_normals
    invoke SendMessage,hwnd,WM_FINISH,0,0
    invoke ExitThread,0
ret

init_pt_norm_th1:
    mov    eax,1
    call   init_point_normals
    invoke SendMessage,hwnd,WM_FINISH,0,0
    invoke ExitThread,0  ;1
ret

init_pt_norm_th2:
    mov    eax,2
    call   init_point_normals
    invoke SendMessage,hwnd,WM_FINISH,0,0
    invoke ExitThread,0  ;2
ret

init_pt_norm_th3:
    mov    eax,3
    call   init_point_normals
    invoke SendMessage,hwnd,WM_FINISH,0,0
    invoke ExitThread,0  ;3
ret



init_point_normals: ; orginally by Mikolaj Feliks
                    ; some changes by macgub
.z equ dword [ebp-8]
.y equ dword [ebp-12]
.x equ [ebp-16]
.point_number equ [ebp-32]
.hit_faces    equ dword [ebp-36]
.end_point    equ dword [ebp-40]


        push      ebp
        mov       ebp,esp
        sub       esp,70
        and       ebp,0xfffffff0
        cmp       eax,100
        jne       @f
        xor       edx,edx
        xor       esi,esi
        push      [points_count_var]
        pop       .end_point
        jmp       .do
      @@:
        mov       edx,[points_count_var]
        mov       edi,[points_count_var]
        and       edi,011b
        shr       edx,2
        mov       ecx,edx
        imul      edx,eax    ; edx  - first vert no
        add       ecx,edx    ; ecx  - end vert no

        cmp       eax,3
        jne       @f
        add       ecx,edi
      @@:
        mov       .end_point,ecx

     .do:
        movd      xmm0,edx
        pshufd    xmm0,xmm0,0
        movdqa    .point_number,xmm0

    .ipn_loop:
        mov       .hit_faces,0
        mov       dword .x,0
        mov       .y,0
        mov       .z,0
        mov       esi,[triangles_ptr]
        mov       ecx,[triangles_count_var]
        lea       ecx,[ecx*3]
    .ipn_check_vertices:
        cmp       ecx,32
        jle       @f
        mov       ebx,32
        movdqa    xmm0,[esi]
        movdqa    xmm1,[esi+16]
        movdqa    xmm2,[esi+32]
        movdqa    xmm3,[esi+48]
        movdqa    xmm4,[esi+64]
        movdqa    xmm5,[esi+80]
        movdqa    xmm6,[esi+96]
        movdqa    xmm7,[esi+112]
        pcmpeqd   xmm0,.point_number
        pcmpeqd   xmm1,.point_number
        pcmpeqd   xmm2,.point_number
        pcmpeqd   xmm3,.point_number
        pcmpeqd   xmm4,.point_number
        pcmpeqd   xmm5,.point_number
        pcmpeqd   xmm6,.point_number
        pcmpeqd   xmm7,.point_number
        por       xmm0,xmm1
        por       xmm2,xmm3
        por       xmm4,xmm5
        por       xmm6,xmm7
        por       xmm0,xmm2
        por       xmm4,xmm6
        por       xmm0,xmm4
        movmskps  eax,xmm0
        or        eax,eax
        jnz       @f
        add       esi,128
        sub       ecx,32
        jmp       .ipn_check_vertices
      @@:
      .check_one_vertex:
        mov       eax,dword[esi]    ;  eax - point_number
        cmp       eax,.point_number
        je        .inc_h_f ;   .ipn_next_vertex
      .inc_counters:
        add       esi,4
        sub       ecx,1
        jz        .calc_one_normal
        sub       ebx,1
        jnz       .check_one_vertex
        jmp       .ipn_check_vertices
      .inc_h_f:
        push      eax
        push      edx

        mov       eax,esi
        sub       eax,[triangles_ptr]
        cdq
        mul       [irecipr3]  ; div eax by 3
        and       edx,0xfffffffc
        lea       edx,[edx*3]
        add       edx,[triangles_normals_ptr]
        movups    xmm0,[edx]
        addps     xmm0,.x
        movaps    .x,xmm0

        pop      edx
        pop      eax
        add       .hit_faces,1
        jmp       .inc_counters

     .calc_one_normal:
        mov       edi,.point_number
        imul      edi,[i12]
        add       edi,[points_normals_ptr]


        cvtsi2ss   xmm1,.hit_faces
        movaps     xmm0,.x
        rcpss      xmm1,xmm1
        shufps     xmm1,xmm1,0
        mulps      xmm0,xmm1
        movhlps    xmm2,xmm0
        movlps     [edi],xmm0
        movss      [edi+8],xmm2

        call      normalize_vector
        movdqa    xmm0,[i1x4]
        paddd     xmm0,.point_number
        movdqa    .point_number,xmm0
        movd      edx,xmm0
        cmp       edx,.end_point
        jne       .ipn_loop

        add       esp,70
        pop       ebp

ret


draw_menu:

    mov    edi,menu_screen
    mov    ecx,MXRES*MYRES
    xor    eax,eax
    call   fill_aligned


    mov    esi,menu_data
  .lab:
    push  esi

    mov   ax,[esi+2]
    shl   eax,16
    mov   ax,[esi+6]
    mov   bx,[esi+4]
    shl   ebx,16
    mov   bx,[esi+8]
    push  esi

    emms
    mov    esi,MXRES
    movd   mm2,esi
    mov    esi,menu_screen
    movd   mm1,esi
    mov    esi,0x0000ff00
    movd   mm3,esi



    call  write_frame
    pop   esi



    mov   eax,[esi+2]      ; write button desc
    ror   eax,16
    add   eax,0x00030000
    mov   ebx,font_table + 62
    mov   ecx,MXRES shl 16 + MYRES
    add   esi,10
    mov   edi,menu_screen
    pushad
    call  write_text
    popad

    add   eax,90 shl 16 + 0      ; write flag of button
    cmp   dword[esi+13],-1
    je    @f
    movzx edi,byte[esi+12]
    imul  edi,[i5]
    mov   esi,[esi+13]
    add   esi,edi
    mov   edi,menu_screen
    call  write_text
   @@:

    pop   esi
    add   esi, 33
    cmp   word[esi],-1
    jne   .lab


    movzx    eax,[whole_xres]
    mov      ecx,eax
    sub      eax,MXRES
    invoke BitBlt,[hdc],eax,0,MXRES,MYRES,menu_screen,Mbmi,0,SRCCOPY
  ;  invoke StretchDIBits,[hdc],eax,0,MXRES,MYRES,0,0,MXRES,MYRES,menu_screen,Mbmi,0,SRCCOPY
ret

random_point_light_position:

  rdtsc
  mov      [rand_seed],ax
  xor      ebx,ebx
 @@:
  push     ebx
  mov      ecx, -1000
  mov      edx, 1000
  call     random
  cvtsi2ss xmm0,eax
  pop      ebx
  movss    [point_light_coords+ebx],xmm0
  add      ebx,4
  cmp      ebx,8
  jnz      @b

  xor      ebx,ebx
  mov      eax,255
  cvtsi2ss xmm1,[i255d]  ;eax
 @@:
  push     ebx
  mov      ecx, 0
  mov      edx, 255
  call     random
  cvtsi2ss xmm0,eax
  divss    xmm0,xmm1
  pop      ebx
  movss    [plane_equation+ebx],xmm0
  add      ebx,4
  cmp      ebx,8
  jnz      @b
ret

include 'grd_cat.inc'
include 'tex_cat.inc'
include 'flat_cat.inc'
if Ext >= AVX
  include 'flat_cat_floatzavx.inc'
else
  include 'flat_cat_floatz.inc'
end if
include '3dmath.inc'
include 'procs_a.inc'
include 'procs_b.inc'
include 'procs_c.inc'
include 'bump_cat.inc'
include 'bump_tex.inc'
include 'two_tex.inc'
include 'grd_tex.inc'
include '3r_phg_col.inc'
include '3r_phg_col_stencil.inc'
include '3bi_fil.inc'
;include '3bi_filavx.inc'
include 'grd_line.inc'
include 'tex_line.inc'
include 'line.inc'
include  'shd3.inc'

include 'b_curve.inc'
;include 'Bline.inc'
include 'arbitrary_vector.inc'
include 'rotated_bezier.inc'
;include 'b_surface.inc'
include 'asc.inc'
include 'asc_header.inc'
include 'menu.inc'
include 'edit.inc'
include 'draw_thrd.inc'
;include 'bssurface.inc'
include 'bezier_surface.inc'
include 'stencil3.inc'
include '3glass.inc'
include '3glass_tex.inc'
include 'long_pipe.inc'
include '3glass_tex_slices.inc'
include '3slices_counter.inc'
include '3stencil_slices.inc'
include '3stencil_slices_var.inc'
include '3glass_tex_slices_var.inc'
;include 'line_grd_trans.inc'
;include 'ray_shadow.inc'
include 'normals.inc'


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; data
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
section '.data' data readable writeable
    _class     TCHAR     'Win3ds',0
    _title     TCHAR     'Ave Cruce Salus Mea'
  if Ext = SSE2
     .exxt     TCHAR     ' SSE2'
  else if Ext = SSE3
     .exxt     TCHAR     ' SSE3'
  else if Ext = SSE4.1
     .exxt     TCHAR     ' SSE4.1'
  else if Ext = AVX
     .exxt     TCHAR     ' AVX'
  end if
  fakebitmap   dd  0,0,0,0
    .end TCHAR 0
    _error     TCHAR     'Startup failed.',0
    _error_mem TCHAR     'Memory fail.',0
    _tsc_title TCHAR     'Clocks number',0
    sh_model   TCHAR     'Shadow + spot light                 '
        .fl    TCHAR     'Flat Shading                        '
        .gr    TCHAR     'Gouraud Shading                     '
        .sp    TCHAR     'Spherical Environment Mapping       '
        .bm    TCHAR     'Bump Mapping                        '
        .tx    TCHAR     'Texture Maping                      '
        .grtx  TCHAR     'Gouraud Shading and Texture Mapping '
        .dtxx  TCHAR     'Double Texture Mapping              '
        .bmtx  TCHAR     'Bump and Texture Mapping Pararell   '
        .grl   TCHAR     'Gouraud Shaded Line                 '
        .flf   TCHAR     'Flat Shading with Float zBuffer     '
        .txl   TCHAR     'Textured Line - Env. Mapping        '
        .tx12  TCHAR     'Textured Line                       '
        .line  TCHAR     'Simple line without z coord. interp.'
        .shad  TCHAR     'Nodes of object                     '
        .ph    TCHAR     'Real Phongs Shading                 '
        .btp   TCHAR     'Tex. with bilinear filtering + Phong'
        .gls   TCHAR     'Glass like rendering                '
        .gls_t TCHAR     'Glass + texture mapping rendering   '
        .wtr   TCHAR     'Weighted transparency               '
        .dp    TCHAR     'Depth peeling transparency          '
    File_ptr     dd ?
    File_tex_ptr dd ?
    EndFile      dd ?
    asc_file_name db 'object.asc'
    up          dd 0
    s_time      dd 0
    e_time      dd 0
    num_animations dd 3
    ccounter       dd 0
    temp_fur       dd 0
  ;  fur_dir        db 0
;    trans:  dd 1.0, 0.0

;    screen_width_var     dd ?  ; SIZE_X
    triangles_count_var  dd TRIANGLES_COUNT
    points_count_var     dd POINTS_COUNT
    edges_count          dd ?
;    NextMtriangles_count dw 0  ; next mesh variables
;    NextMpoints_count    dw 0
;    NextMscale           dd 1.0


    convert_muler:
    dd 1, 10, 100, 1000, 10000, 100000, 1000000, 10000000, 100000000
    XYZpartices:
    db 'X','Y','Z'



    peeling_particle dd 0.1
  ;  aprox     dd   5.1  ; glass it used
    tex_x_div2  dd  TEX_X / 2
    tex_y_div2  dd  TEX_Y / 2
    i3          dd  3
    i2          dd  2
    i5          dd  5
    i6          dd  6
    i9          dd  9
    i10         dw  10
    i12         dd  12
    i14         dd  14
    i15         dd  15
    i24         dd  24
    i27         dd  27
    i36         dd  36
    i38         dd  38
    i180        dw  180
    i256        dw  256
    i255d       dd  255
    f256d       dd  256.0
    irecipr3    dd  0x55555555
    f1          dd  1.0
    f3          dd  3.0
    recipr6     dd  0.1666666666666667
    f6          dd  6.0
    env_const   dd  1.05
    two_pi      dd  6.2831853
    dot_max:    dd  1.0 ;, 1.0, 1.0, 1.0
    dot_min:    dd  0.0 ;, 0, 0, 0
    correct_tex dw  255
    displac_div dd  5800 ;* 3 ; displacement divisior
    recip9000   dd  1 / 9000
    plane_equation:
     .A dd 0.5
     .B dd 0.5
     .C dd 1.0
     .D dd 0.0

    scale       dd 200.0
    scale_plus  dd 1.2
    scale_minus dd 0.8
    line_color  dd 0x0000ff00
    xxadd:
                dd   ?  ;XRES/2
    yyadd       dd   ?  ;YRES/2
    zzadd       dd 0
                dd 0
    NextMxadd:  dw BXRES/2  ; next mesh variables
    NextMyadd   dw BYRES/2
    NextMzadd   dw BYRES/2
                dw 0
    NextMsub    dq (BYRES/2) shl 32+ (BYRES/2) shl 16 + BXRES/2
    NextScale   dw BYRES/2
    NextScaleR  dd 1.0
 ;   NestScaleadd dd 1.07
 ;   NextScalesub dd 0.93
    NextMotion   dw 0
    NextPointsCount    dd 0
    NextTrianglesCount dd 0


    seg_rot_flag_button_down db 0
    tex_button_pres          db 0

    current_seg_add_flag     db 0

    piD128:
    one_deg     dd   0.024543692
    STRdata     db   '-1        '
    main_rotary :
               dd 0
               dd 0
               dd 0
    all_lights_size dw lightsend-lights
    align 16
    sli_cnt_const: times 16 db 1 ; slices counter constans
    point_light_coords:
              dd -500.0
              dd -500.0
              dd -900.0
              dd 0.0



    tspoon_nodes:
include 'tspoon_nodes.txt'
    tpot_nodes:
include 'tpot_nodes.txt'
    tcup_nodes:
include 'tcup_nodes.txt'

    tspoon_rect:
include 'tspoon_rect.txt'
    tpot_rect:
include 'tpot_rect.txt'
    tcup_rect:
include 'tcup_rect.txt'

tex_width  dd 1, 128
if 0
long_pipe_derv:
Bspline_derives:
  dd 1.0,0.0,0.0
  dd 2.0,0.0,0.0
  dd 3.0,0.0,0.0
  dd 4.0,0.0,0.0

  dd 1.0,1.0,0.2
  dd 1.0,2.0,0.3
  dd 1.0,3.0,0.4
  dd 1.0,4.0,-0.1

  dd 4.0,1.0,0.1
  dd 4.0,2.0,0.2
  dd 4.0,3.0,0.3
  dd 4.0,4.0,0.2

  dd 1.0,8.0,2.0
  dd 2.0,8.0,5.0
  dd 1.0,8.0,4.0
  dd 1.0,8.0,2.0

  dd 1.0,0.0,0.1
  dd 2.0,0.0,0.2
  dd 3.0,0.0,0.3
  dd 4.0,0.0,0.4

  dd 1.0,1.0,0.9
  dd 1.0,2.0,0.8
  dd 1.0,3.0,0.7
  dd 1.0,4.0,-0.6

  dd 4.0,1.0,0.3
  dd 4.0,2.0,0.6
  dd 4.0,3.0,0.3
  dd 4.0,4.0,0.7

  dd 1.0,8.0,2.0
  dd 2.0,8.0,5.0
  dd 1.0,8.0,4.0
  dd 1.0,8.0,2.0
end if
long_pipe_segs_c      dd 4
long_pipe_one_seg_c   dd 12  ;12 ; must be divisable by 4


  lights:

    light_vect      dd   0.0,0.0,-1.0
    light_c_o:                       ; +12
    p_light_o:
                    dw   51,71,11,0
    light_c_min:                     ; +20
    p_light_min:
                    dw   15,99,17,0
    light_c_max:
    p_light_max:
                    dw   255,255,5,0 ; +28
    shiness:
                    dw   5           ; +36
    LIGHT_SIZE = ($-lights)

        dd   -0.5,-0.5,-1.0
                                    ;   +12
        dw   1,255,1,0
                                    ;   +20
        dw   1,1,1,0
                                    ;   +28
        dw   255,255,255,0
        dw   5                      ;   +36



       dd   0.5,0.5,-1.0
                                   ;   +12
       dw   13,1,1,0
                                   ;   +20
       dw   70,30,70,0
                                   ;   +28
       dw   25,255,255,0
       dw   5                      ;   +36
   lightsend:
    align 16
  lights_aligned:






    .light_vect  dd   0.0, 0.0, -1.0, 0.0
    .light_c_o:                     ;   +16
                 dd   50.0, 49.0, 44.0, 0.0
    .light_c_min:                   ;   +32
                 dd   30.0, 41.0, 30.0, 0.0
    .light_c_max:                   ;   +48
                 dd   255.0, 255.0, 5.0, 1.0




       dd   -0.5,-0.5,-1.0, 0.0

       dd   32.0, 27.0, 17.0, 0.0

       dd   1.0,1.0,1.0,0.0

       dd   255.0,255.0,255.0,0.0







       dd   0.5,0.5,-1.0,0.0
                                   ;   +12
       dd   13.0,1.0,1.0,0.0
                                   ;   +20
       dd   7.0,3.0,7.0,0.0
                                   ;   +28
       dd   25.0, 255.0, 255.0, 0.0

  lights_aligned_end:

 align 16
     aprox:          dd 1.1, 1.1, 1.1, 1.1
     slices_factor1: dd 0.8, 0.8, 0.8, 0.8
     slices_factor2: dd 0.5, 0.5, 0.5, 0.5
     slices_factor3: dd 0.3, 0.3, 0.3, 0.3
     slices_factor4: dd 0.2, 0.2, 0.2, 0.2
     slices_factor5: dd 0.1, 0.1, 0.1, 0.1

    minimum0:
       dw   0,0,0,0,0,0,0,0
    maximum255:
       dw   255,255,255,255,255,255,255,255

    mask1:   dd -1, 0, 0, 0
    mask2:   dd 0, -1, 0, 0
    mask3:   dd 0, 0, -1, 0


 align 16
;    the_two:
;          dd 4 dup 2.0
    mask_blend:
       dd -1,-1,-1,-1
    i1x4:
       dd  1,1,1,1
    tex_cutter:
       dd TEXTURE_SIZE, TEXTURE_SIZE, TEXTURE_SIZE, TEXTURE_SIZE
    blur_sub:
       dd 0x01010101, 0x01010101, 0x01010101, 0x01010101
 align 16
    muler1234:
       dd 1.0
       dd 2.0
       dd 3.0
       dd 4.0
align 16
    muler0_7:
    muler0123:
       dd 0.0
       dd 1.0
       dd 2.0
       dd 3.0
       dd 4.0
       dd 5.0
       dd 6.0
       dd 7.0
i4x4:
       dd 4, 4, 4, 4
align 16
word_max_f: times 4 dd 65535.0
test_ph:
  dd 110, 390,110,390
mask_65535f:
      times 4 dd 65535.0
mask_line:
           dq 0xffffffff00ffffff
           dq 0xffffffffffffffff
zero_hgst_dd:
           dd 0xffffffff, 0xffffffff, 0xffffffff, 0x00000000
one_hgst_dd:

           dd 0.0, 0.0, 0.0, 1.0
;lpipe_init_add:
;           dd 0,  0,  0.05,  0
f2x4:
           dd 2.0, 2.0, 2.0, 2.0
zero_lowest:
           dd 0x0, 0xffffffff, 0xffffffff, 0xffffffff
zero_hgst:
           dd -1,-1,-1,0
draw_thread_mask:
           dw 0xffff, 0x0000, 0xffff, 0x0000
           dw 0xffff, 0x0000, 0xffff, 0x0000
epsilon:   times 4 dd 0.0000001
abs_mask:
            dd 0x7fffffff, 0x7fffffff, 0x7fffffff, 0x7fffffff
abs_z_coof: dd -1, -1,  0x7fffffff, -1
correct_tex_x4f:
              dd 255.0, 255.0, 255.0, 255.0
z_minus_mask: dd 0 , 0 ,0x80000000, 0
align 32
the_zero:
           dd 0, 0, 0, 0, 0, 0, 0, 0
tex_m2:
           dd 510.0, 510.0, 510.0, 510.0
           dd 510.0, 510.0, 510.0, 510.0
mask_255f:
            dd 255.0, 255.0, 255.0, 255.0
            dd 255.0, 255.0, 255.0, 255.0
tex_fact:   dd 4,2048,4,2048
tex_bi_add: dd 0,0,TEX_X*4,TEX_X*4
tex_bi_shufb db 0,1,8,9,2,3,10,11,4,5,12,13,6,7,14,15
word_max:    dd  65535,65535,65535,65535,65535,65535,65535,65535
big_number:   times 4 dd 999999.0
the_one:
           dd 1.0, 1.0, 1.0 ,1.0

filter     TCHAR   '3ds objects (*.3ds)',0,'*.3ds',0,\
                   'asc objects (*.asc)',0,'*.asc',0,\
                   'All files (*.*)',0,'*.*',0,0
_not3ds    TCHAR   'Not an 3d object file or read error.',0
_notraw    TCHAR   'Not an RAW 512X512 file or read error.',0
filter_tex TCHAR   'raw file 512X512 (*.raw)',0,'*.raw',0,\
                   'All files (*.*)',0,'*.*',0,0

; Sanctssimi Cordis Christi section
   hrt_points:
;cordis
dw 125,175,0, 175,175,0, 212,212,0, 250,175,0, 300,175,0, 325,200,0, 325,250,0
dw 300,300,0, 212,350,0, 125,300,0, 100,250,0, 100,200,0,175,250,-20,250,250,-20
;crux
dw 175,125,0, 200,125,0, 200,100,0, 225,100,0, 225,125,0, 250,125,0, 250,150,0
dw 225,150,0, 225,200,0, 200,200,0, 200,150,0, 175,150,0
;cordis next layer
dw 125,175,20, 175,175,20, 212,212,20, 250,175,20, 300,175,20, 325,200,20
dw 325,250,20, 300,300,20, 212,350,20, 125,300,20, 100,250,20, 100,200,20
dw 175,250,40, 250,250,40
;crux
dw 175,125,20, 200,125,20, 200,100,20, 225,100,20, 225,125,20, 250,125,20
dw 250,150,20, 225,150,20, 225,200,20, 200,200,20, 200,150,20, 175,150,20
;crown
dw 50,250,-25, 50,250,25, 100,250,75, 175,250,100,  275,250,100, 350,250,75
dw 380,250,25, 380,250,-25,350,250,-75,275,250,-100,175,250,-100, 100,250,-75
dw 50,265,-25, 100,265,75, 275,265,100,380,265,25,  350,265,-75, 175,265,-100
;thorns
dw 50,250,-50, 75,260,103, 275,240,150, 365,250,25, 350,260,-41, 150,240,-125

    hrt_points_count = ($-hrt_points)/6
dd -1
    hrt_triangles:
dw 1,0,12, 2,1,12, 9,8,12, 10,9,12, 11,10,12, 0,11,12, 12,13,2, 13,12,8, 3,2,13
dw 4,3,13, 5,4,13, 6,5,13, 7,6,13, 8,7,13
;crux
dw 22,21,23, 24,23,21, 25,24,14, 14,24,18, 18,24,20, 19,18,20, 16,15,18
dw 17,16,18
;cordis next layer
dw 26,27,38, 27,28,38, 34,35,38, 35,36,38, 36,37,38, 37,26,38, 38,39,34, 28,29,39
dw 29,30,39, 30,31,39, 31,32,39, 32,33,39, 33,34,39, 39,38,28
; crux next layer
dw 47,48,49, 49,50,47, 50,51,40, 50,40,44, 50,42,44
dw 42,43,44, 46,50,44, 44,45,46
;border cordis
dw 0,1,26, 1,27,26, 1,2,27, 28,27,2, 2,3,28, 29,28,3, 3,4,29, 30,29,4, 4,5,30
dw 31,30,5, 5,6,31, 32,31,6, 6,7,32, 33,32,7, 7,8,33, 34,33,8, 8,9,34, 35,34,9
dw 9,10,35, 36,35,10, 10,11,36, 37,36,11, 11,0,37, 26,37,0
;border crux
dw 14,15,40, 41,40,15, 15,16,41, 42,41,16, 16,17,42, 43,42,17, 17,18,43
dw 44,43,18, 18,19,44, 45,44,19, 19,20,45, 46,45,20, 20,21,46, 47,46,21
dw 21,22,47, 48,47,22, 23,24,49, 50,49,24, 24,25,50, 51,50,25, 25,14,51
dw 14,40,51
;crown
dw 52,53,64, 53,54,65, 54,55,65, 55,56,66, 56,57,66, 57,58,67, 58,59,67
dw 59,60,68, 60,61,68, 61,62,69, 62,63,69, 63,64,52
;thorns
dw 52,64,70, 54,65,71, 56,66,72, 58,67,73, 68,60,74, 69,62,75
        hrt_triangles_count = (($-hrt_triangles)/6)
dd -1
;box


tetra_points_r:
dd -1.0, -1.0, 1.0
dd  1.0, -1.0, 1.0
dd  0.0,  1.0, 0.0
dd  0.0, -1.0,-1.0
tetra_points_count = 4
dd -1 ;<- end marker

tetra_triangles:
dw 0,1,2, 2,3,0, 2,1,3,  1,0,3
tetra_triangles_count = 4
dd -1 ;<- end marker
tetra_tex_points:
    dw 1,256, 511,511, 511,1
    dw 1,256, 511,511, 511,1
    dw 1,256, 511,511, 511,1
    dw 1,256, 511,511, 511,1
    dd -1
current_seg db 0      ; 255 - no segment selected
max_clip_seg:
    dd ?  ;XRES - 3
    dd ?  ;YRES - 3
min_clip_seg:
    dd 1.0
    dd 1.0
current_seg_add:
    dd 0       ; x ,y ,z
    dd 0
    dd 0
current_seg_add_start:
    dd 0
    dd 0
    dd 0
seg_backward_rot:
    dd 0
    dd 0
    dd 0
maxscaleseg dd 1.0
allsegcenter:
    dd 0, 0, 0, 0
rot_pipe_points_number dd  0

include  'b_rot_var_init.inc'

font_table:
file 'font8x16.bmp'
;mvincit:
;file 'mvincit1.raw'
menu_button_number  dw 0
menu_data:

               dw 0                ; mark if button chosen
               dw 2                ; x1
               dw 2                ; y1
               dw 85               ; x2
               dw 20               ; y2
               db 'rotary    ',0   ; label
               db 4                ; max flag  + 1 , if = 255, no flag
  rotary_flag  db 0                ; flag
               dd axl_f            ; offset to flags description
               dw 1                ; button number
               dw VK_F12           ; key 1
               dw 0                ; key 2 if '0' - none

               dw 0
               dw 2                ; x1
               dw 22               ; y1
               dw 85               ; x2
               dw 40
               db 'mot. blur ',0
               db 2
   blur_flag   db 1
               dd on_of_f
               dw 2
               dw VK_F1
               dw 0

               dw 0
               dw 2                ; x1
               dw 42               ; y1
               dw 85               ; x2
               dw 60
               db 'zoom in   ',0
               db 2
  zoom_in_flag db 0
               dd -1              ; if 0xffffffff - no flag desc
               dw 3
               dw VK_F3
               dw VK_ADD

               dw 0
               dw 2                ; x1
               dw 62               ; y1
               dw 85               ; x2
               dw 80
               db 'zoom out  ',0
               db 2
 zoom_out_flag db 0
               dd -1
               dw 4
               dw VK_F4
               dw VK_SUBTRACT

               dw 0
               dw 2                ; x1
               dw 82               ; y1
               dw 85               ; x2
               dw 100
               db 'tex. mapp.',0
               db 4
    map_flag   db 0
               dd tex_map_f
               dw 5
               dw VK_F5
               dw 0

               dw 0
               dw 2                ; x1
               dw 102               ; y1
               dw 85               ; x2
               dw 120
               db 'ran. light',0
               db 2
       rd_flag db 0
               dd -1
               dw 6
               dw VK_TAB
               dw 0

               dw 0
               dw 2                ; x1
               dw 122               ; y1
               dw 85               ; x2
               dw 140
               db 'dr. model ',0
               db 21
     draw_flag db 0   ; 14 - only nodes (vertices) dbg
               dd -1
               dw 7
               dw VK_SPACE
               dw 0

               dw 0
               dw 2                ; x1
               dw 142               ; y1
               dw 85               ; x2
               dw 160
               db 'bumps     ',0
               db 2
     bump_flag db 0
               dd bumps_f
               dw 8
               dw VK_F6
               dw 0

               dw 0
               dw 2                ; x1
               dw 162              ; y1
               dw 85               ; x2
               dw 180
               db 'texture   ',0
               db 3
     tex_flag  db 0
               dd tex_f
               dw 9
               dw VK_F7
               dw 0

               dw 0
               dw 2                ; x1
               dw 182              ; y1
               dw 85               ; x2
               dw 200
               db 'save obj. ',0
               db 2
     sav_flag  db 0
               dd -1
               dw 10
               dw VK_F8
               dw 0

               dw 0
               dw 2                ; x1
               dw 202              ; y1
               dw 85               ; x2
               dw 220
               db 'run / stop',0
               db 2
  ccounter_inc db 1
               dd run_st_f
               dw 11
               dw VK_F11
               dw 0

               dw 0
               dw 2                ; x1
               dw 222              ; y1
               dw 85               ; x2
               dw 240
               db 'cur.segm. ',0
               db 2 ;BSEGMENTS
               db 0
               dd -1
               dw 12
               dw VK_F9
               dw 0

;               dw 0
;               dw 130                ; x1
;               dw 2               ; y1
;               dw 212               ; x2
;               dw 20
;               db 'load tex. ',0
;               db 2
;               db 0
;               dd -1
;               dw 13
;               dw 0x58          ; x
;               dw 0

               dw 0
               dw 130             ; x1
               dw 2               ; y1
               dw 212             ; x2
               dw 20
               db 'do displ. ',0  ; displacement
               db 2
               db 0
               dd -1
               dw 13
               dw 0x44          ; D
               dw 0

               dw 0
               dw 130              ; x1
               dw 22               ; y1
               dw 212              ; x2
               dw 40
               db 'do triang.',0
               db 2
               db 0
               dd -1
               dw 14
               dw 0x54             ; T
               dw 0

               dw 0
               dw 130               ; x1
               dw 42                ; y1
               dw 212               ; x2
               dw 60                ; y2
               db 'cur object',0
               db 3  ; 6
      hrt_flag db 0
               dd cur_obj_f
               dw 15
               dw VK_F2
               dw 0

if 1
               dw 0                ; mark if button chosen
               dw 130              ; x1
               dw 62               ; y1
               dw 212              ; x2
               dw 80               ; y2
               db 'rand. shd.',0   ; label
               db 255              ; max flag  + 1 , if = 255, no flag
    xchg_flag  db 0                ; flag
               dd -1               ; offset to flags description
               dw 16               ; button number
               dw 0x55             ; key 1
               dw 0                ; key 2 if '0' - none

end if

               dw 0                ; mark if button chosen
               dw 130              ; x1
               dw 82               ; y1
               dw 212              ; x2
               dw 100              ; y2
               db 'edit mesh.',0   ; label
               db 4                ; max flag  + 1 , if = 255, no flag
    edit_flag  db 0                ; flag
               dd edit_f           ; offset to flags description
               dw 17               ; button number
               dw 0x56             ; key 1    'V'
               dw 0                ; key 2 if '0' - none


               dw 0
               dw 130                ; x1
               dw 102                ; y1
               dw 212                ; x2
               dw 120
               db 'set t.area',0
               db 2
 set_tri_area  db 1
               dd on_of_f          ; if 0xffffffff - no flag desc
               dw 18
               dw 0x57             ; W
               dw 0

               dw 0           ; set displacement transparent color
               dw 130         ; x1
               dw 122         ; y1
               dw 212         ; x2
               dw 140
               db 'set t. col',0
               db 2
 disp_col_flag  db 0
               dd displac_f          ; if 0xffffffff - no flag desc
               dw 19
               dw 0x43
               dw 0

               dw 0                ; mark if button chosen
               dw 130              ; x1
               dw 142               ; y1
               dw 212              ; x2
               dw 160              ; y2
               db 'dsp factor',0   ; label
               db 4                ; max flag  + 1 , if = 255, no flag
  disp_fac_flag  db 0                ; flag
               dd d_factor_f          ; offset to flags description
               dw 20               ; button number
               dw 0x46             ; key 1    'F'
               dw 0                ; key 2 if '0' - none


               dw 0                ; mark if button chosen
               dw 130              ; x1
               dw 162              ; y1
               dw 212              ; x2
               dw 180              ; y2
               db 'bezier sr.',0   ; label
               db 4                ; max flag  + 1 , if = 255, no flag
  bezier_flag  db 0                ; flag
               dd bezier_f         ; offset to flags description
               dw 21               ; button number
               dw 0x42             ; 'B'
               dw 0

               dw 0                ; mark if button chosen
               dw 130              ; x1
               dw 182              ; y1
               dw 212              ; x2
               dw 200              ; y2
               db 'set ap tol',0   ; label
               db 7                ; max flag  + 1 , if = 255, no flag
tolerancy_flag db 1                ; flag
               dd approx_toler_f   ; offset to flags description
               dw 22               ; button number
               dw 0x41             ;   'A'
               dw 0

               dw 0                ; mark if button chosen
               dw 130              ; x1
               dw 202              ; y1
               dw 212              ; x2
               dw 220              ; y2
               db 'b.der.edit',0   ; label
               db 2                ; max flag  + 1 , if = 255, no flag
derive_flag    db 1                ; flag
               dd on_of_f          ; offset to flags description
               dw 23               ; button number
               dw 0x45             ;   'e'
               dw 0





               dw 0                ; mark if button chosen
               dw 130              ; x1
               dw 222              ; y1
               dw 212              ; x2
               dw 240              ; y2
               db 't&d wh|pos',0   ; label
               db 2                ; max flag  + 1 , if = 255, no flag
td_wp_flag     db 0                ; triangulation & displacement flag
                                   ; if set on 'posi' - only positive
                                   ; z cooficient vertices are affected,
                                   ; if set on 'whol' all vertices are
                                   ; in triangulation and displacement
                                   ; process affected
               dd t_and_d_f        ; offset to flags description
               dw 24               ; button number
               dw 0x46             ;
               dw 0

               dw 0                ; mark if button chosen
               dw 258              ; x1
               dw 2                ; y1
               dw 340              ; x2
               dw 20               ; y2
               db 'speed     ',0   ; label
               db 2                ; max flag  + 1 , if = 255, no flag
speed_flag     db 0                ; flag
               dd speed_f          ; offset to flags description
               dw 25               ; button number
               dw 0x53             ;   'S'
               dw 0


               dw 0                ; mark if button chosen
               dw 258              ; x1
               dw 22               ; y1
               dw 340              ; x2
               dw 40               ; y2
               db 'rm unu ver',0   ; label
               db 2                ; max flag  + 1 , if = 255, no flag
               db 0                ; flag
               dd -1               ; offset to flags description
               dw 26               ; button number
               dw 0x48             ;   ''
               dw 0

               dw 0                ; mark if button chosen
               dw 258              ; x1
               dw 42               ; y1
               dw 340              ; x2
               dw 60               ; y2
               db 'Long pipe ',0   ; label
               db 3                ; max flag  + 1 , if = 255, no flag
  lpipe_flag   db 0                ; flag
               dd lpipe_f          ; offset to flags description
               dw 27               ; button number
               dw 0x50             ;   'P'
               dw 0

               dw 0                ; mark if button chosen
               dw 258              ; x1
               dw 62               ; y1
               dw 340              ; x2
               dw 80               ; y2
               db 'Lp segs cn',0   ; label
               db 4                ; max flag  + 1 , if = 255, no flag
  lpsegs_flag  db 2                ; flag
               dd lp_segs_f        ; offset to flags description
               dw 28               ; button number
               dw 0x49             ;   ''
               dw 0



               dw 0                ; mark if button chosen
               dw 258              ; x1
               dw 82               ; y1
               dw 340              ; x2
               dw 100              ; y2
               db 'Lp Bez/Bsp',0   ; label
               db 2                ; max flag  + 1 , if = 255, no flag
  lpcurve_tp   db 0                ; flag
               dd lp_cv_type_f     ; offset to flags description
               dw 29               ; button number
               dw 0x49             ;   ''
               dw 0

               dw 0                ; mark if button chosen
               dw 258              ; x1
               dw 102              ; y1
               dw 340              ; x2
               dw 120              ; y2
               db 'Lp quality',0   ; label
               db 4                ; max flag  + 1 , if = 255, no flag
  lp_quality   db 0                ; flag
               dd lp_qual          ; offset to flags description
               dw 30               ; button number
               dw 0x49             ;   ''
               dw 0

               dw 0                ; mark if button chosen
               dw 258              ; x1
               dw 122              ; y1
               dw 340              ; x2
               dw 140              ; y2
               db 'NextM edit',0   ; label   ; next mesh edit
               db 2                ; max flag  + 1 , if = 255, no flag
 NextMed_flag  db 1                ; flag
               dd on_of_f          ; offset to flags description
               dw 31               ; button number
               dw 0x16             ;   ''
               dw 72

               dw 0                ; mark if button chosen
               dw 258              ; x1
               dw 142              ; y1
               dw 340              ; x2
               dw 160              ; y2
               db 'Rm red ver',0   ; label   ; next mesh edit
               db 255              ; max flag  + 1 , if = 255, no flag
               db 0                ; flag
               dd -1               ; offset to flags description
               dw 32               ; button number
               dw 0x49             ;   ''
               dw 0

               dw 0                ; mark if button chosen
               dw 258              ; x1
               dw 162              ; y1
               dw 340              ; x2
               dw 180              ; y2
               db 'Draw norm.',0   ; label
               db 3                ; max flag  + 1 , if = 255, no flag
  normals_flag db 0                ; flag
               dd normals_f        ; offset to flags description
               dw 33               ; button number
               dw 0x49
               dw 0

               dw 0                ; mark if button chosen
               dw 258              ; x1
               dw 182              ; y1
               dw 340              ; x2
               dw 200              ; y2
               db 'Norm. care',0   ; label
               db 2                ; max flag  + 1 , if = 255, no flag
  nr_care_flag db 0                ; flag
               dd of_on_f          ; offset to flags description
               dw 34               ; button number
               dw 0x49
               dw 0

               dw -1  ; <- end flag
normals_f:
     db 'off ',0
     db 'tri.',0
     db 'vert',0
lp_cv_type_f:
     db 'Bezr',0
     db 'Bspl',0

lp_segs_f:
     db ' 2  ',0
     db ' 4  ',0
     db ' 8  ',0
     db ' 11 ',0
lpipe_f:
     db 'off ',0
     db 'edit',0
     db 'calc',0
speed_f:
     db 'idle',0
     db 'full',0

t_and_d_f:
     db 'whol',0
     db 'posi',0



bezier_f:
     db 'off ',0
     db 'pot ',0
     db 'cup ',0
     db 'spoo',0
lp_qual:
approx_toler_f:
d_factor_f:
     db '1   ',0
     db '2   ',0
     db '3   ',0
     db '4   ',0
     db '5   ',0
     db '6   ',0
     db '7   ',0

displac_f:
     db 'whit',0
     db 'blck',0
edit_f:
     db 'off ',0
     db 'vert',0
     db 'face',0
     db 'edge',0
cur_obj_f:
     db 'file',0
     db 'hrt ',0
     db 'tetr',0
     db 'r.s.',0
     db 'pipe',0
     db 'r.sp',0

tex_map_f:
     db 'p.xy',0
     db 'p.xz',0
     db 'p.yz',0
     db 'sph.',0

axl_f:
     db 'y   ',0
     db 'x   ',0
     db 'z   ',0
     db 'cust',0
on_of_f:
     db 'on  ',0
     db 'off ',0
of_on_f:
     db 'off ',0
     db 'on  ',0
bumps_f:
     db 'rand',0
     db 'tex ',0
tex_f:
     db 'str.',0
     db 'xor ',0
     db 'none',0
run_st_f:
    db  'stop',0
    db  'run ',0

    custom_obj_angle   dd 0
    custom_rot_start_x dw 0, 0
    edit_vertex_no  dd 0
    edit_face_no    dd 0
    edit_edge_no    dd 0
    edit_tri_area   db 0
    edit_bezier_x   dw 0
    edit_bezier_y   dw 0
    displ_transparent_col dd 0x00ffffff
;    edit_spline_vert_no   dw 0xffff ;edit Bspline/Bezier curve vars
;    edit_spline_x:         dd 0.0
;    edit_spline_y         dd 0.0

align 16
    tri_area_x1     dd 7.0  ; triangulize area
    tri_area_y1     dd 7.0
    tri_area_x2     dd 2000.0
    tri_area_y2     dd 2000.0
    min_work_x      dw 6
    min_work_y      dw 6
    max_work_x      dw 2000
    max_work_y      dw 2000
    ed_bpatch_derives_x_start: dw 0   ; bezier patch edition
    ed_bpatch_derives_y_start  dw 0
    ed_bpatch_derives_x_end:   dw 0
    ed_bpatch_derives_y_end    dw 0
    ed_bpatch_derive_no        dw 2  ; number
    ed_bpatch_derive_on        dw 0  ; 0 - off, 1 - on


    markshd dd ?

;    mv:
;    file '512x512.raw'
    msg     MSG
    rect    RECT
    ofn     OPENFILENAME sizeof.OPENFILENAME,0,0,filter,0,0,0,fname,260 ;MAXSIZE
    ofntex  OPENFILENAME sizeof.OPENFILENAME,0,0,filter_tex,0,0,0,fnametex,260 ;MAXSIZE
    ofnnext  OPENFILENAME sizeof.OPENFILENAME,0,0,filter,0,0,0,fnamenext,260 ;MAXSIZE
    wc WNDCLASS 0,WindowProc,0,0,NULL,NULL,NULL,1,NULL,_class
    ps PAINTSTRUCT
    ;Bspline_derives:
    long_pipe_derv_init:
       dd 1.0,0.0,0.0
    long_pipe_derv:
       rd 65536   ; mem for control points of
                 ; long pipe
    long_pipe_derv_rotated rd 65536





    Bsplines_vertices   rb 65536

    ThreadID0   dd ?
    ThreadID1   dd ?
    ThreadID2   dd ?
    ThreadID3   dd ?
    ThreadID4   dd ?

    hthread0        rd 1
    hthread1        rd 1
    hthread2        rd 1
    hthread3        rd 1
    hthread4        rd 1

    ThreadID5   dd ?
    ThreadID6   dd ?
    ThreadID7   dd ?
    ThreadID8   dd ?
    ThreadID9   dd ?

    hthread5        rd 1
    hthread6        rd 1
    hthread7        rd 1
    hthread8        rd 1
    hthread9        rd 1

    ThreadID10  dd ?
    ThreadID11   dd ?
    ThreadID12   dd ?
    ThreadID13   dd ?
    ThreadID14   dd ?


    hthread10        rd 1
    hthread11        rd 1
    hthread12        rd 1
    hthread13        rd 1
    hthread14        rd 1
    thread_array:    rd 64



    whole_xres      dw ?
    whole_yres      dw ?
  xres_vard:
    xres_var        dw ?
    yres_var        dw ?
    mwxres          dw ?
    mwyres          dw ?

    edit_start_x    dw ?
    edit_start_y    dw ?
    edit_end_x      dw ?
    edit_end_y      dw ?
    edit_derives_x_start dw ?
    edit_derives_y_stsrt dw ?
    new_tri_ptr     dd ?
    screen_ptr      dd ?
  ;  screenDIB_ptr   dd ?
    Zbuffer_ptr     dd ?
    heap_start_ptr  dd ?
    points_r_ptr    dd ?
    triangles_ptr   dd ?
    edges_ptr       dd ?
    shd_buffer_ptr  dd ?
    edit_buf_ptr    dd ?
    triangles_normals_ptr      dd ?
    triangles_normals_rotated_ptr dd ?
    points_normals_ptr         dd ?
    points_normals_rotated_ptr dd ?
    points_rotated_ptr         dd ?
    tex_points_ptr             dd ?  ; bump_map  and texture coords
                                    ; each point word x, word y
    shadow_vertices_ptr        dd ?

    derv_rect_ptr              dd ?
    derv_nodes_ptr             dd ?
    long_pipe_vert_ptr         dd ?
    slices_counter_ptr         dd ?
    slices_ptrs_buff_ptr       dd ?
    slices_ptr                 dd ?
    voxels_depth_ptr           dd ?
    derv_nodes_rot             rb 9148
    bssurface_derv_unpacked    rb 9148
    bssurface_derv_unpacked_c  rb 9148
    bez_patch_count            dw ?
    bez_nodes_count            dw ?
    bmiheader                  BITMAPINFOHEADER




    matrix                     rb 36
    matrix_scaled              rb 36
    p_light_vec:
              rd 12
    shd_counter dd ?

    segments_add:
              rd BSEGMENTS*3
    segments_rotary:
              rd BSEGMENTS*3    ; x ; angles (in degrees) rotary single Bezier segment
  ;  cur_seg_max:       ; current edited (rotary) bezier segment min and max coordinate
    cur_seg_max_rot:
              rd 8
    cur_seg_center:
              rd 4
    center_scale:
              rd 4
    max_scale dd ?

    tex_mov_begin_x:
                     dw ? ;  \ dont change order
    tex_mov_begin_y  dw ? ;  /
    file_ext_flag    db ?   ; 0 - 3ds, 1 - asc
    fname     rb     260 ;MAXSIZE
    fnametex  rb     260
    fnamenext rb     260
    BytesRead dd ?
    hfile     dd ?
    hfiletex  dd ?
    FileSize  dd ?
    byteswritten  dd ?
    hheap         dd ?
    pmem          dd ?
    pmemtex       dd ?
align 16
    cos_x:
                  rd 1
    sin_x         dd ?
    neg_cos_x     dd ?
    neg_sin_x     dd ?

    cos_y:
                  rd 1
    sin_y         dd ?
    neg_cos_y     dd ?
    neg_sin_y     dd ?

    cos_z:
                  rd 1
    sin_z         dd ?
    neg_cos_z     dd ?
    neg_sin_z     dd ?

    rand_seed dw ?
    hwnd      dd ?
    hdc       dd ?
    bmi       BITMAPINFOHEADER
    Mbmi      BITMAPINFOHEADER
  align 16
    bump_map  rb TEX_X * TEX_Y
    envmap    rd TEX_X * TEX_Y
    color_map rd TEX_X * TEX_Y
    texmap    rd TEX_X * TEX_Y
  align 16
    menu_screen rd MXRES * MYRES
    edges_th_xmm7: rd 4
    edges_th_ebx   dd ?
    edges_th_esi   dd ?
    edges_th_ecx   dd ?

    hrt_points_r:
       rb 65536
    angles  rb 24
    cos_tab rd 360
    sin_tab rd 360
include  'b_rot_var_uninit.inc'


section '.idata' import data readable writeable

    library kernel32,'KERNEL32',user32,'USER32',gdi32,'GDI32',winmm,'WINMM', comdlg32,'COMDLG32'    ;   timing,'timing.dll'
    include '\include\api\kernel32.inc'
    include '\include\api\user32.inc'
    include '\include\api\gdi32.inc'
    include '\include\api\comdlg32.inc'


